@@ -63,9 +63,9 @@
 static char * program_name = "mkswap";
 static char * device_name = NULL;
 static int DEV = -1;
-static long PAGES = 0;
+static unsigned long PAGES = 0;
+static unsigned long badpages = 0;
 static int check = 0;
-static int badpages = 0;
 static int version = -1;
 
 #define MAKE_VERSION(p,q,r)	(65536*(p) + 256*(q) + (r))
@@ -203,10 +203,16 @@ write_signature(char *sig) {
 	strncpy(sp+pagesize-10, sig, 10);
 }
 
+/*
+ * Find out what the maximum amount of swap space is that the kernel will
+ * handle.  This wouldn't matter if the kernel just used as much of the
+ * swap space as it can handle, but until 2.3.4 it would return an error
+ * to swapon() if the swapspace was too large.
+ */
 #define V0_MAX_PAGES	(8 * (pagesize - 10))
 /* Before 2.2.0pre9 */
 #define V1_OLD_MAX_PAGES	((0x7fffffff / pagesize) - 1)
-/* Since 2.2.0pre9:
+/* Since 2.2.0pre9, before 2.3.4:
    error if nr of pages >= SWP_OFFSET(SWP_ENTRY(0,~0UL))
    with variations on
 	#define SWP_ENTRY(type,offset) (((type) << 1) | ((offset) << 8))
@@ -280,8 +286,7 @@ bit_set (unsigned long *addr, unsigned int nr) {
 
 		bitmap += bitnum / (8 * sizeof(long long));
 		rl = *bitmap;
-		ml = 1ULL << (bitnum &
-			      (8ULL * (unsigned long long)sizeof(long long) - 1ULL));
+		ml = 1ULL << (bitnum & (8ULL * sizeof(long long) - 1ULL));
 		*bitmap = rl | ml;
 		return;
 	}
@@ -303,8 +308,7 @@ bit_test_and_clear (unsigned long *addr, unsigned int nr) {
 
 		bitmap += bitnum / (8 * sizeof(long long));
 		rl = *bitmap;
-		ml = 1ULL << (bitnum &
-			      (8ULL * (unsigned long long)sizeof(long long) - 1ULL));
+		ml = 1ULL << (bitnum & (8ULL * sizeof(long long) - 1ULL));
 		*bitmap = rl & ~ml;
 		return ((rl & ml) != 0ULL);
 	}
@@ -375,7 +379,7 @@ check_blocks(void) {
 	if (badpages == 1)
 		printf(_("one bad page\n"));
 	else if (badpages > 1)
-		printf(_("%d bad pages\n"), badpages);
+		printf(_("%lu bad pages\n"), badpages);
 }
 
 static long
@@ -408,10 +412,10 @@ find_size (int fd) {
 }
 
 /* return size in pages, to avoid integer overflow */
-static long
+static unsigned long
 get_size(const char  *file) {
 	int	fd;
-	long	size;
+	unsigned long	size;
 
 	fd = open(file, O_RDONLY);
 	if (fd < 0) {
@@ -436,9 +440,10 @@ isnzdigit(char c) {
 int
 main(int argc, char ** argv) {
 	struct stat statbuf;
-	int i, sz;
-	long maxpages;
-	long goodpages;
+	int i;
+	unsigned long maxpages;
+	unsigned long goodpages;
+	unsigned long sz;
 	off_t offset;
 	int force = 0;
 	char *block_count = 0;
@@ -503,7 +508,7 @@ main(int argc, char ** argv) {
 		   explicitly */
 		char *tmp;
 		int blocks_per_page = pagesize/1024;
-		PAGES = strtol(block_count,&tmp,0)/blocks_per_page;
+		PAGES = strtoul(block_count,&tmp,0)/blocks_per_page;
 		if (*tmp)
 			usage();
 	}
@@ -513,7 +518,7 @@ main(int argc, char ** argv) {
 	} else if (PAGES > sz && !force) {
 		fprintf(stderr,
 			_("%s: error: "
-			  "size %ld is larger than device size %d\n"),
+			  "size %lu is larger than device size %lu\n"),
 			program_name,
 			PAGES*(pagesize/1024), sz*(pagesize/1024));
 		exit(1);
@@ -535,29 +540,28 @@ main(int argc, char ** argv) {
 			program_name, version);
 		usage();
 	}
+
 	if (PAGES < 10) {
 		fprintf(stderr,
 			_("%s: error: swap area needs to be at least %ldkB\n"),
-			program_name, (long)(10 * pagesize / 1024));
+			program_name, (long)(10 * pagesize / 1000));
 		usage();
 	}
-#if 0
-	maxpages = ((version == 0) ? V0_MAX_PAGES : V1_MAX_PAGES);
-#else
-	if (!version)
+
+	if (version == 0)
 		maxpages = V0_MAX_PAGES;
+	else if (linux_version_code() >= MAKE_VERSION(2,3,4))
+		maxpages = PAGES;
 	else if (linux_version_code() >= MAKE_VERSION(2,2,1))
 		maxpages = V1_MAX_PAGES;
-	else {
+	else
 		maxpages = V1_OLD_MAX_PAGES;
-		if (maxpages > V1_MAX_PAGES)
-			maxpages = V1_MAX_PAGES;
-	}
-#endif
+
 	if (PAGES > maxpages) {
 		PAGES = maxpages;
-		fprintf(stderr, _("%s: warning: truncating swap area to %ldkB\n"),
-			program_name, PAGES * pagesize / 1024);
+		fprintf(stderr,
+			_("%s: warning: truncating swap area to %ldkB\n"),
+			program_name, PAGES * pagesize / 1000);
 	}
 
 	DEV = open(device_name,O_RDWR);
@@ -565,6 +569,8 @@ main(int argc, char ** argv) {
 		perror(device_name);
 		exit(1);
 	}
+
+	/* Want a block device. Probably not /dev/hda or /dev/hdb. */
 	if (!S_ISBLK(statbuf.st_mode))
 		check=0;
 	else if (statbuf.st_rdev == 0x0300 || statbuf.st_rdev == 0x0340)
@@ -606,10 +612,10 @@ the -f option to force it.\n"),
 	}
 
 	goodpages = PAGES - badpages - 1;
-	if (goodpages <= 0)
+	if ((long) goodpages <= 0)
 		die(_("Unable to set up swap-space: unreadable"));
-	printf(_("Setting up swapspace version %d, size = %lu KiB\n"),
-		version, (unsigned long)goodpages * pagesize / 1024);
+	printf(_("Setting up swapspace version %d, size = %llu kB\n"),
+		version, (unsigned long long)goodpages * pagesize / 1000);
 	write_signature((version == 0) ? "SWAP-SPACE" : "SWAPSPACE2");
 
 	offset = ((version == 0) ? 0 : 1024);
