@@ -51,6 +51,8 @@
  *  XFS label recognition.
  * Thu Nov 22 15:42:56 CET 2001 <flavio.stanchina@tin.it>
  *  ext3 and ReiserFS recognition.
+ * Sun Oct 12 17:43:43 CEST 2003 <flavio.stanchina@tin.it>
+ *  JFS recognition; ReiserFS label recognition.
  *
  ****************************************************************************/
 
@@ -82,14 +84,8 @@
 #include "xstrncpy.h"
 #include "common.h"
 
-#if defined(__GNUC__) || defined(HAS_LONG_LONG)
-typedef long long ext2_loff_t;
-#else
-typedef long      ext2_loff_t;
-#endif
-
-extern ext2_loff_t ext2_llseek(unsigned int fd, ext2_loff_t offset,
-			       unsigned int origin);
+extern long long ext2_llseek(unsigned int fd, long long offset,
+			     unsigned int origin);
 
 #define VERSION UTIL_LINUX_VERSION
 
@@ -107,8 +103,6 @@ extern ext2_loff_t ext2_llseek(unsigned int fd, ext2_loff_t offset,
 
 #define SECTOR_SIZE 512
 
-#define MAX_CYLINDERS 65535	/* there is no maximum anymore */
-				/* the kernel max is 65535 */
 #define MAX_HEADS 255
 #define MAX_SECTORS 63
 
@@ -126,28 +120,6 @@ extern ext2_loff_t ext2_llseek(unsigned int fd, ext2_loff_t offset,
 #define LINUX_SWAP     0x82
 #define LINUX          0x83
 
-/* There used to be defined error messages here. However, it turns out
- * that gettext cannot handle constructions like
- *
- * #define ADD_EXISTS _("This partition is already in use")
- * ...
- * 	print_warning(ADD_EXISTS);
- *
- * So, now the messages are spread over the source again.
- * Another thing which gettext cannot cope with are multi-line strings:
- *
- *	printf("Usage:
- * Print version:
- *	cfdisk -v
- * Print partition table:
- *	cfdisk -P{r|s|t} device
- * ");
- *
- * (This is a commonly used gnu extension of the C syntax, but not ANSI-C.)
- * Another reason to uglify the source a little.
- */
-
-
 #define PRI_OR_LOG -1
 #define PRIMARY -2
 #define LOGICAL -3
@@ -197,12 +169,13 @@ struct partition {
 
 int heads = 0;
 int sectors = 0;
-int cylinders = 0;
+long long cylinders = 0;
 int cylinder_size = 0;		/* heads * sectors */
-int total_size = 0;		/* actual_size rounded down */
-long actual_size = 0;		/* set using ioctl */
+long long total_size = 0;	/* actual_size rounded down */
+long long actual_size = 0;	/* (in 512-byte sectors) - set using ioctl */
 				/* explicitly given user values */
-int user_heads = 0, user_sectors = 0, user_cylinders = 0;
+int user_heads = 0, user_sectors = 0;
+long long user_cylinders = 0;
 				/* kernel values; ignore the cylinders */
 int kern_heads = 0, kern_sectors = 0;
 				/* partition-table derived values */
@@ -210,7 +183,7 @@ int pt_heads = 0, pt_sectors = 0;
 
 
 static void
-set_hsc0(unsigned char *h, unsigned char *s, int *c, int sector) {
+set_hsc0(unsigned char *h, unsigned char *s, int *c, long long sector) {
 	if (sector >= 1024*cylinder_size)
 		sector = 1024*cylinder_size - 1;
 	*s = sector % sectors + 1;
@@ -221,7 +194,8 @@ set_hsc0(unsigned char *h, unsigned char *s, int *c, int sector) {
 }
 
 static void
-set_hsc(unsigned char *h, unsigned char *s, unsigned char *c, int sector) {
+set_hsc(unsigned char *h, unsigned char *s, unsigned char *c,
+	long long sector) {
 	int cc;
 
 	set_hsc0(h, s, &cc, sector);
@@ -230,12 +204,12 @@ set_hsc(unsigned char *h, unsigned char *s, unsigned char *c, int sector) {
 }
 
 static void
-set_hsc_begin(struct partition *p, int sector) {
+set_hsc_begin(struct partition *p, long long sector) {
 	set_hsc(& p->head, & p->sector, & p->cyl, sector);
 }
 
 static void
-set_hsc_end(struct partition *p, int sector) {
+set_hsc_end(struct partition *p, long long sector) {
 	set_hsc(& p->end_head, & p->end_sector, & p->end_cyl, sector);
 }
 
@@ -259,8 +233,9 @@ store4_little_endian(unsigned char *cp, unsigned int val) {
 
 static unsigned int
 read4_little_endian(unsigned char *cp) {
-	return (uint)(cp[0]) + ((uint)(cp[1]) << 8)
-		+ ((uint)(cp[2]) << 16) + ((uint)(cp[3]) << 24);
+	return (unsigned int)(cp[0]) + ((unsigned int)(cp[1]) << 8)
+		+ ((unsigned int)(cp[2]) << 16)
+		+ ((unsigned int)(cp[3]) << 24);
 }
 
 static void
@@ -298,9 +273,9 @@ typedef union {
 } partition_table;
 
 typedef struct {
-    int first_sector;	/* first sector in partition */
-    int last_sector;	/* last sector in partition */
-    int offset;		/* offset from first sector to start of data */
+    long long first_sector;	/* first sector in partition */
+    long long last_sector;	/* last sector in partition */
+    long offset;		/* offset from first sector to start of data */
     int flags;		/* active == 0x80 */
     int id;		/* filesystem type */
     int num;		/* number of partition -- primary vs. logical */
@@ -353,7 +328,7 @@ int FLAGS_START = 16;
 int PTYPE_START = 28;
 int FSTYPE_START = 38;
 int LABEL_START = 54;
-int SIZE_START = 70;
+int SIZE_START = 68;
 int COMMAND_LINE_X = 5;
 
 static void die_x(int ret);
@@ -404,6 +379,8 @@ partition_type_text(int i) {
 	      return _("Linux ext3");
 	 else if (!strcmp(p_info[i].fstype, "xfs"))
 	      return _("Linux XFS");
+	 else if (!strcmp(p_info[i].fstype, "jfs"))
+	      return _("Linux JFS");
 	 else if (!strcmp(p_info[i].fstype, "reiserfs"))
 	      return _("Linux ReiserFS");
 	 else
@@ -428,8 +405,10 @@ fdexit(int ret) {
 
     if (changed) {
 	fprintf(stderr, _("Disk has been changed.\n"));
+#if 0
 	fprintf(stderr, _("Reboot the system to ensure the partition "
 			"table is correctly updated.\n"));
+#endif
 	
 	fprintf( stderr, _("\nWARNING: If you have created or modified any\n"
 		         "DOS 6.x partitions, please see the cfdisk manual\n"
@@ -441,7 +420,7 @@ fdexit(int ret) {
 
 static int
 get_string(char *str, int len, char *def) {
-    char c;
+    unsigned char c;
     int i = 0;
     int x, y;
     int use_def = FALSE;
@@ -572,16 +551,16 @@ die_x(int ret) {
 }
 
 static void
-read_sector(char *buffer, int sect_num) {
-    if (ext2_llseek(fd, ((ext2_loff_t) sect_num)*SECTOR_SIZE, SEEK_SET) < 0)
+read_sector(char *buffer, long long sect_num) {
+    if (ext2_llseek(fd, sect_num*SECTOR_SIZE, SEEK_SET) < 0)
 	fatal(_("Cannot seek on disk drive"), 2);
     if (read(fd, buffer, SECTOR_SIZE) != SECTOR_SIZE)
 	fatal(_("Cannot read disk drive"), 2);
 }
 
 static void
-write_sector(char *buffer, int sect_num) {
-    if (ext2_llseek(fd, ((ext2_loff_t) sect_num)*SECTOR_SIZE, SEEK_SET) < 0)
+write_sector(char *buffer, long long sect_num) {
+    if (ext2_llseek(fd, sect_num*SECTOR_SIZE, SEEK_SET) < 0)
 	fatal(_("Cannot seek on disk drive"), 2);
     if (write(fd, buffer, SECTOR_SIZE) != SECTOR_SIZE)
 	fatal(_("Cannot write disk drive"), 2);
@@ -605,10 +584,9 @@ get_dos_label(int i) {
 #define DOS_OSTYPE_SZ 8
 #define DOS_LABEL_SZ 11
 #define DOS_FSTYPE_SZ 8
-	ext2_loff_t offset;
+	long long offset;
 
-	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
-		* SECTOR_SIZE;
+	offset = (p_info[i].first_sector + p_info[i].offset) * SECTOR_SIZE;
 	if (ext2_llseek(fd, offset, SEEK_SET) == offset
 	    && read(fd, &sector, sizeof(sector)) == sizeof(sector)) {
 		dos_copy_to_info(p_info[i].ostype, OSTYPESZ,
@@ -623,17 +601,26 @@ get_dos_label(int i) {
 #define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
 #define REISER2FS_SUPER_MAGIC_STRING "ReIsEr2Fs"
 struct reiserfs_super_block {
-	char s_dummy0[ 52];
-	char s_magic [ 12];
-	char s_dummy1[140];
+	char s_dummy0[52];
+	char s_magic [10];
+	char s_dummy1[38];
+	u_char s_label[16];
 };
+#define REISERFSLABELSZ sizeof(reiserfsb.s_label)
 
 static int
-is_reiserfs_magic_string(const struct reiserfs_super_block *rs) {
-	return (!strncmp(rs->s_magic, REISERFS_SUPER_MAGIC_STRING,
-			 strlen(REISERFS_SUPER_MAGIC_STRING)) ||
-		!strncmp(rs->s_magic, REISER2FS_SUPER_MAGIC_STRING,
-			 strlen(REISER2FS_SUPER_MAGIC_STRING)));
+has_reiserfs_magic_string(const struct reiserfs_super_block *rs, int *is_3_6) {
+	if (!strncmp(rs->s_magic, REISERFS_SUPER_MAGIC_STRING,
+		     strlen(REISERFS_SUPER_MAGIC_STRING))) {
+		*is_3_6 = 0;
+		return 1;
+	}
+	if (!strncmp(rs->s_magic, REISER2FS_SUPER_MAGIC_STRING,
+		     strlen(REISER2FS_SUPER_MAGIC_STRING))) {
+		*is_3_6 = 1;
+		return 1;
+	}
+	return 0;
 }
 
 static void
@@ -655,6 +642,20 @@ get_linux_label(int i) {
 
 #define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)
 	struct reiserfs_super_block reiserfsb;
+	int reiserfs_is_3_6;
+
+#define JFS_SUPER1_OFF	0x8000
+#define JFS_MAGIC	"JFS1"
+#define JFSLABELSZ	16
+	struct jfs_super_block {
+		char    s_magic[4];
+		u_char  s_version[4];
+		u_char  s_dummy1[93];
+		char    s_fpack[11];
+		u_char  s_dummy2[24];
+		u_char  s_uuid[16];
+		char    s_label[JFSLABELSZ];
+	} jfsb;
 
 #define XFS_SUPER_MAGIC "XFSB"
 #define XFSLABELSZ 12
@@ -666,11 +667,11 @@ get_linux_label(int i) {
 	} xfsb;
 
 	char *label;
-	ext2_loff_t offset;
+	long long offset;
 	int j;
 
-	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
-		* SECTOR_SIZE + 1024;
+	offset = (p_info[i].first_sector + p_info[i].offset) * SECTOR_SIZE
+		+ 1024;
 	if (ext2_llseek(fd, offset, SEEK_SET) == offset
 	    && read(fd, &e2fsb, sizeof(e2fsb)) == sizeof(e2fsb)
 	    && e2fsb.s_magic[0] + (e2fsb.s_magic[1]<<8) == EXT2_SUPER_MAGIC) {
@@ -686,11 +687,10 @@ get_linux_label(int i) {
 		return;
 	}
 
-	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
-		* SECTOR_SIZE + 0;
+	offset = (p_info[i].first_sector + p_info[i].offset) * SECTOR_SIZE + 0;
 	if (ext2_llseek(fd, offset, SEEK_SET) == offset
 	    && read(fd, &xfsb, sizeof(xfsb)) == sizeof(xfsb)
-	    && !strcmp(xfsb.s_magic, XFS_SUPER_MAGIC)) {
+	    && !strncmp(xfsb.s_magic, XFS_SUPER_MAGIC, 4)) {
 		label = xfsb.s_fname;
 		for(j=0; j<XFSLABELSZ && j<LABELSZ && isprint(label[j]); j++)
 			p_info[i].volume_label[j] = label[j];
@@ -699,12 +699,34 @@ get_linux_label(int i) {
 		return;
 	}
 
+	/* jfs? */
+	offset = (p_info[i].first_sector + p_info[i].offset) * SECTOR_SIZE
+		+ JFS_SUPER1_OFF;
+	if (ext2_llseek(fd, offset, SEEK_SET) == offset
+	    && read(fd, &jfsb, sizeof(jfsb)) == sizeof(jfsb)
+	    && !strncmp(jfsb.s_magic, JFS_MAGIC, strlen(JFS_MAGIC))) {
+		label = jfsb.s_label;
+		for(j=0; j<JFSLABELSZ && j<LABELSZ && isprint(label[j]); j++)
+			p_info[i].volume_label[j] = label[j];
+		p_info[i].volume_label[j] = 0;
+		strncpy(p_info[i].fstype, "jfs", FSTYPESZ);
+		return;
+	}
+
 	/* reiserfs? */
-	offset = ((ext2_loff_t) p_info[i].first_sector + p_info[i].offset)
-		* SECTOR_SIZE + REISERFS_DISK_OFFSET_IN_BYTES;
+	offset = (p_info[i].first_sector + p_info[i].offset) * SECTOR_SIZE
+		+ REISERFS_DISK_OFFSET_IN_BYTES;
 	if (ext2_llseek(fd, offset, SEEK_SET) == offset
 	    && read(fd, &reiserfsb, 1024) == 1024
-	    && is_reiserfs_magic_string(&reiserfsb)) {
+	    && has_reiserfs_magic_string(&reiserfsb, &reiserfs_is_3_6)) {
+		if (reiserfs_is_3_6) {
+			/* label only on version 3.6 onward */
+			label = reiserfsb.s_label;
+			for(j=0; j<REISERFSLABELSZ && j<LABELSZ &&
+				    isprint(label[j]); j++)
+				p_info[i].volume_label[j] = label[j];
+			p_info[i].volume_label[j] = 0;
+		}
 		strncpy(p_info[i].fstype, "reiserfs", FSTYPESZ);
 		return;
 	}
@@ -804,7 +826,7 @@ remove_part(int i) {
 }
 
 static void
-insert_empty_part(int i, int first, int last) {
+insert_empty_part(int i, long long first, long long last) {
     int p;
 
     for (p = num_parts; p > i; p--)
@@ -879,8 +901,8 @@ del_part(int i) {
 }
 
 static int
-add_part(int num, int id, int flags, int first, int last, int offset,
-	 int want_label, char **errmsg) {
+add_part(int num, int id, int flags, long long first, long long last,
+	 long long offset, int want_label, char **errmsg) {
     int i, pri = 0, log = 0;
 
     if (num_parts == MAXIMUM_PARTS) {
@@ -903,11 +925,16 @@ add_part(int num, int id, int flags, int first, int last, int offset,
 	return -1;
     }
 
-    if (last >= total_size) {
+    if (last >= actual_size) {
 	*errmsg = _("Partition ends after end-of-disk");
 	return -1;
     }
 
+    if (last >= total_size) {
+	*errmsg = _("Partition ends in the final partial cylinder");
+	return -1;
+    }
+
     for (i = 0; i < num_parts; i++) {
 	if (p_info[i].id > 0 && IS_PRIMARY(p_info[i].num))
 	    pri++;
@@ -933,6 +960,8 @@ add_part(int num, int id, int flags, int first, int last, int offset,
 		  p_info[i].first_sector + p_info[i].offset <= last)
 	      *errmsg = _("logical partitions overlap");
 	 else
+	      /* the enlarged logical partition starts at the
+		 partition table sector that defines it */
 	      *errmsg = _("enlarged logical partitions overlap");
 	 return -1;
     }
@@ -1062,16 +1091,10 @@ find_logical(int i) {
     return num;
 }
 
-static void
-inc_logical(int i) {
-    int j;
-
-    for (j = i; j < num_parts; j++)
-	if (p_info[j].id > 0 && IS_LOGICAL(p_info[j].num))
-	    p_info[j].num++;
-}
-
-/* Command menu support by Janne Kukonlehto <jtklehto@phoenix.oulu.fi> September 1994 */
+/*
+ * Command menu support by Janne Kukonlehto <jtklehto@phoenix.oulu.fi>
+ * September 1994
+ */
 
 /* Constants for menuType parameter of menuSelect function */
 #define MENU_HORIZ 1
@@ -1138,6 +1161,9 @@ menuUpdate( int y, int x, struct MenuItem *menuItems, int itemLength,
 #endif
 	if (lenName >= sizeof(buff)) {	/* truncate ridiculously long string */
 	    xstrncpy(buff, mi, sizeof(buff));
+	} else if (lenName >= itemLength) {
+            snprintf(buff, sizeof(buff),
+		     (menuType & MENU_BUTTON) ? "[%s]" : "%s", mi);
 	} else {
             snprintf(buff, sizeof(buff),
 		     (menuType & MENU_BUTTON) ? "[%*s%-*s]" : "%*s%-*s",
@@ -1179,60 +1205,57 @@ menuUpdate( int y, int x, struct MenuItem *menuItems, int itemLength,
 }
 
 /* This function takes a list of menu items, lets the user choose one *
- * and returns the value keyboard shortcut of the selected menu item  */
+ * and returns the keyboard shortcut value of the selected menu item  */
 
 static int
 menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
 	    char *available, int menuType, int menuDefault ) {
     int i, ylast = y, key = 0, current = menuDefault;
 
-    if( !( menuType & ( MENU_HORIZ | MENU_VERT ) ) )	
-    {
+    if( !( menuType & ( MENU_HORIZ | MENU_VERT ) ) ) {
         print_warning(_("Menu without direction. Defaulting horizontal."));
         menuType |= MENU_HORIZ;
     }
 
     /* Make sure that the current is one of the available items */
-    while( !strchr(available, menuItems[current].key) )
-    {
+    while( !strchr(available, menuItems[current].key) ) {
         current ++ ;
         if( !menuItems[current].key ) current = 0;
     }
 
     /* Repeat until allowable choice has been made */
-    while( !key )
-    {
-        /* Display the menu */
+    while( !key ) {
+        /* Display the menu and read a command */
         ylast = menuUpdate( y, x, menuItems, itemLength, available,
 			    menuType, current );
         refresh();
         key = getch();
+
         /* Clear out all prompts and such */
         clear_warning();
-        for( i = y; i < ylast; i ++ )
-        {
-            move( i, x );
+        for (i = y; i < ylast; i++) {
+            move(i, x);
             clrtoeol();
         }
         move( WARNING_START + 1, 0 );
         clrtoeol();
-        /* Cursor keys */
-        if( key == ESC )
-        {
+
+        /* Cursor keys - possibly split by slow connection */
+        if( key == ESC ) {
             /* Check whether this is a real ESC or one of extended keys */
             /*nodelay(stdscr, TRUE);*/
             key = getch();
             /*nodelay(stdscr, FALSE);*/
-            if( key == /*ERR*/ ESC )
-            {
+
+            if( key == /*ERR*/ ESC ) {
                 /* This is a real ESC */
                 key = ESC;
             }
-            if( key == '[' )
-            {
+            if(key == '[' || key == 'O') {
                 /* This is one extended keys */
-                switch( getch() )
-                {
+		key = getch();
+
+                switch(key) {
                     case 'A': /* Up arrow */
 			key = MENU_UP;
                         break;
@@ -1245,21 +1268,23 @@ menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
                     case 'D': /* Left arrow */
                         key = MENU_LEFT;
                         break;
+		    default:
+			key = 0;
                 }
             }
         }
 
-        /* Enter equals to the keyboard shortcut of current menu item */
-        if( key == CR)
+        /* Enter equals the keyboard shortcut of current menu item */
+        if (key == CR)
             key = menuItems[current].key;
 
 	/* Give alternatives for arrow keys in case the window manager
 	   swallows these */
-	if ( key == TAB )
+	if (key == TAB)
 	    key = MENU_RIGHT;
-        if ( key == UPKEY )	/* ^P */
+        if (key == UPKEY)	/* ^P */
 	    key = MENU_UP;
-	if ( key == DOWNKEY )	/* ^N */
+	if (key == DOWNKEY)	/* ^N */
 	    key = MENU_DOWN;
 
 	if (key == MENU_UP) {
@@ -1310,21 +1335,22 @@ menuSelect( int y, int x, struct MenuItem *menuItems, int itemLength,
 
         /* Should all keys to be accepted? */
         if( key && (menuType & MENU_ACCEPT_OTHERS) ) break;
-        /* Is pressed key among acceptable ones */
+
+        /* Is pressed key among acceptable ones? */
         if( key && (strchr(available, tolower(key)) || strchr(available, key)))
 	    break;
+
         /* The key has not been accepted so far -> let's reject it */
-        if( key )
-        {
+        if (key) {
             key = 0;
             putchar( BELL );
             print_warning(_("Illegal key"));
         }
     }
+
     /* Clear out prompts and such */
     clear_warning();
-    for( i = y; i <= ylast; i ++ )
-    {
+    for( i = y; i <= ylast; i ++ ) {
         move( i, x );
         clrtoeol();
     }
@@ -1374,13 +1400,13 @@ static void
 new_part(int i) {
     char response[LINE_LENGTH], def[LINE_LENGTH];
     char c;
-    int first = p_info[i].first_sector;
-    int last = p_info[i].last_sector;
-    int offset = 0;
+    long long first = p_info[i].first_sector;
+    long long last = p_info[i].last_sector;
+    long long offset = 0;
     int flags = 0;
     int id = LINUX;
     int num = -1;
-    int num_sects = last - first + 1;
+    long long num_sects = last - first + 1;
     int len, ext, j;
     char *errmsg;
     double sectors_per_MB = K*K / 512.0;
@@ -1465,19 +1491,35 @@ new_part(int i) {
 	    print_warning(_("No room to create the extended partition"));
 	    return;
 	}
-	(void) add_part(ext, DOS_EXTENDED, 0, first, last,
-		       (first == 0 ? sectors : 0), 0, &errmsg);
+	errmsg = 0;
+	if (add_part(ext, DOS_EXTENDED, 0, first, last,
+		     (first == 0 ? sectors : 0), 0, &errmsg) && errmsg)
+		print_warning(errmsg);
 	first = ext_info.first_sector + ext_info.offset;
     }
 
-    if (IS_LOGICAL(num))
-	inc_logical(i);
+    /* increment number of all partitions past this one */
+    if (IS_LOGICAL(num)) {
+#if 0
+	/* original text - ok, but fails when partitions not in disk order */
+	for (j = i; j < num_parts; j++)
+	    if (p_info[j].id > 0 && IS_LOGICAL(p_info[j].num))
+		p_info[j].num++;
+#else
+	/* always ok */
+	for (j = 0; j < num_parts; j++)
+	    if (p_info[j].id > 0 && p_info[j].num >= num)
+		p_info[j].num++;
+#endif
+    }
 
     /* Now we have a complete partition to ourselves */
     if (first == 0 || IS_LOGICAL(num))
 	offset = sectors;
 
-    (void) add_part(num, id, flags, first, last, offset, 0, &errmsg);
+    errmsg = 0;
+    if (add_part(num, id, flags, first, last, offset, 0, &errmsg) && errmsg)
+	    print_warning(errmsg);
 }
 
 static void
@@ -1492,57 +1534,85 @@ get_kernel_geometry(void) {
 #endif
 }
 
+static int
+said_yes(char answer) {
+#ifdef HAVE_rpmatch
+	char reply[2];
+	int yn;
+
+	reply[0] = answer;
+	reply[1] = 0;
+	yn = rpmatch(reply);	/* 1: yes, 0: no, -1: ? */
+	if (yn >= 0)
+		return yn;
+#endif
+	return (answer == 'y' || answer == 'Y');
+}
+
 static void
 get_partition_table_geometry(partition_table *bufp) {
-    struct partition *p;
-    int i,h,s,hh,ss;
-    int first = TRUE;
-    int bad = FALSE;
+	struct partition *p;
+	int i,h,s,hh,ss;
+	int first = TRUE;
+	int bad = FALSE;
 
-    if (bufp->p.magicflag[0] != PART_TABLE_FLAG0 ||
-	bufp->p.magicflag[1] != PART_TABLE_FLAG1) {
-	    /* Matthew Wilcox: slightly friendlier version of
-	       fatal(_("Bad signature on partition table"), 3);
-	    */
-	    int cont;
-	    mvaddstr(WARNING_START, 0,
-	      _("No partition table or unknown signature on partition table"));
-	    mvaddstr(WARNING_START+1, 0,
-	      _("Do you wish to start with a zero table [y/N] ?"));
-	    putchar(BELL);
-	    refresh();
-	    cont = getch();
-	    if ((cont != 'y') && (cont != 'Y'))
-		    die_x(3);
-	    zero_table = TRUE;
-	    return;
+	for (i=0; i<66; i++)
+		if (bufp->c.b[446+i])
+			goto nonz;
 
-	    /* Oskar Liljeblad suggested:
-	         Bad signature blah blah
-		 If this is a brand new harddrive that has not been partitioned
-		 before, please run cfdisk -z.
-	    */
-    }
-
-    hh = ss = 0;
-    for (i=0; i<4; i++) {
-	p = &(bufp->p.part[i]);
-	if (p->sys_ind != 0) {
-	    h = p->end_head + 1;
-	    s = (p->end_sector & 077);
-	    if (first) {
-		hh = h;
-		ss = s;
-		first = FALSE;
-	    } else if (hh != h || ss != s)
-		bad = TRUE;
+	/* zero table */
+	if (!curses_started) {
+		fatal(_("No partition table.\n"), 3);
+		return;
+	} else {
+		mvaddstr(WARNING_START, 0,
+			 _("No partition table. Starting with zero table."));
+		putchar(BELL);
+		refresh();
+		zero_table = TRUE;
+		return;
 	}
-    }
+ nonz:
+	if (bufp->p.magicflag[0] != PART_TABLE_FLAG0 ||
+	    bufp->p.magicflag[1] != PART_TABLE_FLAG1) {
+		if (!curses_started)
+			fatal(_("Bad signature on partition table"), 3);
 
-    if (!first && !bad) {
-	pt_heads = hh;
-	pt_sectors = ss;
-    }
+		/* Matthew Wilcox */
+		mvaddstr(WARNING_START, 0,
+			 _("Unknown partition table type"));
+		mvaddstr(WARNING_START+1, 0,
+			 _("Do you wish to start with a zero table [y/N] ?"));
+		putchar(BELL);
+		refresh();
+		{
+			int cont = getch();
+			if (cont == EOF || !said_yes(cont))
+				die_x(3);
+		}
+		zero_table = TRUE;
+		return;
+	}
+
+	hh = ss = 0;
+	for (i=0; i<4; i++) {
+		p = &(bufp->p.part[i]);
+		if (p->sys_ind != 0) {
+			h = p->end_head + 1;
+			s = (p->end_sector & 077);
+			if (first) {
+				hh = h;
+				ss = s;
+				first = FALSE;
+			} else if (hh != h || ss != s)
+				bad = TRUE;
+		}
+	}
+
+	if (!first && !bad) {
+		pt_heads = hh;
+		pt_sectors = ss;
+	}
 }
 
 static void
@@ -1555,7 +1625,7 @@ decide_on_geometry(void) {
 	       kern_sectors ? kern_sectors : 63);
     cylinder_size = heads*sectors;
     cylinders = actual_size/cylinder_size;
-    if (user_cylinders > 0 && user_cylinders <= 0x7fffffff/cylinder_size)
+    if (user_cylinders > 0)
 	    cylinders = user_cylinders;
 
     total_size = cylinder_size*cylinders;
@@ -1583,7 +1653,9 @@ clear_p_info(void) {
 
 static void
 fill_p_info(void) {
-    int pn, i, bs, bsz;
+    int pn, i;
+    long long bs, bsz;
+    unsigned long long llsectors;
     struct partition *p;
     partition_table buffer;
     partition_info tmp_ext = { 0, 0, 0, 0, FREE_SPACE, PRIMARY };
@@ -1612,8 +1684,9 @@ fill_p_info(void) {
     ioctl(fd, BLKFLSBUF);	/* ignore errors */
 				/* e.g. Permission Denied */
 
-    if (ioctl(fd, BLKGETSIZE, &actual_size))
-	fatal(_("Cannot get disk size"), 3);
+    if (disksize(fd, &llsectors))
+	    fatal(_("Cannot get disk size"), 3);
+    actual_size = llsectors;
 
     read_sector(buffer.c.b, 0);
 
@@ -1670,7 +1743,7 @@ fill_p_info(void) {
 			     bs, 1, &errmsg)) {
 				char *bad = _("Bad logical partition");
 				char *msg = (char *) xmalloc(strlen(bad) + strlen(errmsg) + 30);
-				sprintf(msg, "%s %d: %s", bad, i-1, errmsg);
+				sprintf(msg, "%s %d: %s", bad, i, errmsg);
 				fatal(msg, 4);
 			}
 		}
@@ -1692,7 +1765,7 @@ fill_p_info(void) {
 
 static void
 fill_part_table(struct partition *p, partition_info *pi) {
-    int begin;
+    long long begin;
 
     p->boot_ind = pi->flags;
     p->sys_ind = pi->id;
@@ -1945,7 +2018,7 @@ print_raw_table(void) {
 
 static void
 print_p_info_entry(FILE *fp, partition_info *p) {
-    int size;
+    long long size;
     char part_str[40];
 
     if (p->id == UNUSABLE)
@@ -1962,17 +2035,17 @@ print_p_info_entry(FILE *fp, partition_info *p) {
 
     fp_printf(fp, " ");
 
-    fp_printf(fp, "%8d%c", p->first_sector,
+    fp_printf(fp, "%11lld%c", p->first_sector,
 	      ((p->first_sector/cylinder_size) !=
 	       ((float)p->first_sector/cylinder_size) ?
 	       '*' : ' '));
 
-    fp_printf(fp, "%8d%c", p->last_sector,
+    fp_printf(fp, "%11lld%c", p->last_sector,
 	      (((p->last_sector+1)/cylinder_size) !=
 	       ((float)(p->last_sector+1)/cylinder_size) ?
 	       '*' : ' '));
 
-    fp_printf(fp, "%7d%c", p->offset,
+    fp_printf(fp, "%6ld%c", p->offset,
 	      ((((p->first_sector == 0 || IS_LOGICAL(p->num)) &&
 		 (p->offset != sectors)) ||
 		(p->first_sector != 0 && IS_PRIMARY(p->num) &&
@@ -1980,30 +2053,30 @@ print_p_info_entry(FILE *fp, partition_info *p) {
 	       '#' : ' '));
 
     size = p->last_sector - p->first_sector + 1;
-    fp_printf(fp, "%8d%c", size,
+    fp_printf(fp, "%11lld%c", size,
 	      ((size/cylinder_size) != ((float)size/cylinder_size) ?
 	       '*' : ' '));
 
-    fp_printf(fp, " ");
+    /* fp_printf(fp, " "); */
 
     if (p->id == UNUSABLE)
-	sprintf(part_str, "%.17s", _("Unusable"));
+	sprintf(part_str, "%.15s", _("Unusable"));
     else if (p->id == FREE_SPACE)
-	sprintf(part_str, "%.17s", _("Free Space"));
+	sprintf(part_str, "%.15s", _("Free Space"));
     else if (partition_type_name(p->id))
-	sprintf(part_str, "%.17s (%02X)", partition_type_name(p->id), p->id);
+	sprintf(part_str, "%.15s (%02X)", partition_type_name(p->id), p->id);
     else
-	sprintf(part_str, "%.17s (%02X)", _("Unknown"), p->id);
-    fp_printf(fp, "%-22.22s", part_str);
+	sprintf(part_str, "%.15s (%02X)", _("Unknown"), p->id);
+    fp_printf(fp, "%-20.20s", part_str);
 
     fp_printf(fp, " ");
 
     if (p->flags == ACTIVE_FLAG)
-	fp_printf(fp, _("Boot (%02X)"), p->flags);
+	fp_printf(fp, _("Boot"), p->flags);
     else if (p->flags != 0)
-	fp_printf(fp, _("Unknown (%02X)"), p->flags);
+	fp_printf(fp, _("(%02X)"), p->flags);
     else
-	fp_printf(fp, _("None (%02X)"), p->flags);
+	fp_printf(fp, _("None"), p->flags);
 
     fp_printf(fp, "\n");
 }
@@ -2040,9 +2113,9 @@ print_p_info(void) {
 
     fp_printf(fp, _("Partition Table for %s\n"), disk_device);
     fp_printf(fp, "\n");
-    fp_printf(fp, _("            First    Last\n"));
-    fp_printf(fp, _(" # Type     Sector   Sector   Offset  Length   Filesystem Type (ID)   Flags\n"));
-    fp_printf(fp, _("-- ------- -------- --------- ------ --------- ---------------------- ---------\n"));
+    fp_printf(fp, _("               First       Last\n"));
+    fp_printf(fp, _(" # Type       Sector      Sector   Offset    Length   Filesystem Type (ID) Flag\n"));
+    fp_printf(fp, _("-- ------- ----------- ----------- ------ ----------- -------------------- ----\n"));
 
     for (i = 0; i < num_parts; i++) {
 	if (pext && (p_info[i].first_sector >= ext_info.first_sector)) {
@@ -2062,7 +2135,7 @@ print_p_info(void) {
 
 static void
 print_part_entry(FILE *fp, int num, partition_info *pi) {
-    int first = 0, start = 0, end = 0, size = 0;
+    long long first = 0, start = 0, end = 0, size = 0;
     unsigned char ss, es, sh, eh;
     int sc, ec;
     int flags = 0, id = 0;
@@ -2087,7 +2160,7 @@ print_part_entry(FILE *fp, int num, partition_info *pi) {
 	set_hsc0(&eh, &es, &ec, end);
     }
 
-    fp_printf(fp, "%2d  0x%02X %4d %4d %4d 0x%02X %4d %4d %4d %8d %9d\n",
+    fp_printf(fp, "%2d  0x%02X %4d %4d %4d 0x%02X %4d %4d %4d %11lld %11lld\n",
 	      num+1, flags, sh, ss, sc, id, eh, es, ec, first, size);
 }
 
@@ -2125,9 +2198,9 @@ print_part_table(void) {
     fp_printf(fp, _("Partition Table for %s\n"), disk_device);
     fp_printf(fp, "\n");
     /* Three-line heading. Read "Start Sector" etc vertically. */
-    fp_printf(fp, _("         ---Starting---      ----Ending----    Start Number of\n"));
-    fp_printf(fp, _(" # Flags Head Sect Cyl   ID  Head Sect Cyl    Sector  Sectors\n"));
-    fp_printf(fp, _("-- ----- ---- ---- ---- ---- ---- ---- ---- -------- ---------\n"));
+    fp_printf(fp, _("         ---Starting---      ----Ending----    Start     Number of\n"));
+    fp_printf(fp, _(" # Flags Head Sect Cyl   ID  Head Sect Cyl     Sector    Sectors\n"));
+    fp_printf(fp, _("-- ----- ---- ---- ---- ---- ---- ---- ---- ----------- -----------\n"));
 
     for (i = 0; i < 4; i++) {
 	for (j = 0;
@@ -2256,7 +2329,8 @@ change_geometry(void) {
     int done = FALSE;
     char def[LINE_LENGTH];
     char response[LINE_LENGTH];
-    int tmp_val, max_cyls, i;
+    long long tmp_val;
+    int i;
 
     while (!done) {
         static struct MenuItem menuGeometry[]=
@@ -2275,7 +2349,7 @@ change_geometry(void) {
 
 	switch (toupper( menuSimple(menuGeometry, 3) )) {
 	case 'C':
-	    sprintf(def, "%ld", actual_size/cylinder_size);
+	    sprintf(def, "%llu", actual_size/cylinder_size);
 	    mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X,
 		     _("Enter the number of cylinders: "));
 	    i = get_string(response, LINE_LENGTH, def);
@@ -2283,9 +2357,8 @@ change_geometry(void) {
 		user_cylinders = actual_size/cylinder_size;
 		ret_val = TRUE;
 	    } else if (i > 0) {
-		tmp_val = atoi(response);
-		max_cyls = 0x7fffffff / cylinder_size;
-		if (tmp_val > 0 && tmp_val <= max_cyls) {
+		tmp_val = atoll(response);
+		if (tmp_val > 0) {
 		    user_cylinders = tmp_val;
 		    ret_val = TRUE;
 		} else
@@ -2297,7 +2370,7 @@ change_geometry(void) {
 	    mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X,
 		     _("Enter the number of heads: "));
 	    if (get_string(response, LINE_LENGTH, def) > 0) {
-		tmp_val = atoi(response);
+		tmp_val = atoll(response);
 		if (tmp_val > 0 && tmp_val <= MAX_HEADS) {
 		    user_heads = tmp_val;
 		    ret_val = TRUE;
@@ -2310,7 +2383,7 @@ change_geometry(void) {
 	    mvaddstr(COMMAND_LINE_Y, COMMAND_LINE_X,
 		     _("Enter the number of sectors per track: "));
 	    if (get_string(response, LINE_LENGTH, def) > 0) {
-		tmp_val = atoi(response);
+		tmp_val = atoll(response);
 		if (tmp_val > 0 && tmp_val <= MAX_SECTORS) {
 		    user_sectors = tmp_val;
 		    ret_val = TRUE;
@@ -2334,7 +2407,7 @@ change_geometry(void) {
     }
 
     if (ret_val) {
-	int disk_end;
+	long long disk_end;
 
 	disk_end = total_size-1;
 
@@ -2445,9 +2518,10 @@ change_id(int i) {
 
 static void
 draw_partition(int i) {
-    int size, j;
+    int j;
     int y = i + DISK_TABLE_START + 2 - (cur_part/NUM_ON_SCREEN)*NUM_ON_SCREEN;
     char *t;
+    long long size;
     double fsize;
 
     if (!arrow_cursor) {
@@ -2509,13 +2583,13 @@ draw_partition(int i) {
     size = p_info[i].last_sector - p_info[i].first_sector + 1;
     fsize = (double) size * SECTOR_SIZE;
     if (display_units == SECTORS)
-	mvprintw(y, SIZE_START, "%9d", size);
+	mvprintw(y, SIZE_START, "%11lld", size);
     else if (display_units == CYLINDERS)
-	mvprintw(y, SIZE_START, "%9d", size/cylinder_size);
+	mvprintw(y, SIZE_START, "%11lld", size/cylinder_size);
     else if (display_units == MEGABYTES)
-	mvprintw(y, SIZE_START, "%9.2f", ceiling((100*fsize)/(K*K))/100);
+	mvprintw(y, SIZE_START, "%11.2f", ceiling((100*fsize)/(K*K))/100);
     else if (display_units == GIGABYTES)
-	mvprintw(y, SIZE_START, "%9.2f", ceiling((100*fsize)/(K*K*K))/100);
+	mvprintw(y, SIZE_START, "%11.2f", ceiling((100*fsize)/(K*K*K))/100);
     if (size % cylinder_size != 0 ||
 	p_info[i].first_sector % cylinder_size != 0)
 	mvprintw(y, COLUMNS-1, "*");
@@ -2570,16 +2644,17 @@ draw_screen(void) {
     mvaddstr(HEADER_START+2, (COLS-strlen(line))/2, line);
     {
 	    long long bytes = actual_size*(long long) SECTOR_SIZE;
-	    long megabytes = bytes/1000000;
+	    long long megabytes = bytes/(K*K);
+
 	    if (megabytes < 10000)
-		    sprintf(line, _("Size: %lld bytes, %ld MB"),
+		    sprintf(line, _("Size: %lld bytes, %lld MB"),
 			    bytes, megabytes);
 	    else
-		    sprintf(line, _("Size: %lld bytes, %ld.%ld GB"),
-			    bytes, megabytes/1000, (megabytes/100)%10);
+		    sprintf(line, _("Size: %lld bytes, %lld.%lld GB"),
+			    bytes, megabytes/K, (10*megabytes/K)%10);
     }
     mvaddstr(HEADER_START+3, (COLS-strlen(line))/2, line);
-    sprintf(line, _("Heads: %d   Sectors per Track: %d   Cylinders: %d"),
+    sprintf(line, _("Heads: %d   Sectors per Track: %d   Cylinders: %lld"),
 	    heads, sectors, cylinders);
     mvaddstr(HEADER_START+4, (COLS-strlen(line))/2, line);
 
@@ -2589,13 +2664,13 @@ draw_screen(void) {
     mvaddstr(DISK_TABLE_START, FSTYPE_START, _("FS Type"));
     mvaddstr(DISK_TABLE_START, LABEL_START+1, _("[Label]"));
     if (display_units == SECTORS)
-	mvaddstr(DISK_TABLE_START, SIZE_START, _("  Sectors"));
+	mvaddstr(DISK_TABLE_START, SIZE_START, _("    Sectors"));
     else if (display_units == CYLINDERS)
-	mvaddstr(DISK_TABLE_START, SIZE_START, _("Cylinders"));
+	mvaddstr(DISK_TABLE_START, SIZE_START, _("  Cylinders"));
     else if (display_units == MEGABYTES)
-	mvaddstr(DISK_TABLE_START, SIZE_START, _("Size (MB)"));
+	mvaddstr(DISK_TABLE_START, SIZE_START, _("  Size (MB)"));
     else if (display_units == GIGABYTES)
-	mvaddstr(DISK_TABLE_START, SIZE_START, _("Size (GB)"));
+	mvaddstr(DISK_TABLE_START, SIZE_START, _("  Size (GB)"));
 
     move(DISK_TABLE_START+1, 1);
     for (i = 1; i < COLS-1; i++)
@@ -2648,8 +2723,7 @@ do_curses_fdisk(void) {
     int done = FALSE;
     char command;
 
-    static struct MenuItem menuMain[]=
-    {
+    static struct MenuItem menuMain[] = {
         { 'b', N_("Bootable"), N_("Toggle bootable flag of the current partition") },
         { 'd', N_("Delete"), N_("Delete the current partition") },
         { 'g', N_("Geometry"), N_("Change disk geometry (experts only)") },
@@ -2815,7 +2889,7 @@ do_curses_fdisk(void) {
 
 static void
 copyright(void) {
-    fprintf(stderr, _("Copyright (C) 1994-2000 Kevin E. Martin & aeb\n"));
+    fprintf(stderr, _("Copyright (C) 1994-2002 Kevin E. Martin & aeb\n"));
 }
 
 static void
@@ -2857,7 +2931,7 @@ main(int argc, char **argv)
 	    arrow_cursor = TRUE;
 	    break;
 	case 'c':
-	    user_cylinders = cylinders = atoi(optarg);
+	    user_cylinders = cylinders = atoll(optarg);
 	    if (cylinders <= 0) {
 		fprintf(stderr, "%s: %s\n", argv[0], _("Illegal cylinders value"));
 		exit(1);
@@ -2902,7 +2976,7 @@ main(int argc, char **argv)
 		    break;
 		default:
 		    usage(argv[0]);
-		    break;
+		    exit(1);
 		}
 	    }
 	    break;
