@@ -6,18 +6,8 @@
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation: either version 1 or
  * (at your option) any later version.
- *
- * For detailed old history, see older versions.
- * Contributions before 2001 by faith@cs.unc.edu, Michael Bischoff,
- * LeBlanc@mcc.ac.uk, martin@cs.unc.edu, leisner@sdsp.mc.xerox.com,
- * esr@snark.thyrsus.com, aeb@cwi.nl, quinlan@yggdrasil.com,
- * fasten@cs.bonn.edu, orschaer@cip.informatik.uni-erlangen.de,
- * jj@sunsite.mff.cuni.cz, fasten@shw.com, ANeuper@GUUG.de,
- * kgw@suse.de, kalium@gmx.de, dhuggins@linuxcare.com,
- * michal@ellpspace.math.ualberta.ca and probably others.
  */
 
-
 #include <unistd.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -71,7 +61,7 @@ static void delete_partition(int i);
 /* A valid partition table sector ends in 0x55 0xaa */
 static unsigned int
 part_table_flag(char *b) {
-	return ((uint) b[510]) + (((uint) b[511]) << 8);
+	return ((unsigned int) b[510]) + (((unsigned int) b[511]) << 8);
 }
 
 int
@@ -97,8 +87,9 @@ store4_little_endian(unsigned char *cp, unsigned int val) {
 
 static unsigned int
 read4_little_endian(unsigned char *cp) {
-	return (uint)(cp[0]) + ((uint)(cp[1]) << 8)
-		+ ((uint)(cp[2]) << 16) + ((uint)(cp[3]) << 24);
+	return (unsigned int)(cp[0]) + ((unsigned int)(cp[1]) << 8)
+		+ ((unsigned int)(cp[2]) << 16)
+		+ ((unsigned int)(cp[3]) << 24);
 }
 
 static void
@@ -141,9 +132,9 @@ char MBRbuffer[MAX_SECTOR_SIZE];
 struct pte {
 	struct partition *part_table;	/* points into sectorbuffer */
 	struct partition *ext_pointer;	/* points into sectorbuffer */
-	char changed;		/* boolean */
-	uint offset;		/* disk sector number */
-	char *sectorbuffer;	/* disk sector contents */
+	char changed;			/* boolean */
+	unsigned int offset;		/* disk sector number */
+	char *sectorbuffer;		/* disk sector contents */
 } ptes[MAXIMUM_PARTS];
 
 char	*disk_device,			/* must be specified */
@@ -158,11 +149,11 @@ int	fd,				/* the disk */
 	dos_changed = 0,
 	partitions = 4;			/* maximum partition + 1 */
 
-uint	user_cylinders, user_heads, user_sectors;
-uint	pt_heads, pt_sectors;
-uint	kern_heads, kern_sectors;
+unsigned int	user_cylinders, user_heads, user_sectors;
+unsigned int	pt_heads, pt_sectors;
+unsigned int	kern_heads, kern_sectors;
 
-uint	heads,
+unsigned int	heads,
 	sectors,
 	cylinders,
 	sector_size = DEFAULT_SECTOR_SIZE,
@@ -172,11 +163,14 @@ uint	heads,
 	display_in_cyl_units = 1,
 	extended_offset = 0;		/* offset of link pointers */
 
+unsigned long long total_number_of_sectors;
+
 #define dos_label (!sun_label && !sgi_label && !aix_label && !osf_label)
 int     sun_label = 0;                  /* looking at sun disklabel */
 int	sgi_label = 0;			/* looking at sgi disklabel */
 int	aix_label = 0;			/* looking at aix disklabel */
-int	osf_label = 0;			/* looking at osf disklabel */
+int	osf_label = 0;			/* looking at OSF/1 disklabel */
+int	possibly_osf_label = 0;
 
 jmp_buf listingbuf;
 
@@ -244,21 +238,21 @@ void fatal(enum failure why) {
 }
 
 static void
-seek_sector(int fd, uint secno) {
-	ext2_loff_t offset = (ext2_loff_t) secno * sector_size;
-	if (ext2_llseek(fd, offset, SEEK_SET) == (ext2_loff_t) -1)
+seek_sector(int fd, unsigned int secno) {
+	long long offset = (long long) secno * sector_size;
+	if (ext2_llseek(fd, offset, SEEK_SET) == (long long) -1)
 		fatal(unable_to_seek);
 }
 
 static void
-read_sector(int fd, uint secno, char *buf) {
+read_sector(int fd, unsigned int secno, char *buf) {
 	seek_sector(fd, secno);
 	if (read(fd, buf, sector_size) != sector_size)
 		fatal(unable_to_read);
 }
 
 static void
-write_sector(int fd, uint secno, char *buf) {
+write_sector(int fd, unsigned int secno, char *buf) {
 	seek_sector(fd, secno);
 	if (write(fd, buf, sector_size) != sector_size)
 		fatal(unable_to_write);
@@ -266,7 +260,7 @@ write_sector(int fd, uint secno, char *buf) {
 
 /* Allocate a buffer and read a partition table sector */
 static void
-read_pte(int fd, int pno, uint offset) {
+read_pte(int fd, int pno, unsigned int offset) {
 	struct pte *pe = &ptes[pno];
 
 	pe->offset = offset;
@@ -337,7 +331,7 @@ menu(void) {
 	   puts(_("   w   write table to disk and exit"));
 	   puts(_("   x   extra functionality (experts only)"));
 	}
-	else if(sgi_label) {
+	else if (sgi_label) {
 	   puts(_("Command action"));
 	   puts(_("   a   select bootable partition"));    /* sgi flavour */
 	   puts(_("   b   edit bootfile entry"));          /* sgi */
@@ -355,7 +349,7 @@ menu(void) {
 	   puts(_("   v   verify the partition table"));
 	   puts(_("   w   write table to disk and exit"));
 	}
-	else if(aix_label) {
+	else if (aix_label) {
 	   puts(_("Command action"));
 	   puts(_("   m   print this menu"));
 	   puts(_("   o   create a new empty DOS partition table"));
@@ -403,7 +397,7 @@ xmenu(void) {
 	   puts(_("   w   write table to disk and exit"));
 	   puts(_("   y   change number of physical cylinders"));	/*sun*/
 	}
-	else if(sgi_label) {
+	else if (sgi_label) {
 	   puts(_("Command action"));
 	   puts(_("   b   move beginning of data in a partition")); /* !sun */
 	   puts(_("   c   change number of cylinders"));
@@ -419,7 +413,7 @@ xmenu(void) {
 	   puts(_("   v   verify the partition table"));
 	   puts(_("   w   write table to disk and exit"));
 	}
-	else if(aix_label) {
+	else if (aix_label) {
 	   puts(_("Command action"));
 	   puts(_("   b   move beginning of data in a partition")); /* !sun */
 	   puts(_("   c   change number of cylinders"));
@@ -484,7 +478,7 @@ char *partition_type(unsigned char type)
 
 void list_types(struct systypes *sys)
 {
-	uint last[4], done = 0, next = 0, size;
+	unsigned int last[4], done = 0, next = 0, size;
 	int i;
 
 	for (i = 0; sys[i].name; i++);
@@ -530,9 +524,10 @@ clear_partition(struct partition *p) {
 }
 
 static void
-set_partition(int i, int doext, uint start, uint stop, int sysid) {
+set_partition(int i, int doext, unsigned int start, unsigned int stop,
+	      int sysid) {
 	struct partition *p;
-	uint offset;
+	unsigned int offset;
 
 	if (doext) {
 		p = ptes[i].ext_pointer;
@@ -710,7 +705,7 @@ create_doslabel(void) {
 	  "content won't be recoverable.\n\n"));
 	sun_nolabel();  /* otherwise always recognised as sun */
 	sgi_nolabel();  /* otherwise always recognised as sgi */
-	aix_label = osf_label = 0;
+	aix_label = osf_label = possibly_osf_label = 0;
 	partitions = 4;
 
 	for (i = 510-64; i < 510; i++)
@@ -806,7 +801,7 @@ get_partition_table_geometry(void) {
 void
 get_geometry(int fd, struct geom *g) {
 	int sec_fac;
-	long longsectors;
+	unsigned long long bytes;	/* really u64 */
 
 	get_sectorsize(fd);
 	sec_fac = sector_size / 512;
@@ -825,15 +820,23 @@ get_geometry(int fd, struct geom *g) {
 		pt_sectors ? pt_sectors :
 		kern_sectors ? kern_sectors : 63;
 
-	if (ioctl(fd, BLKGETSIZE, &longsectors))
-		longsectors = 0;
+	if (ioctl(fd, BLKGETSIZE64, &bytes) == 0) {
+		/* got bytes */
+	} else {
+		unsigned long longsectors;
+
+		if (ioctl(fd, BLKGETSIZE, &longsectors))
+			longsectors = 0;
+		bytes = ((unsigned long long) longsectors) << 9;
+	}
+
+	total_number_of_sectors = (bytes >> 9);
 
 	sector_offset = 1;
 	if (dos_compatible_flag)
 		sector_offset = sectors;
 
-	cylinders = longsectors / (heads * sectors);
-	cylinders /= sec_fac;
+	cylinders = total_number_of_sectors / (heads * sectors * sec_fac);
 	if (!cylinders)
 		cylinders = user_cylinders;
 
@@ -856,10 +859,23 @@ get_boot(enum action what) {
 
 	partitions = 4;
 
+	for (i = 0; i < 4; i++) {
+		struct pte *pe = &ptes[i];
+
+		pe->part_table = pt_offset(MBRbuffer, i);
+		pe->ext_pointer = NULL;
+		pe->offset = 0;
+		pe->sectorbuffer = MBRbuffer;
+		pe->changed = (what == create_empty_dos);
+	}
+
+	if (what == create_empty_sun && check_sun_label())
+		return 0;
+
+	memset(MBRbuffer, 0, 512);
+
 	if (what == create_empty_dos)
 		goto got_dos_table;		/* skip reading disk */
-	if (what == create_empty_sun)
-		goto got_table;
 
 	if ((fd = open(disk_device, type_open)) < 0) {
 	    if ((fd = open(disk_device, O_RDONLY)) < 0) {
@@ -867,7 +883,8 @@ get_boot(enum action what) {
 		    return 1;
 		fatal(unable_to_open);
 	    } else
-		printf(_("You will not be able to write the partition table.\n"));
+		printf(_("You will not be able to write "
+			 "the partition table.\n"));
 	}
 
 	if (512 != read(fd, MBRbuffer, 512)) {
@@ -880,8 +897,6 @@ get_boot(enum action what) {
 
 	update_units();
 
-got_table:
-
 	if (check_sun_label())
 		return 0;
 
@@ -891,8 +906,15 @@ got_table:
 	if (check_aix_label())
 		return 0;
 
-	if (check_osf_label())
-		return 0;
+	if (check_osf_label()) {
+		possibly_osf_label = 1;
+		if (!valid_part_table_flag(MBRbuffer)) {
+			osf_label = 1;
+			return 0;
+		}
+		printf(_("This disk has both DOS and BSD magic.\n"
+			 "Give the 'b' command to go to BSD mode.\n"));
+	}
 
 got_dos_table:
 
@@ -900,8 +922,9 @@ got_dos_table:
 		switch(what) {
 		case fdisk:
 			fprintf(stderr,
-				_("Device contains neither a valid DOS partition"
-			 	  " table, nor Sun, SGI or OSF disklabel\n"));
+				_("Device contains neither a valid DOS "
+				  "partition table, nor Sun, SGI or OSF "
+				  "disklabel\n"));
 #ifdef __sparc__
 			create_sunlabel();
 #else
@@ -915,25 +938,15 @@ got_dos_table:
 		case create_empty_dos:
 		case create_empty_sun:
 			break;
+		default:
+			fprintf(stderr, _("Internal error\n"));
+			exit(1);
 		}
-
-		fprintf(stderr, _("Internal error\n"));
-		exit(1);
 	}
 
 	warn_cylinders();
 	warn_geometry();
 
-	for (i = 0; i < 4; i++) {
-		struct pte *pe = &ptes[i];
-
-		pe->part_table = pt_offset(MBRbuffer, i);
-		pe->ext_pointer = NULL;
-		pe->offset = 0;
-		pe->sectorbuffer = MBRbuffer;
-		pe->changed = (what == create_empty_dos);
-	}
-
 	for (i = 0; i < 4; i++) {
 		struct pte *pe = &ptes[i];
 
@@ -1033,10 +1046,11 @@ read_hex(struct systypes *sys)
  *
  * There is no default if DFLT is not between LOW and HIGH.
  */
-uint
-read_int(uint low, uint dflt, uint high, uint base, char *mesg)
+unsigned int
+read_int(unsigned int low, unsigned int dflt, unsigned int high,
+	 unsigned int base, char *mesg)
 {
-	uint i;
+	unsigned int i;
 	int default_ok = 1;
 	static char *ms = NULL;
 	static int mslen = 0;
@@ -1051,10 +1065,10 @@ read_int(uint low, uint dflt, uint high, uint base, char *mesg)
 		default_ok = 0;
 
 	if (default_ok)
-		snprintf(ms, mslen, _("%s (%d-%d, default %d): "),
+		snprintf(ms, mslen, _("%s (%u-%u, default %u): "),
 			 mesg, low, high, dflt);
 	else
-		snprintf(ms, mslen, "%s (%d-%d): ",
+		snprintf(ms, mslen, "%s (%u-%u): ",
 			 mesg, low, high);
 
 	while (1) {
@@ -1066,38 +1080,49 @@ read_int(uint low, uint dflt, uint high, uint base, char *mesg)
 			continue;
 
 		if (*line_ptr == '+' || *line_ptr == '-') {
+			int minus = (*line_ptr == '-');
+			int absolute = 0;
+
 			i = atoi(line_ptr+1);
-			if (*line_ptr == '-')
-				i = -i;
+
  			while (isdigit(*++line_ptr))
 				use_default = 0;
+
 			switch (*line_ptr) {
 				case 'c':
 				case 'C':
 					if (!display_in_cyl_units)
 						i *= heads * sectors;
 					break;
-				case 'k':
 				case 'K':
-					i *= 2;
-					i /= (sector_size / 512);
-					i /= units_per_sector;
+					absolute = 1024;
+					break;
+				case 'k':
+					absolute = 1000;
 					break;
 				case 'm':
 				case 'M':
-					i *= 2048;
-					i /= (sector_size / 512);
-					i /= units_per_sector;
+					absolute = 1000000;
 					break;
 				case 'g':
 				case 'G':
-					i *= 2048000;
-					i /= (sector_size / 512);
-					i /= units_per_sector;
+					absolute = 1000000000;
 					break;
 				default:
 					break;
 			}
+			if (absolute) {
+				unsigned long long bytes;
+				unsigned long unit;
+
+				bytes = (unsigned long long) i * absolute;
+				unit = sector_size * units_per_sector;
+				bytes += unit/2;	/* round */
+				bytes /= unit;
+				i = bytes;
+			}
+			if (minus)
+				i = -i;
 			i += base;
 		} else {
 			i = atoi(line_ptr);
@@ -1107,7 +1132,7 @@ read_int(uint low, uint dflt, uint high, uint base, char *mesg)
 			}
 		}
 		if (use_default)
-			printf(_("Using default value %d\n"), i = dflt);
+			printf(_("Using default value %u\n"), i = dflt);
 		if (i >= low && i <= high)
 			break;
 		else
@@ -1124,16 +1149,72 @@ get_partition(int warn, int max) {
 	i = read_int(1, 0, max, 0, _("Partition number")) - 1;
 	pe = &ptes[i];
 
-	if (warn
-	    && ((!sun_label && !sgi_label && !pe->part_table->sys_ind)
-	     || (sun_label &&
-		(!sunlabel->partitions[i].num_sectors ||
-		 !sunlabel->infos[i].id))
-		|| (sgi_label && (!sgi_get_num_sectors(i))))
-	) fprintf(stderr, _("Warning: partition %d has empty type\n"), i+1);
+	if (warn) {
+		if ((!sun_label && !sgi_label && !pe->part_table->sys_ind)
+		    || (sun_label &&
+			(!sunlabel->partitions[i].num_sectors ||
+			 !sunlabel->infos[i].id))
+		    || (sgi_label && (!sgi_get_num_sectors(i)))
+		   )
+			fprintf(stderr,
+				_("Warning: partition %d has empty type\n"),
+				i+1);
+	}
 	return i;
 }
 
+static int
+get_existing_partition(int warn, int max) {
+	int pno = -1;
+	int i;
+
+	for (i = 0; i < max; i++) {
+		struct pte *pe = &ptes[i];
+		struct partition *p = pe->part_table;
+
+		if (p && !is_cleared_partition(p)) {
+			if (pno >= 0)
+				goto not_unique;
+			pno = i;
+		}
+	}
+	if (pno >= 0) {
+		printf(_("Selected partition %d\n"), pno+1);
+		return pno;
+	}
+	printf(_("No partition is defined yet!\n"));
+	return -1;
+
+ not_unique:
+	return get_partition(warn, max);
+}
+
+static int
+get_nonexisting_partition(int warn, int max) {
+	int pno = -1;
+	int i;
+
+	for (i = 0; i < max; i++) {
+		struct pte *pe = &ptes[i];
+		struct partition *p = pe->part_table;
+
+		if (p && is_cleared_partition(p)) {
+			if (pno >= 0)
+				goto not_unique;
+			pno = i;
+		}
+	}
+	if (pno >= 0) {
+		printf(_("Selected partition %d\n"), pno+1);
+		return pno;
+	}
+	printf(_("All primary partitions have been defined already!\n"));
+	return -1;
+
+ not_unique:
+	return get_partition(warn, max);
+}
+
 char * const
 str_units(int n) {	/* n==1: use singular */
 	if (n == 1)
@@ -1229,7 +1310,7 @@ delete_partition(int i) {
 			/* the first logical in a longer chain */
 			struct pte *pe = &ptes[5];
 
-			if(pe->part_table) /* prevent SEGFAULT */
+			if (pe->part_table) /* prevent SEGFAULT */
 				set_start_sect(pe->part_table,
 					       get_partition_start(pe) -
 					       extended_offset);
@@ -1252,12 +1333,26 @@ delete_partition(int i) {
 static void
 change_sysid(void) {
 	char *temp;
-	int i = get_partition(0, partitions), sys, origsys;
-	struct partition *p = ptes[i].part_table;
+	int i, sys, origsys;
+	struct partition *p;
 
+	/* If sgi_label then don't use get_existing_partition,
+	   let the user select a partition, since get_existing_partition()
+	   only works for Linux like partition tables. */
+	if (!sgi_label) { 
+		i = get_existing_partition(0, partitions);
+	} else {
+		i = get_partition(0, partitions);
+	}
+
+	if (i == -1)
+		return;
+	p = ptes[i].part_table;
 	origsys = sys = get_sysid(i);
 
-	if (!sys && !sgi_label && !sun_label)
+	/* if changing types T to 0 is allowed, then
+	   the reverse change must be allowed, too */
+	if (!sys && !sgi_label && !sun_label && !get_nr_sects(p))
                 printf(_("Partition %d does not exist yet!\n"), i + 1);
         else while (1) {
 		sys = read_hex (get_sys_types());
@@ -1292,7 +1387,7 @@ change_sysid(void) {
 				       "partition 11 as entire volume (6)"
 				       "as IRIX expects it.\n\n"));
                         if (sys == origsys)
-                            break;
+				break;
 			if (sun_label) {
 				sun_change_sysid(i, sys);
 			} else
@@ -1318,8 +1413,9 @@ change_sysid(void) {
  * Jan.  1990 (version 1.2.1 by Gordon W. Ross Aug. 1990; Modified by S.
  * Lubkin Oct.  1991). */
 
-static void long2chs(ulong ls, uint *c, uint *h, uint *s) {
-	int	spc = heads * sectors;
+static void
+long2chs(ulong ls, unsigned int *c, unsigned int *h, unsigned int *s) {
+	int spc = heads * sectors;
 
 	*c = ls / spc;
 	ls = ls % spc;
@@ -1328,10 +1424,10 @@ static void long2chs(ulong ls, uint *c, uint *h, uint *s) {
 }
 
 static void check_consistency(struct partition *p, int partition) {
-	uint	pbc, pbh, pbs;		/* physical beginning c, h, s */
-	uint	pec, peh, pes;		/* physical ending c, h, s */
-	uint	lbc, lbh, lbs;		/* logical beginning c, h, s */
-	uint	lec, leh, les;		/* logical ending c, h, s */
+	unsigned int pbc, pbh, pbs;	/* physical beginning c, h, s */
+	unsigned int pec, peh, pes;	/* physical ending c, h, s */
+	unsigned int lbc, lbh, lbs;	/* logical beginning c, h, s */
+	unsigned int lec, leh, les;	/* logical ending c, h, s */
 
 	if (!heads || !sectors || (partition >= 4))
 		return;		/* do not check extended partitions */
@@ -1380,19 +1476,36 @@ static void check_consistency(struct partition *p, int partition) {
 
 /* Ending on cylinder boundary? */
 	if (peh != (heads - 1) || pes != sectors) {
-		printf(_("Partition %i does not end on cylinder boundary:\n"),
+		printf(_("Partition %i does not end on cylinder boundary.\n"),
 			partition + 1);
+#if 0
 		printf(_("     phys=(%d, %d, %d) "), pec, peh, pes);
 		printf(_("should be (%d, %d, %d)\n"),
 		pec, heads - 1, sectors);
+#endif
 	}
 }
 
 static void
 list_disk_geometry(void) {
-	printf(_("\nDisk %s: %d heads, %d sectors, %d cylinders\nUnits = "
-	       "%s of %d * %d bytes\n\n"), disk_device, heads, sectors,
-	       cylinders, str_units(PLURAL), units_per_sector, sector_size);
+	long long bytes = (total_number_of_sectors << 9);
+	long megabytes = bytes/1000000;
+
+	if (megabytes < 10000)
+		printf(_("\nDisk %s: %ld MB, %lld bytes\n"),
+		       disk_device, megabytes, bytes);
+	else
+		printf(_("\nDisk %s: %ld.%ld GB, %lld bytes\n"),
+		       disk_device, megabytes/1000, (megabytes/100)%10, bytes);
+	printf(_("%d heads, %d sectors/track, %d cylinders"),
+	       heads, sectors, cylinders);
+	if (units_per_sector == 1)
+		printf(_(", total %llu sectors"),
+		       total_number_of_sectors / (sector_size/512));
+	printf("\n");
+	printf(_("Units = %s of %d * %d = %d bytes\n\n"),
+	       str_units(PLURAL),
+	       units_per_sector, sector_size, units_per_sector * sector_size);
 }
 
 /*
@@ -1404,7 +1517,7 @@ static int
 wrong_p_order(int *prev) {
 	struct pte *pe;
 	struct partition *p;
-	uint last_p_start_pos = 0, p_start_pos;
+	unsigned int last_p_start_pos = 0, p_start_pos;
 	int i, last_i = 0;
 
 	for (i = 0 ; i < partitions; i++) {
@@ -1450,7 +1563,7 @@ fix_chain_of_logicals(void) {
 	/* Stage 1: sort sectors but leave sector of part 4 */
 	/* (Its sector is the global extended_offset.) */
  stage1:
-	for(j = 5; j < partitions-1; j++) {
+	for (j = 5; j < partitions-1; j++) {
 		oj = ptes[j].offset;
 		ojj = ptes[j+1].offset;
 		if (oj > ojj) {
@@ -1470,7 +1583,7 @@ fix_chain_of_logicals(void) {
 
 	/* Stage 2: sort starting sectors */
  stage2:
-	for(j = 4; j < partitions-1; j++) {
+	for (j = 4; j < partitions-1; j++) {
 		pj = ptes[j].part_table;
 		pjj = ptes[j+1].part_table;
 		sj = get_start_sect(pj);
@@ -1488,7 +1601,7 @@ fix_chain_of_logicals(void) {
 	}
 
 	/* Probably something was changed */
-	for(j = 4; j < partitions; j++)
+	for (j = 4; j < partitions; j++)
 		ptes[j].changed = 1;
 }
 
@@ -1497,7 +1610,7 @@ fix_partition_table_order(void) {
 	struct pte *pei, *pek;
 	int i,k;
 
-	if(!wrong_p_order(NULL)) {
+	if (!wrong_p_order(NULL)) {
 		printf(_("Nothing to do. Ordering is correct already.\n\n"));
 		return;
 	}
@@ -1562,10 +1675,10 @@ list_table(int xtra) {
 	if (w < 5)
 		w = 5;
 
-	printf(_("%*s Boot    Start       End    Blocks   Id  System\n"),
+	printf(_("%*s Boot      Start         End      Blocks   Id  System\n"),
 	       w+1, _("Device"));
 
-	for (i = 0 ; i < partitions; i++) {
+	for (i = 0; i < partitions; i++) {
 		struct pte *pe = &ptes[i];
 
 		p = pe->part_table;
@@ -1581,14 +1694,14 @@ list_table(int xtra) {
 			if (sector_size > 1024)
 				pblocks *= (sector_size / 1024);
                         printf(
-			    "%s  %c %9ld %9ld %9ld%c  %2x  %s\n",
+			    "%s  %c %11lu %11lu %11lu%c  %2x  %s\n",
 			partname(disk_device, i+1, w+2),
 /* boot flag */		!p->boot_ind ? ' ' : p->boot_ind == ACTIVE_FLAG
 			? '*' : '?',
-/* start */		(long) cround(get_partition_start(pe)),
-/* end */		(long) cround(get_partition_start(pe) + psects
+/* start */		(unsigned long) cround(get_partition_start(pe)),
+/* end */		(unsigned long) cround(get_partition_start(pe) + psects
 				- (psects ? 1 : 0)),
-/* odd flag on end */	(long) pblocks, podd ? '+' : ' ',
+/* odd flag on end */	(unsigned long) pblocks, podd ? '+' : ' ',
 /* type id */		p->sys_ind,
 /* type name */		(type = partition_type(p->sys_ind)) ?
 			type : _("Unknown"));
@@ -1612,12 +1725,12 @@ x_list_table(int extend) {
 
 	printf(_("\nDisk %s: %d heads, %d sectors, %d cylinders\n\n"),
 		disk_device, heads, sectors, cylinders);
-        printf(_("Nr AF  Hd Sec  Cyl  Hd Sec  Cyl    Start     Size ID\n"));
+        printf(_("Nr AF  Hd Sec  Cyl  Hd Sec  Cyl     Start      Size ID\n"));
 	for (i = 0 ; i < partitions; i++) {
 		pe = &ptes[i];
 		p = (extend ? pe->ext_pointer : pe->part_table);
 		if (p != NULL) {
-                        printf("%2d %02x%4d%4d%5d%4d%4d%5d%9d%9d %02x\n",
+                        printf("%2d %02x%4d%4d%5d%4d%4d%5d%11u%11u %02x\n",
 				i + 1, p->boot_ind, p->head,
 				sector(p->sector),
 				cylinder(p->sector, p->cyl), p->end_head,
@@ -1631,7 +1744,7 @@ x_list_table(int extend) {
 }
 
 static void
-fill_bounds(uint *first, uint *last) {
+fill_bounds(unsigned int *first, unsigned int *last) {
 	int i;
 	struct pte *pe = &ptes[0];
 	struct partition *p;
@@ -1649,8 +1762,9 @@ fill_bounds(uint *first, uint *last) {
 }
 
 static void
-check(int n, uint h, uint s, uint c, uint start) {
-	uint total, real_s, real_c;
+check(int n, unsigned int h, unsigned int s, unsigned int c,
+      unsigned int start) {
+	unsigned int total, real_s, real_c;
 
 	real_s = sector(s) - 1;
 	real_c = cylinder(s, c);
@@ -1676,8 +1790,8 @@ check(int n, uint h, uint s, uint c, uint start) {
 static void
 verify(void) {
 	int i, j;
-	uint total = 1;
-	uint first[partitions], last[partitions];
+	unsigned int total = 1;
+	unsigned int first[partitions], last[partitions];
 	struct partition *p;
 
 	if (warn_geometry())
@@ -1721,7 +1835,7 @@ verify(void) {
 
 	if (extended_offset) {
 		struct pte *pex = &ptes[ext_index];
-		uint e_last = get_start_sect(pex->part_table) +
+		unsigned int e_last = get_start_sect(pex->part_table) +
 			get_nr_sects(pex->part_table) - 1;
 
 		for (i = 4; i < partitions; i++) {
@@ -1752,7 +1866,8 @@ add_partition(int n, int sys) {
 	int i, read = 0;
 	struct partition *p = ptes[n].part_table;
 	struct partition *q = ptes[ext_index].part_table;
-	uint start, stop = 0, limit, temp,
+	long long llimit;
+	unsigned int start, stop = 0, limit, temp,
 		first[partitions], last[partitions];
 
 	if (p && p->sys_ind) {
@@ -1763,7 +1878,13 @@ add_partition(int n, int sys) {
 	fill_bounds(first, last);
 	if (n < 4) {
 		start = sector_offset;
-		limit = heads * sectors * cylinders - 1;
+		if (display_in_cyl_units || !total_number_of_sectors)
+			llimit = heads * sectors * cylinders - 1;
+		else
+			llimit = total_number_of_sectors - 1;
+		limit = llimit;
+		if (limit != llimit)
+			limit = 0x7fffffff;
 		if (extended_offset) {
 			first[ext_index] = extended_offset;
 			last[ext_index] = get_start_sect(q) +
@@ -1797,8 +1918,8 @@ add_partition(int n, int sys) {
 			read = 0;
 		}
 		if (!read && start == temp) {
-			uint i;
-			i = start;
+			unsigned int i = start;
+
 			start = read_int(cround(i), cround(i), cround(limit),
 					 0, mesg);
 			if (display_in_cyl_units) {
@@ -1910,13 +2031,14 @@ new_partition(void) {
 		return;
 	}
 
-	if (partitions >= MAXIMUM_PARTS) {
+	for (i = 0; i < 4; i++)
+		free_primary += !ptes[i].part_table->sys_ind;
+
+	if (!free_primary && partitions >= MAXIMUM_PARTS) {
 		printf(_("The maximum number of partitions has been created\n"));
 		return;
 	}
 
-	for (i = 0; i < 4; i++)
-		free_primary += !ptes[i].part_table->sys_ind;
 	if (!free_primary) {
 		if (extended_offset)
 			add_logical();
@@ -1931,8 +2053,9 @@ new_partition(void) {
 			 _("l   logical (5 or over)") : _("e   extended"));
 		while (1) {
 			if ((c = tolower(read_char(line))) == 'p') {
-				add_partition(get_partition(0, 4),
-					      LINUX_NATIVE);
+				int i = get_nonexisting_partition(0, 4);
+				if (i >= 0)
+					add_partition(i, LINUX_NATIVE);
 				return;
 			}
 			else if (c == 'l' && extended_offset) {
@@ -1940,8 +2063,9 @@ new_partition(void) {
 				return;
 			}
 			else if (c == 'e' && !extended_offset) {
-				add_partition(get_partition(0, 4),
-					      EXTENDED);
+				int i = get_nonexisting_partition(0, 4);
+				if (i >= 0)
+					add_partition(i, EXTENDED);
 				return;
 			}
 			else
@@ -1957,7 +2081,7 @@ write_table(void) {
 
 	if (dos_label) {
 		for (i=0; i<3; i++)
-			if(ptes[i].changed)
+			if (ptes[i].changed)
 				ptes[3].changed = 1;
 		for (i = 3; i < partitions; i++) {
 			struct pte *pe = &ptes[i];
@@ -1975,7 +2099,7 @@ write_table(void) {
 		int needw = 0;
 
 		for (i=0; i<8; i++)
-			if(ptes[i].changed)
+			if (ptes[i].changed)
 				needw = 1;
 		if (needw)
 			sun_write_table();
@@ -2001,7 +2125,7 @@ reread_partition_table(int leave) {
 		   twice, the second time works. - biro@yggdrasil.com */
                 sync();
                 sleep(2);
-                if((i = ioctl(fd, BLKRRPART)) != 0)
+                if ((i = ioctl(fd, BLKRRPART)) != 0)
                         error = errno;
         }
 
@@ -2065,7 +2189,7 @@ static void
 move_begin(int i) {
 	struct pte *pe = &ptes[i];
 	struct partition *p = pe->part_table;
-	uint new, first;
+	unsigned int new, first;
 
 	if (warn_geometry())
 		return;
@@ -2123,7 +2247,7 @@ xselect(void) {
 				x_list_table(1);
 			break;
 		case 'f':
-			if(dos_label)
+			if (dos_label)
 				fix_partition_table_order();
 			break;
 		case 'g':
@@ -2247,7 +2371,7 @@ try(char *device, int user_specified) {
 		/* Ignore other errors, since we try IDE
 		   and SCSI hard disks which may not be
 		   installed on the system. */
-		if(errno == EACCES) {
+		if (errno == EACCES) {
 			fprintf(stderr, _("Cannot open %s\n"), device);
 			return;
 		}
@@ -2272,7 +2396,7 @@ tryprocpt(void) {
 		if (sscanf (line, " %d %d %d %[^\n ]",
 			    &ma, &mi, &sz, ptname) != 4)
 			continue;
-		for(s = ptname; *s; s++);
+		for (s = ptname; *s; s++);
 		if (isdigit(s[-1]))
 			continue;
 		snprintf(devname, sizeof(devname), "/dev/%s", ptname);
@@ -2304,12 +2428,17 @@ main(int argc, char **argv) {
 	 *  fdisk -l [-b sectorsize] [-u] device ...
 	 *  fdisk -s [partition] ...
 	 *  fdisk [-b sectorsize] [-u] device
+	 *
+	 * Options -C, -H, -S set the geometry.
+	 * 
 	 */
-	while ((c = getopt(argc, argv, "b:lsuvV")) != -1) {
+	while ((c = getopt(argc, argv, "b:C:H:lsS:uvV")) != -1) {
 		switch (c) {
 		case 'b':
-			/* ugly: this sector size is really per device,
-			   so cannot be combined with multiple disks */
+			/* Ugly: this sector size is really per device,
+			   so cannot be combined with multiple disks,
+			   and te same goes for the C/H/S options.
+			*/
 			sector_size = atoi(optarg);
 			if (sector_size != 512 && sector_size != 1024 &&
 			    sector_size != 2048)
@@ -2317,6 +2446,19 @@ main(int argc, char **argv) {
 			sector_offset = 2;
 			user_set_sector_size = 1;
 			break;
+		case 'C':
+			user_cylinders = atoi(optarg);
+			break;
+		case 'H':
+			user_heads = atoi(optarg);
+			if (user_heads <= 0 || user_heads >= 256)
+				user_heads = 0;
+			break;
+		case 'S':
+			user_sectors = atoi(optarg);
+			if (user_sectors <= 0 || user_sectors >= 64)
+				user_sectors = 0;
+			break;
 		case 'l':
 			optl = 1;
 			break;
@@ -2336,7 +2478,8 @@ main(int argc, char **argv) {
 	}
 
 #if 0
-	printf(_("This kernel finds the sector size itself - -b option ignored\n"));
+	printf(_("This kernel finds the sector size itself - "
+		 "-b option ignored\n"));
 #else
 	if (user_set_sector_size && argc-optind != 1)
 		printf(_("Warning: the -b (set sector size) option should"
@@ -2352,7 +2495,7 @@ main(int argc, char **argv) {
 			   variable `k' might be clobbered by `longjmp' */
 			dummy(&k);
 			listing = 1;
-			for(k=optind; k<argc; k++)
+			for (k=optind; k<argc; k++)
 				try(argv[k], 1);
 		} else {
 			/* we no longer have default device names */
@@ -2396,18 +2539,16 @@ main(int argc, char **argv) {
 
 	get_boot(fdisk);
 
-#ifdef __alpha__
-	/* On alpha, if we detect a disklabel, go directly to
-	   disklabel mode (typically you'll be switching from DOS
-	   partition tables to disklabels, not the other way around)
-	   - dhuggins@linuxcare.com */
 	if (osf_label) {
-		printf(_("Detected an OSF/1 disklabel on %s, entering disklabel mode.\n"
-			 "To return to DOS partition table mode, use the 'r' command.\n"),
+		/* OSF label, and no DOS label */
+		printf(_("Detected an OSF/1 disklabel on %s, entering "
+			 "disklabel mode.\n"),
 		       disk_device);
 		bselect();
+		osf_label = 0;
+		/* If we return we may want to make an empty DOS label? */
 	}
-#endif
+
 	while (1) {
 		putchar('\n');
 		c = tolower(read_char(_("Command (m for help): ")));
@@ -2449,8 +2590,17 @@ main(int argc, char **argv) {
 				unknown_command(c);
 			break;
 		case 'd':
-			delete_partition(
-				get_partition(1, partitions));
+        		/* If sgi_label then don't use get_existing_partition,
+			   let the user select a partition, since
+			   get_existing_partition() only works for Linux-like
+			   partition tables */
+        		if (!sgi_label) {
+                		j = get_existing_partition(1, partitions);
+        		} else {
+                		j = get_partition(1, partitions);
+        		}
+			if (j >= 0)
+				delete_partition(j);
 			break;
 		case 'i':
 			if (sgi_label)
@@ -2492,7 +2642,7 @@ main(int argc, char **argv) {
 			write_table(); 		/* does not return */
 			break;
 		case 'x':
-			if(sgi_label) {
+			if (sgi_label) {
 				fprintf(stderr,
 					_("\n\tSorry, no experts menu for SGI "
 					"partition tables available.\n\n"));
