@@ -33,12 +33,12 @@
 
 #include <linux/types.h>	/* for __u32, __u16, __u8, __s16 */
 
-#ifndef BSD_DISKMAGIC	/* perhaps from <linux/genhd.h> */
+#ifndef BSD_DISKMAGIC
 #define BSD_DISKMAGIC     ((__u32) 0x82564557)
 #endif
 
 #ifndef BSD_MAXPARTITIONS
-#define	BSD_MAXPARTITIONS 8
+#define	BSD_MAXPARTITIONS 16
 #endif
 
 #define BSD_LINUX_BOOTDIR "/usr/ucb/mdec"
@@ -49,6 +49,9 @@
 #elif defined (__alpha__) || defined (__powerpc__) || defined (__ia64__) || defined (__hppa__)
 #define BSD_LABELSECTOR   0
 #define BSD_LABELOFFSET   64
+#elif defined (__s390__) || defined (__s390x__)
+#define BSD_LABELSECTOR   1
+#define BSD_LABELOFFSET   0
 #else
 #error unknown architecture
 #endif
