@@ -14,20 +14,20 @@ struct systypes i386_sys_types[] = {
 	{0x08, N_("AIX")},		/* AIX boot (AIX -- PS/2 port) or SplitDrive */
 	{0x09, N_("AIX bootable")},	/* AIX data or Coherent */
 	{0x0a, N_("OS/2 Boot Manager")},/* OS/2 Boot Manager */
-	{0x0b, N_("Win95 FAT32")},
-	{0x0c, N_("Win95 FAT32 (LBA)")},/* LBA really is `Extended Int 13h' */
-	{0x0e, N_("Win95 FAT16 (LBA)")},
-	{0x0f, N_("Win95 Ext'd (LBA)")},
+	{0x0b, N_("W95 FAT32")},
+	{0x0c, N_("W95 FAT32 (LBA)")},/* LBA really is `Extended Int 13h' */
+	{0x0e, N_("W95 FAT16 (LBA)")},
+	{0x0f, N_("W95 Ext'd (LBA)")},
 	{0x10, N_("OPUS")},
 	{0x11, N_("Hidden FAT12")},
 	{0x12, N_("Compaq diagnostics")},
 	{0x14, N_("Hidden FAT16 <32M")},
 	{0x16, N_("Hidden FAT16")},
 	{0x17, N_("Hidden HPFS/NTFS")},
-	{0x18, N_("AST Windows swapfile")},
-	{0x1b, N_("Hidden Win95 FAT32")},
-	{0x1c, N_("Hidden Win95 FAT32 (LBA)")},
-	{0x1e, N_("Hidden Win95 FAT16 (LBA)")},
+	{0x18, N_("AST SmartSleep")},
+	{0x1b, N_("Hidden W95 FAT32")},
+	{0x1c, N_("Hidden W95 FAT32 (LBA)")},
+	{0x1e, N_("Hidden W95 FAT16 (LBA)")},
 	{0x24, N_("NEC DOS")},
 	{0x39, N_("Plan 9")},
 	{0x3c, N_("PartitionMagic recovery")},
@@ -64,11 +64,16 @@ struct systypes i386_sys_types[] = {
 	{0x94, N_("Amoeba BBT")},	/* (bad block table) */
 	{0x9f, N_("BSD/OS")},		/* BSDI */
 	{0xa0, N_("IBM Thinkpad hibernation")},
-	{0xa5, N_("BSD/386")},
+	{0xa5, N_("FreeBSD")},		/* various BSD flavours */
 	{0xa6, N_("OpenBSD")},
 	{0xa7, N_("NeXTSTEP")},
+	{0xa8, N_("Darwin UFS")},
+	{0xa9, N_("NetBSD")},
+	{0xab, N_("Darwin boot")},
 	{0xb7, N_("BSDI fs")},
 	{0xb8, N_("BSDI swap")},
+	{0xbb, N_("Boot Wizard hidden")},
+	{0xbe, N_("Solaris boot")},
 	{0xc1, N_("DRDOS/sec (FAT-12)")},
 	{0xc4, N_("DRDOS/sec (FAT-16 < 32M)")},
 	{0xc6, N_("DRDOS/sec (FAT-16)")},
@@ -77,6 +82,7 @@ struct systypes i386_sys_types[] = {
 	{0xdb, N_("CP/M / CTOS / ...")},/* CP/M or Concurrent CP/M or
 					   Concurrent DOS or CTOS */
 	{0xde, N_("Dell Utility")},	/* Dell PowerEdge Server utilities */
+	{0xdf, N_("BootIt")},		/* BootIt EMBRM */
 	{0xe1, N_("DOS access")},	/* DOS access or SpeedStor 12-bit FAT
 					   extended partition */
 	{0xe3, N_("DOS R/O")},		/* DOS R/O or SpeedStor */
@@ -85,6 +91,7 @@ struct systypes i386_sys_types[] = {
 	{0xeb, N_("BeOS fs")},
 	{0xee, N_("EFI GPT")},		/* Intel EFI GUID Partition Table */
 	{0xef, N_("EFI (FAT-12/16/32)")},/* Intel EFI System Partition */
+	{0xf0, N_("Linux/PA-RISC boot")},/* Linux/PA-RISC boot loader */
 	{0xf1, N_("SpeedStor")},
 	{0xf4, N_("SpeedStor")},	/* SpeedStor large partition */
 	{0xf2, N_("DOS secondary")},	/* DOS 3.3+ secondary */
