@@ -130,10 +130,10 @@ fatal(char *s, ...) {
  *
  * Note: we use 512-byte sectors here, irrespective of the hardware ss.
  */
-#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__)
+#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__) && !defined (__s390x__)
 static
-_syscall5(int,  _llseek,  uint,  fd, ulong, hi, ulong, lo,
-       loff_t *, res, uint, wh);
+_syscall5(int,  _llseek,  unsigned int,  fd, ulong, hi, ulong, lo,
+       loff_t *, res, unsigned int, wh);
 #endif
 
 static int
@@ -142,7 +142,7 @@ sseek(char *dev, unsigned int fd, unsigned long s) {
     in = ((loff_t) s << 9);
     out = 1;
 
-#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__)
+#if !defined (__alpha__) && !defined (__ia64__) && !defined (__x86_64__) && !defined (__s390x__)
     if (_llseek (fd, in>>32, in & 0xffffffff, &out, SEEK_SET) != 0) {
 #else
     if ((out = lseek(fd, in, SEEK_SET)) != in) {
@@ -154,8 +154,8 @@ sseek(char *dev, unsigned int fd, unsigned long s) {
 
     if (in != out) {
 	error(_("seek error: wanted 0x%08x%08x, got 0x%08x%08x\n"),
-	       (uint)(in>>32), (uint)(in & 0xffffffff),
-	       (uint)(out>>32), (uint)(out & 0xffffffff));
+	       (unsigned int)(in>>32), (unsigned int)(in & 0xffffffff),
+	       (unsigned int)(out>>32), (unsigned int)(out & 0xffffffff));
 	return 0;
     }
     return 1;
@@ -396,7 +396,7 @@ struct geometry {
 static struct geometry
 get_geometry(char *dev, int fd, int silent) {
     struct hd_geometry g;
-    long size;
+    unsigned long size;
     struct geometry R;
 
     if (ioctl(fd, BLKGETSIZE, &size)) {
@@ -882,9 +882,9 @@ out_partition_header(char *dev, int format, struct geometry G) {
 	printf(_("   Device Boot   Start       End    #blocks   Id  System\n"));
 	break;
       case F_MEGABYTE:
-	printf(_("Units = megabytes of 1048576 bytes, blocks of 1024 bytes"
+	printf(_("Units = mebibytes of 1048576 bytes, blocks of 1024 bytes"
 	       ", counting from %d\n\n"), increment);
-	printf(_("   Device Boot Start   End     MB    #blocks   Id  System\n"));
+	printf(_("   Device Boot Start   End    MiB    #blocks   Id  System\n"));
 	break;
     }
 }
@@ -1269,7 +1269,7 @@ partitions_ok(struct disk_desc *z) {
 	b = p->p.begin_chs;
 	aa = chs_to_longchs(a);
 	bb = chs_to_longchs(b);
-	if (!chs_ok(b, PNO(p), "start"))
+	if (!chs_ok(b, PNO(p), _("start")))
 	  return 0;
 	if(a.s && !is_equal_chs(a, b))
 	  warn(_("partition %s: start: (c,h,s) expected (%ld,%ld,%ld) found (%ld,%ld,%ld)\n"),
@@ -1278,7 +1278,7 @@ partitions_ok(struct disk_desc *z) {
 	b = p->p.end_chs;
 	aa = chs_to_longchs(a);
 	bb = chs_to_longchs(b);
-	if (!chs_ok(b, PNO(p), "end"))
+	if (!chs_ok(b, PNO(p), _("end")))
 	  return 0;
 	if(a.s && !is_equal_chs(a, b))
 	  warn(_("partition %s: end: (c,h,s) expected (%ld,%ld,%ld) found (%ld,%ld,%ld)\n"),
@@ -1367,7 +1367,7 @@ extended_partition(char *dev, int fd, struct part_desc *ep, struct disk_desc *z)
 }
 
 #define BSD_DISKMAGIC   (0x82564557UL)
-#define BSD_MAXPARTITIONS       8
+#define BSD_MAXPARTITIONS       16
 #define BSD_FS_UNUSED	   0
 typedef unsigned char u8;
 typedef unsigned short u16;
@@ -1404,11 +1404,11 @@ bsd_partition(char *dev, int fd, struct part_desc *ep, struct disk_desc *z) {
 	if (!(s = get_sector(dev,fd,start+1)))
 		return;
 	l = (struct bsd_disklabel *) (s->data);
-	if (l->d_magic != BSD_DISKMAGIC)
+	if (l->d_magic != BSD_DISKMAGIC || l->d_magic2 != BSD_DISKMAGIC)
 		return;
 
 	bp = bp0 = &l->d_partitions[0];
-	while (bp - bp0 <= BSD_MAXPARTITIONS) {
+	while (bp - bp0 < BSD_MAXPARTITIONS && bp - bp0 < l->d_npartitions) {
 		if (pno+1 >= SIZE(z->partitions)) {
 			printf(_("too many partitions - ignoring those "
 			       "past nr (%d)\n"), pno-1);
@@ -1568,9 +1568,11 @@ write_partitions(char *dev, int fd, struct disk_desc *z) {
 	s = get_sector(dev, fd, p->sector);
 	if (!s) return 0;
 	s->to_be_written = 1;
-	copy_from_part(&(p->p), s->data + p->offset);
-	s->data[510] = 0x55;
-	s->data[511] = 0xaa;
+	if (p->ptype == DOS_TYPE) {
+	    copy_from_part(&(p->p), s->data + p->offset);
+	    s->data[510] = 0x55;
+	    s->data[511] = 0xaa;
+	}
     }
     if (save_sector_file) {
 	if (!save_sectors(dev, fd)) {
@@ -2396,7 +2398,7 @@ static void do_change_id(char *dev, char *part, char *id);
 static void do_unhide(char **av, int ac, char *arg);
 static void do_activate(char **av, int ac, char *arg);
 
-int total_size;
+unsigned long total_size;
 
 int
 main(int argc, char **argv) {
@@ -2525,7 +2527,7 @@ main(int argc, char **argv) {
 	total_size = 0;
 	openproc();
 	while ((dev = nextproc()) != NULL) {
-	    if (!strncmp(dev, "hd", 2) && is_ide_cdrom_or_tape(dev))
+	    if (is_ide_cdrom_or_tape(dev))
 		continue;
 	    if (opt_out_geom)
 		do_geom(dev, 1);
@@ -2536,7 +2538,7 @@ main(int argc, char **argv) {
 	}
 
 	if (opt_size)
-	  printf(_("total: %d blocks\n"), total_size);
+	  printf(_("total: %lu blocks\n"), total_size);
 
 	exit(exit_status);
     }
@@ -2662,7 +2664,7 @@ do_geom (char *dev, int silent) {
 static void
 do_size (char *dev, int silent) {
     int fd;
-    long size;
+    unsigned long size;
 
     fd = my_open(dev, 0, silent);
     if (fd < 0)
@@ -2683,9 +2685,9 @@ do_size (char *dev, int silent) {
       return;
 
     if (silent)
-      printf("%s: %9ld\n", dev, size);
+      printf("%s: %9lu\n", dev, size);
     else
-      printf("%ld\n", size);
+      printf("%lu\n", size);
 
     total_size += size;
 }
@@ -2719,7 +2721,8 @@ set_active (struct disk_desc *z, char *pnam) {
     int pno;
 
     pno = asc_to_index(pnam, z);
-    z->partitions[pno].p.bootable = 0x80;
+    if (z->partitions[pno].ptype == DOS_TYPE)
+	    z->partitions[pno].p.bootable = 0x80;
 }
 
 static void
@@ -2755,7 +2758,8 @@ do_activate (char **av, int ac, char *arg) {
     } else {
 	/* clear `active byte' everywhere */
 	for (pno=0; pno < z->partno; pno++)
-	    z->partitions[pno].p.bootable = 0;
+	    if (z->partitions[pno].ptype == DOS_TYPE)
+		z->partitions[pno].p.bootable = 0;
 
 	/* then set where desired */
 	if (ac == 1)
