@@ -47,6 +47,12 @@
 #include <security/pam_misc.h>
 #endif
 
+#ifdef WITH_SELINUX
+#include <selinux/selinux.h>
+#include <selinux/av_permissions.h>
+#include "selinux_utils.h"
+#endif
+
 typedef unsigned char boolean;
 #define false 0
 #define true 1
@@ -121,6 +127,27 @@ main (int argc, char *argv[]) {
        exit(1);
     }
 
+#ifdef WITH_SELINUX
+    if (is_selinux_enabled()) {
+      if(uid == 0) {
+	if (checkAccess(pw->pw_name,PASSWD__CHSH)!=0) {
+	  security_context_t user_context;
+	  if (getprevcon(&user_context) < 0)
+	    user_context=(security_context_t) strdup(_("Unknown user context"));
+	  fprintf(stderr, _("%s: %s is not authorized to change the shell of %s\n"),
+		  whoami, user_context, pw->pw_name);
+	  freecon(user_context);
+	  exit(1);
+	}
+      }
+      if (setupDefaultContext("/etc/passwd") != 0) {
+	fprintf(stderr,_("%s: Can't set default context for /etc/passwd"),
+		whoami);
+	exit(1);
+      }
+    }
+#endif
+
     oldshell = pw->pw_shell;
     if (!oldshell[0]) oldshell = "/bin/sh";
 
@@ -201,11 +228,8 @@ main (int argc, char *argv[]) {
  *	parse the command line arguments, and fill in "pinfo" with any
  *	information from the command line.
  */
-static void parse_argv (argc, argv, pinfo)
-    int argc;
-    char *argv[];
-    struct sinfo *pinfo;
-{
+static void
+parse_argv (int argc, char *argv[], struct sinfo *pinfo) {
     int index, c;
 
     static struct option long_options[] = {
@@ -220,7 +244,7 @@ static void parse_argv (argc, argv, pinfo)
     while (c != EOF) {
 	c = getopt_long (argc, argv, "s:luv", long_options, &index);
 	switch (c) {
-	case EOF:
+	case -1:
 	    break;
 	case 'v':
 	    printf ("%s\n", util_linux_version);
@@ -257,22 +281,20 @@ static void parse_argv (argc, argv, pinfo)
  *  usage () --
  *	print out a usage message.
  */
-static void usage (fp)
-    FILE *fp;
-{
-    fprintf (fp, _("Usage: %s [ -s shell ] "), whoami);
-    fprintf (fp, _("[ --list-shells ] [ --help ] [ --version ]\n"));
-    fprintf (fp, _("       [ username ]\n"));
+static void
+usage (FILE *fp) {
+    fprintf (fp,
+	     _("Usage: %s [ -s shell ] [ --list-shells ] "
+	       "[ --help ] [ --version ]\n"
+	       "       [ username ]\n"), whoami);
 }
 
 /*
  *  prompt () --
  *	ask the user for a given field and return it.
  */
-static char *prompt (question, def_val)
-    char *question;
-    char *def_val;
-{
+static char *
+prompt (char *question, char *def_val) {
     int len;
     char *ans, *cp;
   
@@ -300,9 +322,8 @@ static char *prompt (question, def_val)
  *	an error and return (-1).
  *	if the shell is a bad idea, print a warning.
  */
-static int check_shell (shell)
-    char *shell;
-{
+static int
+check_shell (char *shell) {
     int i, c;
 
     if (*shell != '/') {
@@ -354,9 +375,8 @@ static int check_shell (shell)
  *	return true.  if not, return false.
  *	if the given shell is NULL, /etc/shells is outputted to stdout.
  */
-static boolean get_shell_list (shell_name)
-    char *shell_name;
-{
+static boolean
+get_shell_list (char *shell_name) {
     FILE *fp;
     boolean found;
     int len;
@@ -391,9 +411,8 @@ static boolean get_shell_list (shell_name)
 /*
  *  xmalloc () -- malloc that never fails.
  */
-static void *xmalloc (bytes)
-    int bytes;
-{
+static void *
+xmalloc (int bytes) {
     void *vp;
 
     vp = malloc (bytes);
