@@ -183,10 +183,6 @@ static void sigint (int);
 static void motd (void);
 static void dolastlog (int quiet);
 
-#ifndef __linux__
-static char *stypeof (char *ttyid);
-#endif
-
 #ifdef CRYPTOCARD
 #include "cryptocard.h"
 #endif
@@ -205,7 +201,6 @@ int	kerror = KSUCCESS, notickets = 1;
 #endif
 
 #define	TTYGRPNAME	"tty"		/* name of group to own ttys */
-/**#  define TTYGRPNAME      "other" **/
 
 #ifndef MAXPATHLEN
 #  define MAXPATHLEN 1024
@@ -242,11 +237,6 @@ struct	ltchars ltc = {
 };
 #endif
 
-const char *months[] = {
-	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
-	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
-};
-
 /* Nice and simple code provided by Linus Torvalds 16-Feb-93 */
 /* Nonblocking stuff by Maciej W. Rozycki, macro@ds2.pg.gda.pl, 1999.
    He writes: "Login performs open() on a tty in a blocking mode.
@@ -280,6 +270,24 @@ opentty(const char * tty) {
 		close(fd);
 }
 
+/* In case login is suid it was possible to use a hardlink as stdin
+   and exploit races for a local root exploit. (Wojciech Purczynski). */
+/* More precisely, the problem is  ttyn := ttyname(0); ...; chown(ttyn);
+   here ttyname() might return "/tmp/x", a hardlink to a pseudotty. */
+/* All of this is a problem only when login is suid, which it isnt. */
+static void
+check_ttyname(char *ttyn) {
+	struct stat statbuf;
+
+	if (lstat(ttyn, &statbuf)
+	    || !S_ISCHR(statbuf.st_mode)
+	    || (statbuf.st_nlink > 1 && strncmp(ttyn, "/dev/", 5))) {
+		syslog(LOG_ERR, _("FATAL: bad tty"));
+		sleep(1);
+		exit(1);
+	}
+}
+
 /* true if the filedescriptor fd is a console tty, very Linux specific */
 static int
 consoletty(int fd) {
@@ -350,8 +358,8 @@ main(int argc, char **argv)
     char *domain, *ttyn;
     char tbuf[MAXPATHLEN + 2], tname[sizeof(_PATH_TTY) + 10];
     char *termenv;
-    char * childArgv[10];
-    char * buff;
+    char *childArgv[10];
+    char *buff;
     int childArgc = 0;
 #ifdef USE_PAM
     int retcode;
@@ -364,9 +372,6 @@ main(int argc, char **argv)
 #ifdef CHOWNVCS
     char vcsn[20], vcsan[20];
 #endif
-#ifndef __linux__
-    int ioctlval;
-#endif
 
     pid = getpid();
 
@@ -451,38 +456,21 @@ main(int argc, char **argv)
 	while(*p)
 	    *p++ = ' ';
     } else
-      ask = 1;
+        ask = 1;
 
-#ifndef __linux__
-    ioctlval = 0;
-    ioctl(0, TIOCLSET, &ioctlval);
-    ioctl(0, TIOCNXCL, 0);
-    fcntl(0, F_SETFL, ioctlval);
-    ioctl(0, TIOCGETP, &sgttyb);
-    sgttyb.sg_erase = CERASE;
-    sgttyb.sg_kill = CKILL;
-    ioctl(0, TIOCSLTC, &ltc);
-    ioctl(0, TIOCSETC, &tc);
-    ioctl(0, TIOCSETP, &sgttyb);
-    
-    /*
-     * Be sure that we're in blocking mode!!!
-     * This is really for HPUX
-     */
-    ioctlval = 0;
-    ioctl(0, FIOSNBIO, &ioctlval);
-#endif /* ! __linux__ */
-    
     for (cnt = getdtablesize(); cnt > 2; cnt--)
       close(cnt);
     
     ttyn = ttyname(0);
+
     if (ttyn == NULL || *ttyn == '\0') {
 	/* no snprintf required - see definition of tname */
 	sprintf(tname, "%s??", _PATH_TTY);
 	ttyn = tname;
     }
 
+    check_ttyname(ttyn);
+
     if (strncmp(ttyn, "/dev/", 5) == 0)
 	tty_name = ttyn+5;
     else
@@ -704,10 +692,6 @@ main(int argc, char **argv)
 #else /* ! USE_PAM */
 
     for (cnt = 0;; ask = 1) {
-#  ifndef __linux__
-	ioctlval = 0;
-	ioctl(0, TIOCSETD, &ioctlval);
-#  endif
 
 	if (ask) {
 	    fflag = 0;
@@ -1268,29 +1252,40 @@ getloginname(void) {
 }
 #endif
 
-void
+/*
+ * Robert Ambrose writes:
+ * A couple of my users have a problem with login processes hanging around
+ * soaking up pts's.  What they seem to hung up on is trying to write out the
+ * message 'Login timed out after %d seconds' when the connection has already
+ * been dropped.
+ * What I did was add a second timeout while trying to write the message so
+ * the process just exits if the second timeout expires.
+ */
+
+static void
+timedout2(int sig) {
+	struct termio ti;
+    
+	/* reset echo */
+	ioctl(0, TCGETA, &ti);
+	ti.c_lflag |= ECHO;
+	ioctl(0, TCSETA, &ti);
+	exit(0);			/* %% */
+}
+
+static void
 timedout(int sig) {
-    struct termio ti;
-    
-    fprintf(stderr, _("Login timed out after %d seconds\n"), timeout);
-    
-    /* reset echo */
-    ioctl(0, TCGETA, &ti);
-    ti.c_lflag |= ECHO;
-    ioctl(0, TCSETA, &ti);
-    exit(0);			/* %% */
+	signal(SIGALRM, timedout2);
+	alarm(10);
+	fprintf(stderr, _("Login timed out after %d seconds\n"), timeout);
+	signal(SIGALRM, SIG_IGN);
+	alarm(0);
+	timedout2(0);
 }
 
 #ifndef USE_PAM
 int
 rootterm(char * ttyn)
-#ifndef __linux__
-{
-    struct ttyent *t;
-    
-    return((t = getttynam(ttyn)) && (t->ty_status&TTY_SECURE));
-}
-#else
 { 
     int fd;
     char buf[100],*p;
@@ -1318,7 +1313,6 @@ rootterm(char * ttyn)
   	}
     }
 }
-#endif /* !__linux__ */
 #endif /* !USE_PAM */
 
 jmp_buf motdinterrupt;
@@ -1411,18 +1405,6 @@ badlogin(const char *name) {
     }
 }
 
-#undef	UNKNOWN
-#define	UNKNOWN	"su"
-
-#ifndef __linux__
-char *
-stypeof(char *ttyid) {
-    struct ttyent *t;
-
-    return(ttyid && (t = getttynam(ttyid)) ? t->ty_type : UNKNOWN);
-}
-#endif 
-
 /* Should not be called from PAM code... */
 void
 sleepexit(int eval) {
