@@ -26,14 +26,12 @@
    59 Bcy, 3161:       PRAISE_BOB and KILL_BOB options split, other minor
                        changes.
 
-   1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+   1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
    - added Native Language Support
 
    2000-03-17 Burt Holzman <bnh@iname.com>
    - added range checks for dates
 */
-#include "../defines.h"		/* for util-linux-version */
-
 
 /* configuration options  VVVVV   READ THIS!!! */
 
@@ -64,6 +62,7 @@
 #include <time.h>
 #include <stdio.h>
 #include "nls.h"
+#include "../defines.h"		/* for util-linux-version */
 
 #ifndef __GNUC__
 #define inline /* foo */
