@@ -67,6 +67,7 @@
 void finish(int);
 void done(void);
 void fail(void);
+void resize(int);
 void fixtty(void);
 void getmaster(void);
 void getslave(void);
@@ -90,18 +91,20 @@ int	l;
 char	line[] = "/dev/ptyXX";
 #endif
 int	aflg = 0;
+char	*cflg = NULL;
 int	fflg = 0;
 int	qflg = 0;
+int	tflg = 0;
 
 static char *progname;
 
 static void
-die_if_symlink(char *fn) {
+die_if_link(char *fn) {
 	struct stat s;
 
-	if (lstat(fn, &s) == 0 && S_ISLNK(s.st_mode)) {
+	if (lstat(fn, &s) == 0 && (S_ISLNK(s.st_mode) || s.st_nlink > 1)) {
 		fprintf(stderr,
-			_("Warning: `%s' is a symlink.\n"
+			_("Warning: `%s' is a link.\n"
 			  "Use `%s [options] %s' if you really "
 			  "want to use it.\n"
 			  "Script not started.\n"),
@@ -133,21 +136,27 @@ main(int argc, char **argv) {
 		}
 	}
 
-	while ((ch = getopt(argc, argv, "afq")) != EOF)
+	while ((ch = getopt(argc, argv, "ac:fqt")) != -1)
 		switch((char)ch) {
 		case 'a':
 			aflg++;
 			break;
+		case 'c':
+			cflg = optarg;
+			break;
 		case 'f':
 			fflg++;
 			break;
 		case 'q':
 			qflg++;
 			break;
+		case 't':
+			tflg++;
+			break;
 		case '?':
 		default:
 			fprintf(stderr,
-				_("usage: script [-a] [-f] [-q] [file]\n"));
+				_("usage: script [-a] [-f] [-q] [-t] [file]\n"));
 			exit(1);
 		}
 	argc -= optind;
@@ -157,7 +166,7 @@ main(int argc, char **argv) {
 		fname = argv[0];
 	else {
 		fname = "typescript";
-		die_if_symlink(fname);
+		die_if_link(fname);
 	}
 	if ((fscript = fopen(fname, aflg ? "a" : "w")) == NULL) {
 		perror(fname);
@@ -189,7 +198,8 @@ main(int argc, char **argv) {
 			dooutput();
 		else
 			doshell();
-	}
+	} else
+		(void) signal(SIGWINCH, resize);
 	doinput();
 
 	return 0;
@@ -201,9 +211,7 @@ doinput() {
 	char ibuf[BUFSIZ];
 
 	(void) fclose(fscript);
-#ifdef HAVE_openpty
-	(void) close(slave);
-#endif
+
 	while ((cc = read(0, ibuf, BUFSIZ)) > 0)
 		(void) write(master, ibuf, cc);
 	done();
@@ -225,22 +233,51 @@ finish(int dummy) {
 		done();
 }
 
+void
+resize(int dummy) {
+	/* transmit window change information to the child */
+	(void) ioctl(0, TIOCGWINSZ, (char *)&win);
+	(void) ioctl(slave, TIOCSWINSZ, (char *)&win);
+
+	kill(child, SIGWINCH);
+}
+
+/*
+ * Stop extremely silly gcc complaint on %c:
+ *  warning: `%c' yields only last 2 digits of year in some locales
+ */
+static void
+my_strftime(char *buf, size_t len, const char *fmt, const struct tm *tm) {
+	strftime(buf, len, fmt, tm);
+}
+
 void
 dooutput() {
 	register int cc;
 	time_t tvec;
 	char obuf[BUFSIZ];
+	struct timeval tv;
+	double oldtime=time(NULL), newtime;
 
 	(void) close(0);
 #ifdef HAVE_openpty
 	(void) close(slave);
 #endif
 	tvec = time((time_t *)NULL);
-	fprintf(fscript, _("Script started on %s"), ctime(&tvec));
+	my_strftime(obuf, sizeof obuf, "%c\n", localtime(&tvec));
+	fprintf(fscript, _("Script started on %s"), obuf);
+
 	for (;;) {
+		if (tflg)
+			gettimeofday(&tv, NULL);
 		cc = read(master, obuf, sizeof (obuf));
 		if (cc <= 0)
 			break;
+		if (tflg) {
+			newtime = tv.tv_sec + (double) tv.tv_usec / 1000000;
+			fprintf(stderr, "%f %i\n", newtime - oldtime, cc);
+			oldtime = newtime;
+		}
 		(void) write(1, obuf, cc);
 		(void) fwrite(obuf, 1, cc, fscript);
 		if (fflg)
@@ -251,7 +288,9 @@ dooutput() {
 
 void
 doshell() {
-	/***
+	char *shname;
+
+#if 0
 	int t;
 
 	t = open(_PATH_TTY, O_RDWR);
@@ -259,7 +298,8 @@ doshell() {
 		(void) ioctl(t, TIOCNOTTY, (char *)0);
 		(void) close(t);
 	}
-	***/
+#endif
+
 	getslave();
 	(void) close(master);
 	(void) fclose(fscript);
@@ -267,11 +307,18 @@ doshell() {
 	(void) dup2(slave, 1);
 	(void) dup2(slave, 2);
 	(void) close(slave);
-#ifdef __linux__
-	execl(shell, strrchr(shell, '/') + 1, "-i", 0);
-#else
-	execl(shell, "sh", "-i", 0);
-#endif
+
+	shname = strrchr(shell, '/');
+	if (shname)
+		shname++;
+	else
+		shname = shell;
+
+	if (cflg)
+		execl(shell, shname, "-c", cflg, 0);
+	else
+		execl(shell, shname, "-i", 0);
+
 	perror(shell);
 	fail();
 }
@@ -299,9 +346,10 @@ done() {
 
 	if (subchild) {
 		if (!qflg) {
+			char buf[BUFSIZ];
 			tvec = time((time_t *)NULL);
-			fprintf(fscript, _("\nScript done on %s"),
-				ctime(&tvec));
+			my_strftime(buf, sizeof buf, "%c\n", localtime(&tvec));
+			fprintf(fscript, _("\nScript done on %s"), buf);
 		}
 		(void) fclose(fscript);
 		(void) close(master);
