@@ -11,7 +11,10 @@
 #include <sys/stat.h>
 #include "mntent.h"
 #include "fstab.h"
-#include "sundries.h"		/* for xmalloc() etc */
+#include "sundries.h"
+#include "xmalloc.h"
+#include "mount_blkid.h"
+#include "paths.h"
 #include "nls.h"
 
 #define streq(s, t)	(strcmp ((s), (t)) == 0)
@@ -26,49 +29,49 @@ static int var_mtab_is_a_symlink = 0;
 
 static void
 get_mtab_info(void) {
-     struct stat mtab_stat;
+	struct stat mtab_stat;
 
-     if (!have_mtab_info) {
-	  if (lstat(MOUNTED, &mtab_stat))
-	       var_mtab_does_not_exist = 1;
-	  else if (S_ISLNK(mtab_stat.st_mode))
-	       var_mtab_is_a_symlink = 1;
-	  have_mtab_info = 1;
-     }
+	if (!have_mtab_info) {
+		if (lstat(MOUNTED, &mtab_stat))
+			var_mtab_does_not_exist = 1;
+		else if (S_ISLNK(mtab_stat.st_mode))
+			var_mtab_is_a_symlink = 1;
+		have_mtab_info = 1;
+	}
 }
 
 int
 mtab_does_not_exist(void) {
-     get_mtab_info();
-     return var_mtab_does_not_exist;
+	get_mtab_info();
+	return var_mtab_does_not_exist;
 }
 
 int
 mtab_is_a_symlink(void) {
-     get_mtab_info();
-     return var_mtab_is_a_symlink;
+	get_mtab_info();
+	return var_mtab_is_a_symlink;
 }
 
 int
 mtab_is_writable() {
-     static int ret = -1;
+	static int ret = -1;
 
-     /* Should we write to /etc/mtab upon an update?
-	Probably not if it is a symlink to /proc/mounts, since that
-	would create a file /proc/mounts in case the proc filesystem
-	is not mounted. */
-     if (mtab_is_a_symlink())
-	  return 0;
+	/* Should we write to /etc/mtab upon an update?
+	   Probably not if it is a symlink to /proc/mounts, since that
+	   would create a file /proc/mounts in case the proc filesystem
+	   is not mounted. */
+	if (mtab_is_a_symlink())
+		return 0;
 
-     if (ret == -1) {
-	  int fd = open(MOUNTED, O_RDWR | O_CREAT, 0644);
-	  if (fd >= 0) {
-	       close(fd);
-	       ret = 1;
-	  } else
-	       ret = 0;
-     }
-     return ret;
+	if (ret == -1) {
+		int fd = open(MOUNTED, O_RDWR | O_CREAT, 0644);
+		if (fd >= 0) {
+			close(fd);
+			ret = 1;
+		} else
+			ret = 0;
+	}
+	return ret;
 }
 
 /* Contents of mtab and fstab ---------------------------------*/
@@ -81,25 +84,45 @@ static void read_mounttable(void), read_fstab(void);
 
 struct mntentchn *
 mtab_head() {
-     if (!got_mtab)
-	  read_mounttable();
-     return &mounttable;
+	if (!got_mtab)
+		read_mounttable();
+	return &mounttable;
 }
 
 struct mntentchn *
 fstab_head() {
-     if (!got_fstab)
-	  read_fstab();
-     return &fstab;
+	if (!got_fstab)
+		read_fstab();
+	return &fstab;
+}
+
+static void
+my_free(const void *s) {
+	if (s)
+		free((void *) s);
+}
+
+static void
+discard_mntentchn(struct mntentchn *mc0) {
+	struct mntentchn *mc, *mc1;
+
+	for (mc = mc0->nxt; mc && mc != mc0; mc = mc1) {
+		mc1 = mc->nxt;
+		my_free(mc->m.mnt_fsname);
+		my_free(mc->m.mnt_dir);
+		my_free(mc->m.mnt_type);
+		my_free(mc->m.mnt_opts);
+		free(mc);
+	}
 }
 
 static void
 read_mntentchn(mntFILE *mfp, const char *fnam, struct mntentchn *mc0) {
 	struct mntentchn *mc = mc0;
-	struct mntent *mnt;
+	struct my_mntent *mnt;
 
-	while ((mnt = my_getmntent (mfp)) != NULL) {
-		if (!streq (mnt->mnt_type, MNTTYPE_IGNORE)) {
+	while ((mnt = my_getmntent(mfp)) != NULL) {
+		if (!streq(mnt->mnt_type, MNTTYPE_IGNORE)) {
 			mc->nxt = (struct mntentchn *) xmalloc(sizeof(*mc));
 			mc->nxt->prev = mc;
 			mc = mc->nxt;
@@ -108,7 +131,7 @@ read_mntentchn(mntFILE *mfp, const char *fnam, struct mntentchn *mc0) {
 		}
 	}
 	mc0->prev = mc;
-	if (ferror (mfp->mntent_fp)) {
+	if (ferror(mfp->mntent_fp)) {
 		int errsv = errno;
 		error(_("warning: error reading %s: %s"),
 		      fnam, strerror (errsv));
@@ -124,48 +147,50 @@ read_mntentchn(mntFILE *mfp, const char *fnam, struct mntentchn *mc0) {
  */
 static void
 read_mounttable() {
-     mntFILE *mfp;
-     const char *fnam;
-     struct mntentchn *mc = &mounttable;
+	mntFILE *mfp;
+	const char *fnam;
+	struct mntentchn *mc = &mounttable;
 
-     got_mtab = 1;
-     mc->nxt = mc->prev = NULL;
+	got_mtab = 1;
+	mc->nxt = mc->prev = NULL;
 
-     fnam = MOUNTED;
-     mfp = my_setmntent (fnam, "r");
-     if (mfp == NULL || mfp->mntent_fp == NULL) {
-	  int errsv = errno;
-	  fnam = PROC_MOUNTS;
-	  mfp = my_setmntent (fnam, "r");
-	  if (mfp == NULL || mfp->mntent_fp == NULL) {
-	       error(_("warning: can't open %s: %s"), MOUNTED, strerror (errsv));
-	       return;
-	  }
-	  if (verbose)
-	       printf (_("mount: could not open %s - using %s instead\n"),
-		       MOUNTED, PROC_MOUNTS);
-     }
-     read_mntentchn(mfp, fnam, mc);
+	fnam = MOUNTED;
+	mfp = my_setmntent (fnam, "r");
+	if (mfp == NULL || mfp->mntent_fp == NULL) {
+		int errsv = errno;
+		fnam = PROC_MOUNTS;
+		mfp = my_setmntent (fnam, "r");
+		if (mfp == NULL || mfp->mntent_fp == NULL) {
+			error(_("warning: can't open %s: %s"),
+			      MOUNTED, strerror (errsv));
+			return;
+		}
+		if (verbose)
+			printf (_("mount: could not open %s - "
+				  "using %s instead\n"),
+				MOUNTED, PROC_MOUNTS);
+	}
+	read_mntentchn(mfp, fnam, mc);
 }
 
 static void
 read_fstab() {
-     mntFILE *mfp = NULL;
-     const char *fnam;
-     struct mntentchn *mc = &fstab;
+	mntFILE *mfp = NULL;
+	const char *fnam;
+	struct mntentchn *mc = &fstab;
 
-     got_fstab = 1;
-     mc->nxt = mc->prev = NULL;
+	got_fstab = 1;
+	mc->nxt = mc->prev = NULL;
 
-     fnam = _PATH_FSTAB;
-     mfp = my_setmntent (fnam, "r");
-     if (mfp == NULL || mfp->mntent_fp == NULL) {
-     	  int errsv = errno;
-	  error(_("warning: can't open %s: %s"),
-		_PATH_FSTAB, strerror (errsv));
-	  return;
-     }
-     read_mntentchn(mfp, fnam, mc);
+	fnam = _PATH_FSTAB;
+	mfp = my_setmntent (fnam, "r");
+	if (mfp == NULL || mfp->mntent_fp == NULL) {
+		int errsv = errno;
+		error(_("warning: can't open %s: %s"),
+		      _PATH_FSTAB, strerror (errsv));
+		return;
+	}
+	read_mntentchn(mfp, fnam, mc);
 }
      
 
@@ -176,35 +201,67 @@ getmntfile (const char *name) {
 
 	mc0 = mtab_head();
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
-		if (streq (mc->m.mnt_dir, name) ||
-		    streq (mc->m.mnt_fsname, name))
+		if (streq(mc->m.mnt_dir, name) ||
+		    streq(mc->m.mnt_fsname, name))
 			return mc;
 	return NULL;
 }
 
 /*
- * Given the name NAME, and the place MCPREV we found it last time,
+ * Given the directory name NAME, and the place MCPREV we found it last time,
  * try to find more occurrences.
  */ 
 struct mntentchn *
-getmntfilesbackward (const char *name, struct mntentchn *mcprev) {
+getmntdirbackward (const char *name, struct mntentchn *mcprev) {
 	struct mntentchn *mc, *mc0;
 
 	mc0 = mtab_head();
 	if (!mcprev)
 		mcprev = mc0;
 	for (mc = mcprev->prev; mc && mc != mc0; mc = mc->prev)
-		if (streq (mc->m.mnt_dir, name) ||
-		    streq (mc->m.mnt_fsname, name))
+		if (streq(mc->m.mnt_dir, name))
 			return mc;
 	return NULL;
 }
 
+/*
+ * Given the device name NAME, and the place MCPREV we found it last time,
+ * try to find more occurrences.
+ */ 
+struct mntentchn *
+getmntdevbackward (const char *name, struct mntentchn *mcprev) {
+	struct mntentchn *mc, *mc0;
+
+	mc0 = mtab_head();
+	if (!mcprev)
+		mcprev = mc0;
+	for (mc = mcprev->prev; mc && mc != mc0; mc = mc->prev)
+		if (streq(mc->m.mnt_fsname, name))
+			return mc;
+	return NULL;
+}
+
+/*
+ * Given the name NAME, check that it occurs precisely once as dir or dev.
+ */
+int
+is_mounted_once(const char *name) {
+	struct mntentchn *mc, *mc0;
+	int ct = 0;
+
+	mc0 = mtab_head();
+	for (mc = mc0->prev; mc && mc != mc0; mc = mc->prev)
+		if (streq(mc->m.mnt_dir, name) ||
+		    streq(mc->m.mnt_fsname, name))
+			ct++;
+	return (ct == 1);
+}
+
 /* Given the name FILE, try to find the option "loop=FILE" in mtab.  */ 
 struct mntentchn *
 getmntoptfile (const char *file) {
 	struct mntentchn *mc, *mc0;
-	char *opts, *s;
+	const char *opts, *s;
 	int l;
 
 	if (!file)
@@ -223,22 +280,64 @@ getmntoptfile (const char *file) {
 	return NULL;
 }
 
+static int
+has_label(const char *device, const char *label) {
+	const char *devlabel;
+	int ret;
+
+	devlabel = mount_get_volume_label_by_spec(device);
+	ret = !strcmp(label, devlabel);
+	/* free(devlabel); */
+	return ret;
+}
+
+static int
+has_uuid(const char *device, const char *uuid){
+	const char *devuuid;
+	int ret;
+
+	devuuid = mount_get_devname_by_uuid(device);
+	ret = !strcmp(uuid, devuuid);
+	/* free(devuuid); */
+	return ret;
+}
+
 /* Find the entry (SPEC,FILE) in fstab */
 struct mntentchn *
 getfsspecfile (const char *spec, const char *file) {
 	struct mntentchn *mc, *mc0;
 
 	mc0 = fstab_head();
+
+	/* first attempt: names occur precisely as given */
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
-		if (streq (mc->m.mnt_dir, file) &&
-		    streq (mc->m.mnt_fsname, spec))
+		if (streq(mc->m.mnt_dir, file) &&
+		    streq(mc->m.mnt_fsname, spec))
 			return mc;
+
+	/* second attempt: names found after symlink resolution */
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
-		if ((streq (mc->m.mnt_dir, file) ||
-		     streq (canonicalize(mc->m.mnt_dir), file))
-		    && (streq (mc->m.mnt_fsname, spec) ||
-			streq (canonicalize(mc->m.mnt_fsname), spec)))
+		if ((streq(mc->m.mnt_dir, file) ||
+		     streq(canonicalize(mc->m.mnt_dir), file))
+		    && (streq(mc->m.mnt_fsname, spec) ||
+			streq(canonicalize(mc->m.mnt_fsname), spec)))
 			return mc;
+
+	/* third attempt: names found after LABEL= or UUID= resolution */
+	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt) {
+		if (!strncmp (mc->m.mnt_fsname, "LABEL=", 6) &&
+		    (streq(mc->m.mnt_dir, file) ||
+		     streq(canonicalize(mc->m.mnt_dir), file))) {
+			if (has_label(spec, mc->m.mnt_fsname+6))
+				return mc;
+		}
+		if (!strncmp (mc->m.mnt_fsname, "UUID=", 5) &&
+		    (streq(mc->m.mnt_dir, file) ||
+		     streq(canonicalize(mc->m.mnt_dir), file))) {
+			if (has_uuid(spec, mc->m.mnt_fsname+5))
+				return mc;
+		}
+	}
 	return NULL;
 }
 
@@ -249,7 +348,7 @@ getfsfile (const char *file) {
 
 	mc0 = fstab_head();
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
-		if (streq (mc->m.mnt_dir, file))
+		if (streq(mc->m.mnt_dir, file))
 			return mc;
 	return NULL;
 }
@@ -261,7 +360,7 @@ getfsspec (const char *spec) {
 
 	mc0 = fstab_head();
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
-		if (streq (mc->m.mnt_fsname, spec))
+		if (streq(mc->m.mnt_fsname, spec))
 			return mc;
 	return NULL;
 }
@@ -301,9 +400,11 @@ static int we_created_lockfile = 0;
 static int signals_have_been_setup = 0;
 
 /* Ensure that the lock is released if we are interrupted.  */
+extern char *strsignal(int sig);	/* not always in <string.h> */
+
 static void
 handler (int sig) {
-     die (EX_USER, "%s", sys_siglist[sig]);
+     die(EX_USER, "%s", strsignal(sig));
 }
 
 static void
@@ -311,6 +412,15 @@ setlkw_timeout (int sig) {
      /* nothing, fcntl will fail anyway */
 }
 
+/* Remove lock file.  */
+void
+unlock_mtab (void) {
+	if (we_created_lockfile) {
+		unlink (MOUNTED_LOCK);
+		we_created_lockfile = 0;
+	}
+}
+
 /* Create the lock file.
    The lock file will be removed if we catch a signal or when we exit. */
 /* The old code here used flock on a lock file /etc/mtab~ and deleted
@@ -327,13 +437,16 @@ setlkw_timeout (int sig) {
 
 /* Where does the link point to? Obvious choices are mtab and mtab~~.
    HJLu points out that the latter leads to races. Right now we use
-   mtab~.<pid> instead. */
-#define MOUNTLOCK_LINKTARGET	MOUNTED_LOCK "%d"
+   mtab~.<pid> instead. Use 20 as upper bound for the length of %d. */
+#define MOUNTLOCK_LINKTARGET		MOUNTED_LOCK "%d"
+#define MOUNTLOCK_LINKTARGET_LTH	(sizeof(MOUNTED_LOCK)+20)
 
 void
 lock_mtab (void) {
 	int tries = 3;
-	char *linktargetfile;
+	char linktargetfile[MOUNTLOCK_LINKTARGET_LTH];
+
+	at_die = unlock_mtab;
 
 	if (!signals_have_been_setup) {
 		int sig = 0;
@@ -354,9 +467,6 @@ lock_mtab (void) {
 		signals_have_been_setup = 1;
 	}
 
-	/* somewhat clumsy, but some ancient systems do not have snprintf() */
-	/* use 20 as upper bound for the length of %d output */
-	linktargetfile = xmalloc(strlen(MOUNTLOCK_LINKTARGET) + 20);
 	sprintf(linktargetfile, MOUNTLOCK_LINKTARGET, getpid ());
 
 	/* Repeat until it was us who made the link */
@@ -382,6 +492,9 @@ lock_mtab (void) {
 
 		(void) unlink(linktargetfile);
 
+		if (j == 0)
+			we_created_lockfile = 1;
+
 		if (j < 0 && errsv != EEXIST) {
 			die (EX_FILEIO, _("can't link lock file %s: %s "
 			     "(use -n flag to override)"),
@@ -415,7 +528,6 @@ lock_mtab (void) {
 				}
 				/* proceed anyway */
 			}
-			we_created_lockfile = 1;
 		} else {
 			static int tries = 0;
 
@@ -443,15 +555,6 @@ lock_mtab (void) {
 	}
 }
 
-/* Remove lock file.  */
-void
-unlock_mtab (void) {
-	if (we_created_lockfile) {
-		unlink (MOUNTED_LOCK);
-		we_created_lockfile = 0;
-	}
-}
-
 /*
  * Update the mtab.
  *  Used by umount with null INSTEAD: remove the last DIR entry.
@@ -463,11 +566,11 @@ unlock_mtab (void) {
  */
 
 void
-update_mtab (const char *dir, struct mntent *instead) {
+update_mtab (const char *dir, struct my_mntent *instead) {
 	mntFILE *mfp, *mftmp;
 	const char *fnam = MOUNTED;
 	struct mntentchn mtabhead;	/* dummy */
-	struct mntentchn *mc, *mc0, absent;
+	struct mntentchn *mc, *mc0, *absent = NULL;
 
 	if (mtab_does_not_exist() || mtab_is_a_symlink())
 		return;
@@ -490,7 +593,7 @@ update_mtab (const char *dir, struct mntent *instead) {
 
 	/* find last occurrence of dir */
 	for (mc = mc0->prev; mc && mc != mc0; mc = mc->prev)
-		if (streq (mc->m.mnt_dir, dir))
+		if (streq(mc->m.mnt_dir, dir))
 			break;
 	if (mc && mc != mc0) {
 		if (instead == NULL) {
@@ -498,6 +601,7 @@ update_mtab (const char *dir, struct mntent *instead) {
 			if (mc && mc != mc0) {
 				mc->prev->nxt = mc->nxt;
 				mc->nxt->prev = mc->prev;
+				free(mc);
 			}
 		} else {
 			/* A remount */
@@ -505,17 +609,18 @@ update_mtab (const char *dir, struct mntent *instead) {
 		}
 	} else if (instead) {
 		/* not found, add a new entry */
-		absent.m = *instead;
-		absent.nxt = mc0;
-		absent.prev = mc0->prev;
-		mc0->prev = &absent;
+		absent = xmalloc(sizeof(*absent));
+		absent->m = *instead;
+		absent->nxt = mc0;
+		absent->prev = mc0->prev;
+		mc0->prev = absent;
 		if (mc0->nxt == NULL)
-			mc0->nxt = &absent;
+			mc0->nxt = absent;
 	}
 
 	/* write chain to mtemp */
 	mftmp = my_setmntent (MOUNTED_TEMP, "w");
-	if (mftmp == NULL || mfp->mntent_fp == NULL) {
+	if (mftmp == NULL || mftmp->mntent_fp == NULL) {
 		int errsv = errno;
 		error (_("cannot open %s (%s) - mtab not updated"),
 		       MOUNTED_TEMP, strerror (errsv));
@@ -530,6 +635,8 @@ update_mtab (const char *dir, struct mntent *instead) {
 		}
 	}
 
+	discard_mntentchn(mc0);
+
 	if (fchmod (fileno (mftmp->mntent_fp),
 		    S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH) < 0) {
 		int errsv = errno;
