@@ -12,6 +12,7 @@
 int mtab_is_writable(void);
 int mtab_does_not_exist(void);
 int mtab_is_a_symlink(void);
+int is_mounted_once(const char *name);
 
 struct mntentchn {
 	struct mntentchn *nxt, *prev;
@@ -21,7 +22,8 @@ struct mntentchn {
 struct mntentchn *mtab_head (void);
 struct mntentchn *getmntfile (const char *name);
 struct mntentchn *getmntoptfile (const char *file);
-struct mntentchn *getmntfilesbackward (const char *file, struct mntentchn *mc);
+struct mntentchn *getmntdirbackward (const char *dir, struct mntentchn *mc);
+struct mntentchn *getmntdevbackward (const char *dev, struct mntentchn *mc);
 
 struct mntentchn *fstab_head (void);
 struct mntentchn *getfsfile (const char *file);
