@@ -21,7 +21,7 @@ struct minix_super_block {
 	u_char   s_dummy[16];
 	u_char   s_magic[2];
 };
-#define minixmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8))
+#define minixmagic(s)	assemble2le(s.s_magic)
 
 #define ISODCL(from, to) (to - from + 1)
 #define ISO_STANDARD_ID "CD001"
@@ -46,7 +46,7 @@ struct ext_super_block {
 	u_char   s_dummy[56];
 	u_char   s_magic[2];
 };
-#define extmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8))
+#define extmagic(s)	assemble2le(s.s_magic)
 
 #define EXT2_PRE_02B_MAGIC  0xEF51
 #define EXT2_SUPER_MAGIC    0xEF53
@@ -63,7 +63,7 @@ struct ext2_super_block {
 	u_char	s_dummy3[88];
 	u_char	s_journal_inum[4];	/* ext3 only */
 };
-#define ext2magic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8))
+#define ext2magic(s)	assemble2le(s.s_magic)
 
 struct reiserfs_super_block
 {
@@ -96,9 +96,7 @@ struct xiafs_super_block {
     u_char     s_dummy[60];
     u_char     s_magic[4];
 };
-#define xiafsmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8) + \
-			(((uint) s.s_magic[2]) << 16) + \
-			(((uint) s.s_magic[3]) << 24))
+#define xiafsmagic(s)	assemble4le(s.s_magic)
 
 /* From jj@sunsite.ms.mff.cuni.cz Mon Mar 23 15:19:05 1998 */
 #define UFS_SUPER_MAGIC_LE 0x00011954
@@ -107,9 +105,7 @@ struct ufs_super_block {
     u_char     s_dummy[0x55c];
     u_char     s_magic[4];
 };
-#define ufsmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8) + \
-			 (((uint) s.s_magic[2]) << 16) + \
-			 (((uint) s.s_magic[3]) << 24))
+#define ufsmagic(s)	assemble4le(s.s_magic)
 
 /* From Richard.Russon@ait.co.uk Wed Feb 24 08:05:27 1999 */
 #define NTFS_SUPER_MAGIC "NTFS"
@@ -145,35 +141,29 @@ struct xfs_super_block {
 };
 
 #define CRAMFS_SUPER_MAGIC 0x28cd3d45
+#define CRAMFS_SUPER_MAGIC_BE 0x453dcd28
 struct cramfs_super_block {
 	u_char    s_magic[4];
 	u_char    s_dummy[12];
 	u_char    s_id[16];
 };
-#define cramfsmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8) + \
-			 (((uint) s.s_magic[2]) << 16) + \
-			 (((uint) s.s_magic[3]) << 24))
+#define cramfsmagic(s)	assemble4le(s.s_magic)
 
 #define HFS_SUPER_MAGIC 0x4244
 struct hfs_super_block {
-	u_char    s_magic[2];
+	u_char    s_magic[2];		/* drSigWord */
 	u_char    s_dummy[18];
-	u_char    s_blksize[4];
+	u_char    s_blksize[4];		/* drAlBlkSiz */
 };
-#define hfsmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8))
-#define hfsblksize(s)	((uint) s.s_blksize[0] + \
-			 (((uint) s.s_blksize[1]) << 8) + \
-			 (((uint) s.s_blksize[2]) << 16) + \
-			 (((uint) s.s_blksize[3]) << 24))
+#define hfsmagic(s)	assemble2be(s.s_magic)
+#define hfsblksize(s)	assemble4be(s.s_blksize)
 
 #define HPFS_SUPER_MAGIC 0xf995e849
 struct hpfs_super_block {
 	u_char    s_magic[4];
 	u_char    s_magic2[4];
 };
-#define hpfsmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8) + \
-			 (((uint) s.s_magic[2]) << 16) + \
-			 (((uint) s.s_magic[3]) << 24))
+#define hpfsmagic(s)	assemble4le(s.s_magic)
 
 struct adfs_super_block {
 	u_char    s_dummy[448];
@@ -187,27 +177,66 @@ struct adfs_super_block {
 struct vxfs_super_block {
 	u_char	s_magic[4];
 };
-#define vxfsmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8) + \
-			 (((uint) s.s_magic[2]) << 16) + \
-			 (((uint) s.s_magic[3]) << 24))
+#define vxfsmagic(s)	assemble4le(s.s_magic)
 #define VXFS_SUPER_MAGIC 0xa501FCF5
 
 struct jfs_super_block {
 	char	s_magic[4];
-	u_char	s_dummy1[97];
+	u_char	s_version[4];
+	u_char	s_dummy1[93];
 	char	s_fpack[11];
+	u_char	s_dummy2[24];
+	u_char	s_uuid[16];
+	char	s_label[16];
 };
 #define JFS_SUPER1_OFF 0x8000
 #define JFS_MAGIC "JFS1"
 
-#define SYSV_SUPER_MAGIC	0xfd187e20
-
 struct sysv_super_block {
 	u_char  s_dummy1[504];
 	u_char  s_magic[4];
 	u_char  type[4];
 };
+#define sysvmagic(s)		assemble4le(s.s_magic)
+#define SYSV_SUPER_MAGIC	0xfd187e20
 
-#define sysvmagic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8) + \
-			 (((uint) s.s_magic[2]) << 16) + \
-			 (((uint) s.s_magic[3]) << 24))
+struct mdp_super_block {
+	u_char	md_magic[4];
+};
+#define MD_SB_MAGIC	0xa92b4efc
+#define mdsbmagic(s)	assemble4le(s.md_magic)
+
+struct ocfs_volume_header {
+	u_char  minor_version[4];
+	u_char  major_version[4];
+	u_char  signature[128];
+};
+
+struct ocfs_volume_label {
+	u_char  disk_lock[48];
+	u_char  label[64];
+	u_char  label_len[2];
+};
+
+#define ocfslabellen(o)	assemble2le(o.label_len)
+#define OCFS_MAGIC	"OracleCFS"
+
+static inline int
+assemble2le(unsigned char *p) {
+	return (p[0] | (p[1] << 8));
+}
+
+static inline int
+assemble2be(unsigned char *p) {
+	return (p[1] | (p[0] << 8));
+}
+
+static inline int
+assemble4le(unsigned char *p) {
+	return (p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24));
+}
+
+static inline int
+assemble4be(unsigned char *p) {
+	return (p[3] | (p[2] << 8) | (p[1] << 16) | (p[0] << 24));
+}
