@@ -37,6 +37,8 @@
  * 2000-05-11 Mark A. Peloquin <peloquin@us.ibm.com>
  * - check_special_mountprog now returns correct status
  * 2000-11-08 aeb: accept nonnumeric uid=, gid= options
+ * 2001-07-13 Michael K. Johnson <johnsonm@redhat.com>
+ * - implemented -a -O
  */
 
 #include <unistd.h>
@@ -66,6 +68,7 @@
 #include "mount_guess_fstype.h"
 #include "mount_by_label.h"
 #include "getusername.h"
+#include "env.h"
 #include "nls.h"
 
 #define DO_PS_FIDDLING
@@ -77,6 +80,9 @@
 /* True for fake mount (-f).  */
 static int fake = 0;
 
+/* True if we are allowed to call /sbin/mount.${FSTYPE} */
+static int external_allowed = 1;
+
 /* Don't write a entry in /etc/mtab (-n).  */
 static int nomtab = 0;
 
@@ -93,7 +99,7 @@ int sloppy = 0;
 static int readwrite = 0;
 
 /* True for all mount (-a).  */
-int all = 0;
+int mount_all = 0;
 
 /* True for fork() during all mount (-F).  */
 static int optfork = 0;
@@ -107,6 +113,9 @@ static int mounttype = 0;
 /* True if ruid != euid.  */
 static int suid = 0;
 
+/* Contains the fd to read the passphrase from, if any. */
+static int pfd = -1;
+
 /* Map from -o and fstab option strings to the flag argument to mount(2).  */
 struct opt_map {
   const char *opt;		/* option name */
@@ -121,10 +130,11 @@ struct opt_map {
 #define MS_USERS	0x40000000
 #define MS_USER		0x20000000
 #define MS_OWNER	0x10000000
+#define MS_NETDEV	0x00020000
 #define MS_LOOP		0x00010000
 
 /* Options that we keep the mount system call from seeing.  */
-#define MS_NOSYS	(MS_NOAUTO|MS_USERS|MS_USER|MS_LOOP)
+#define MS_NOSYS	(MS_NOAUTO|MS_USERS|MS_USER|MS_NETDEV|MS_LOOP)
 
 /* Options that we keep from appearing in the options field in the mtab.  */
 #define MS_NOMTAB	(MS_REMOUNT|MS_NOAUTO|MS_USERS|MS_USER)
@@ -147,6 +157,7 @@ static const struct opt_map opt_map[] = {
   { "nodev",	0, 0, MS_NODEV	},	/* don't interpret devices */
   { "sync",	0, 0, MS_SYNCHRONOUS},	/* synchronous I/O */
   { "async",	0, 1, MS_SYNCHRONOUS},	/* asynchronous I/O */
+  { "dirsync",	0, 0, MS_DIRSYNC},	/* synchronous directory modifications */
   { "remount",  0, 0, MS_REMOUNT},      /* Alter flags of mounted FS */
   { "bind",	0, 0, MS_BIND   },	/* Remount part of tree elsewhere */
   { "auto",	0, 1, MS_NOAUTO	},	/* Can be mounted using -a */
@@ -157,6 +168,7 @@ static const struct opt_map opt_map[] = {
   { "nouser",	0, 1, MS_USER	},	/* Forbid ordinary user to mount */
   { "owner",	0, 0, MS_OWNER  },	/* Let the owner of the device mount */
   { "noowner",	0, 1, MS_OWNER  },	/* Device owner has no special privs */
+  { "_netdev",	0, 0, MS_NETDEV },	/* Device accessible only via network */
   /* add new options here */
 #ifdef MS_NOSUB
   { "sub",	0, 1, MS_NOSUB	},	/* allow submounts */
@@ -200,25 +212,25 @@ static struct string_opt_map {
 
 static void
 clear_string_opts(void) {
-  struct string_opt_map *m;
+	struct string_opt_map *m;
 
-  for (m = &string_opt_map[0]; m->tag; m++)
-    *(m->valptr) = NULL;
+	for (m = &string_opt_map[0]; m->tag; m++)
+		*(m->valptr) = NULL;
 }
 
 static int
 parse_string_opt(char *s) {
-  struct string_opt_map *m;
-  int lth;
+	struct string_opt_map *m;
+	int lth;
 
-  for (m = &string_opt_map[0]; m->tag; m++) {
-    lth = strlen(m->tag);
-    if (!strncmp(s, m->tag, lth)) {
-      *(m->valptr) = xstrdup(s + lth);
-      return 1;
-    }
-  }
-  return 0;
+	for (m = &string_opt_map[0]; m->tag; m++) {
+		lth = strlen(m->tag);
+		if (!strncmp(s, m->tag, lth)) {
+			*(m->valptr) = xstrdup(s + lth);
+			return 1;
+		}
+	}
+	return 0;
 }
 
 int mount_quiet=0;
@@ -244,7 +256,7 @@ print_one (const struct mntent *me) {
 
 /* Report on everything in mtab (of the specified types if any).  */
 static int
-print_all (string_list types) {
+print_all (char *types) {
      struct mntentchn *mc, *mc0;
 
      mc0 = mtab_head();
@@ -263,54 +275,54 @@ print_all (string_list types) {
  */
 static inline void
 parse_opt (const char *opt, int *mask, char *extra_opts) {
-  const struct opt_map *om;
+	const struct opt_map *om;
 
-  for (om = opt_map; om->opt != NULL; om++)
-      if (streq (opt, om->opt)) {
-	if (om->inv)
-	  *mask &= ~om->mask;
-	else
-	  *mask |= om->mask;
-	if ((om->mask == MS_USER || om->mask == MS_USERS)
-	    && !om->inv)
-	  *mask |= MS_SECURE;
-	if ((om->mask == MS_OWNER) && !om->inv)
-	  *mask |= MS_OWNERSECURE;
+	for (om = opt_map; om->opt != NULL; om++)
+		if (streq (opt, om->opt)) {
+			if (om->inv)
+				*mask &= ~om->mask;
+			else
+				*mask |= om->mask;
+			if ((om->mask == MS_USER || om->mask == MS_USERS)
+			    && !om->inv)
+				*mask |= MS_SECURE;
+			if ((om->mask == MS_OWNER) && !om->inv)
+				*mask |= MS_OWNERSECURE;
 #ifdef MS_SILENT
-        if (om->mask == MS_SILENT && om->inv)  {
-          mount_quiet = 1;
-          verbose = 0;
-        }
+			if (om->mask == MS_SILENT && om->inv)  {
+				mount_quiet = 1;
+				verbose = 0;
+			}
 #endif
-	return;
-      }
+			return;
+		}
 
-  if (*extra_opts)
-    strcat(extra_opts, ",");
+	if (*extra_opts)
+		strcat(extra_opts, ",");
 
-  /* convert nonnumeric ids to numeric */
-  if (!strncmp(opt, "uid=", 4) && !isdigit(opt[4])) {
-	  struct passwd *pw = getpwnam(opt+4);
-	  char uidbuf[20];
+	/* convert nonnumeric ids to numeric */
+	if (!strncmp(opt, "uid=", 4) && !isdigit(opt[4])) {
+		struct passwd *pw = getpwnam(opt+4);
+		char uidbuf[20];
 
-	  if (pw) {
-		  sprintf(uidbuf, "uid=%d", pw->pw_uid);
-		  strcat(extra_opts, uidbuf);
-		  return;
-	  }
-  }
-  if (!strncmp(opt, "gid=", 4) && !isdigit(opt[4])) {
-	  struct group *gr = getgrnam(opt+4);
-	  char gidbuf[20];
+		if (pw) {
+			sprintf(uidbuf, "uid=%d", pw->pw_uid);
+			strcat(extra_opts, uidbuf);
+			return;
+		}
+	}
+	if (!strncmp(opt, "gid=", 4) && !isdigit(opt[4])) {
+		struct group *gr = getgrnam(opt+4);
+		char gidbuf[20];
 
-	  if (gr) {
-		  sprintf(gidbuf, "gid=%d", gr->gr_gid);
-		  strcat(extra_opts, gidbuf);
-		  return;
-	  }
-  }
+		if (gr) {
+			sprintf(gidbuf, "gid=%d", gr->gr_gid);
+			strcat(extra_opts, gidbuf);
+			return;
+		}
+	}
 
-  strcat(extra_opts, opt);
+	strcat(extra_opts, opt);
 }
   
 /* Take -o options list and compute 4th and 5th args to mount(2).  flags
@@ -343,30 +355,31 @@ parse_opts (char *opts, int *flags, char **extra_opts) {
 /* Try to build a canonical options string.  */
 static char *
 fix_opts_string (int flags, const char *extra_opts, const char *user) {
-  const struct opt_map *om;
-  const struct string_opt_map *m;
-  char *new_opts;
+	const struct opt_map *om;
+	const struct string_opt_map *m;
+	char *new_opts;
 
-  new_opts = (flags & MS_RDONLY) ? "ro" : "rw";
-  for (om = opt_map; om->opt != NULL; om++) {
-      if (om->skip)
-	continue;
-      if (om->inv || !om->mask || (flags & om->mask) != om->mask)
-	continue;
-      new_opts = xstrconcat3(new_opts, ",", om->opt);
-      flags &= ~om->mask;
-  }
-  for (m = &string_opt_map[0]; m->tag; m++) {
-      if (!m->skip && *(m->valptr))
-	   new_opts = xstrconcat4(new_opts, ",", m->tag, *(m->valptr));
-  }
-  if (extra_opts && *extra_opts) {
-      new_opts = xstrconcat3(new_opts, ",", extra_opts);
-  }
-  if (user) {
-      new_opts = xstrconcat3(new_opts, ",user=", user);
-  }
-  return new_opts;
+	new_opts = (flags & MS_RDONLY) ? "ro" : "rw";
+	for (om = opt_map; om->opt != NULL; om++) {
+		if (om->skip)
+			continue;
+		if (om->inv || !om->mask || (flags & om->mask) != om->mask)
+			continue;
+		new_opts = xstrconcat3(new_opts, ",", om->opt);
+		flags &= ~om->mask;
+	}
+	for (m = &string_opt_map[0]; m->tag; m++) {
+		if (!m->skip && *(m->valptr))
+			new_opts = xstrconcat4(new_opts, ",",
+					       m->tag, *(m->valptr));
+	}
+	if (extra_opts && *extra_opts) {
+		new_opts = xstrconcat3(new_opts, ",", extra_opts);
+	}
+	if (user) {
+		new_opts = xstrconcat3(new_opts, ",user=", user);
+	}
+	return new_opts;
 }
 
 static int
@@ -440,11 +453,16 @@ static int mountcount = 0;
  */
 static int
 do_mount_syscall (struct mountargs *args) {
-     int ret = mount (args->spec, args->node, args->type,
-		      MS_MGC_VAL | (args->flags), args->data);
-     if (ret == 0)
-	  mountcount++;
-     return ret;
+	int flags = args->flags;
+	int ret;
+
+	if ((flags & MS_MGC_MSK) == 0)
+		flags |= MS_MGC_VAL;
+
+	ret = mount (args->spec, args->node, args->type, flags, args->data);
+	if (ret == 0)
+		mountcount++;
+	return ret;
 }
 
 /*
@@ -452,33 +470,51 @@ do_mount_syscall (struct mountargs *args) {
  *	Mount a single file system. Guess the type when unknown.
  * returns: 0: OK, -1: error in errno, 1: other error
  *	don't exit on non-fatal errors.
+ *	on return types is filled with the type used.
  */
 static int
-guess_fstype_and_mount (char *spec, char *node, char **type,
+guess_fstype_and_mount (char *spec, char *node, char **types,
 			int flags, char *mount_opts) {
    struct mountargs args = { spec, node, NULL, flags & ~MS_NOSYS, mount_opts };
    
-   if (*type && strcasecmp (*type, "auto") == 0)
-      *type = NULL;
+   if (*types && strcasecmp (*types, "auto") == 0)
+      *types = NULL;
 
-   if (!*type && (flags & (MS_BIND | MS_MOVE)))
-      *type = "none";		/* random, but not "bind" */
+   if (!*types && (flags & (MS_BIND | MS_MOVE)))
+      *types = "none";		/* random, but not "bind" */
 
-   if (!*type && !(flags & MS_REMOUNT)) {
-      *type = guess_fstype_from_superblock(spec);
-      if (*type && !strcmp(*type, "swap")) {
+   if (!*types && !(flags & MS_REMOUNT)) {
+      *types = guess_fstype(spec);
+      if (*types && !strcmp(*types, "swap")) {
 	  error(_("%s looks like swapspace - not mounted"), spec);
-	  *type = NULL;
+	  *types = NULL;
 	  return 1;
       }
    }
 
-   if (*type || (flags & MS_REMOUNT)) {
-      args.type = *type;
+   /* Accept a comma-separated list of types, and try them one by one */
+   /* A list like "nonfs,.." indicates types not to use */
+   if (*types && strncmp(*types, "no", 2) && index(*types,',')) {
+      char *t = strdup(*types);
+      char *p;
+
+      while((p = index(t,',')) != NULL) {
+	 *p = 0;
+	 args.type = *types = t;
+	 if(do_mount_syscall (&args) == 0)
+	    return 0;
+	 t = p+1;
+      }
+      /* do last type below */
+      *types = t;
+   }
+
+   if (*types || (flags & MS_REMOUNT)) {
+      args.type = *types;
       return do_mount_syscall (&args);
    }
 
-   return procfsloop(do_mount_syscall, &args, type);
+   return procfsloop(do_mount_syscall, &args, types);
 }
 
 /*
@@ -568,7 +604,8 @@ loop_check(char **spec, char **type, int *flags,
       if (verbose)
 	printf(_("mount: going to use the loop device %s\n"), *loopdev);
       offset = opt_offset ? strtoul(opt_offset, NULL, 0) : 0;
-      if (set_loop (*loopdev, *loopfile, offset, opt_encryption, &loopro)) {
+      if (set_loop(*loopdev, *loopfile, offset,
+		   opt_encryption, pfd, &loopro)) {
 	if (verbose)
 	  printf(_("mount: failed setting up loop device\n"));
 	return EX_FAIL;
@@ -626,6 +663,14 @@ update_mtab_entry(char *spec, char *node, char *type, char *opts,
     }
 }
 
+static void
+set_pfd(char *s) {
+	if (!isdigit(*s))
+		die(EX_USAGE,
+		    _("mount: argument to -p or --pass-fd must be a number"));
+	pfd = atoi(optarg);
+}
+
 static void
 cdrom_setspeed(char *spec) {
 #define CDROM_SELECT_SPEED      0x5322  /* Set the CD-ROM speed */
@@ -634,7 +679,7 @@ cdrom_setspeed(char *spec) {
 	int speed = atoi(opt_speed);
 
 	if ((cdrom = open(spec, O_RDONLY | O_NONBLOCK)) < 0)
-	    die(EX_FAIL, _("mount: cannot not open %s for setting speed"),
+	    die(EX_FAIL, _("mount: cannot open %s for setting speed"),
 		spec);
 	if (ioctl(cdrom, CDROM_SELECT_SPEED, speed) < 0)
 	    die(EX_FAIL, _("mount: cannot set speed: %s"),
@@ -648,7 +693,6 @@ cdrom_setspeed(char *spec) {
  *	If there is a special mount program for this type, exec it.
  * returns: 0: no exec was done, 1: exec was done, status has result
  */
-#define ALWAYS_STAT
 
 static int
 check_special_mountprog(char *spec, char *node, char *type, int flags,
@@ -657,16 +701,10 @@ check_special_mountprog(char *spec, char *node, char *type, int flags,
   struct stat statbuf;
   int res;
 
-  if (type &&
-#ifndef ALWAYS_STAT
-     (streq (type, "smb") || streq (type, "ncp")
-      /* these are incorrect but perhaps used by smbmount or so */
-       || streq (type, "smbfs") || streq (type, "ncpfs")
-     )
-#else
-     (strlen (type) < 100)
-#endif
-  ) {
+  if (!external_allowed)
+      return 0;
+
+  if (type && strlen(type) < 100) {
        sprintf(mountprog, "/sbin/mount.%s", type);
        if (stat(mountprog, &statbuf) == 0) {
 	    res = fork();
@@ -715,7 +753,7 @@ check_special_mountprog(char *spec, char *node, char *type, int flags,
  *      return status from wait
  */
 static int
-try_mount_one (const char *spec0, const char *node0, char *type0,
+try_mount_one (const char *spec0, const char *node0, char *types0,
 	       const char *opts0, int freq, int pass, int bg, int ro) {
   int res, status;
   int mnt5_res = 0;		/* only for gcc */
@@ -724,51 +762,58 @@ try_mount_one (const char *spec0, const char *node0, char *type0,
   char *extra_opts;		/* written in mtab */
   char *mount_opts;		/* actually used on system call */
   const char *opts;
-  char *spec, *node, *type;
+  char *spec, *node, *types;
   char *user = 0;
   int loop = 0;
   char *loopdev = 0, *loopfile = 0;
   struct stat statbuf;
+  int nfs_mount_version = 0;	/* any version */
 
   spec = xstrdup(spec0);
   node = xstrdup(node0);
-  type = xstrdup(type0);
+  types = xstrdup(types0);
   opts = xstrdup(opts0);
 
   parse_opts (xstrdup (opts), &flags, &extra_opts);
 
-  suid_check (spec, node, &flags, &user);
-
   /* quietly succeed for fstab entries that don't get mounted automatically */
-  if (all && (flags & MS_NOAUTO))
+  if (mount_all && (flags & MS_NOAUTO))
     return 0;
 
+  suid_check(spec, node, &flags, &user);
+
   mount_opts = extra_opts;
 
   if (opt_speed)
       cdrom_setspeed(spec);
 
   if (!(flags & MS_REMOUNT)) {
-      /* don't set up a (new) loop device if we only remount - this left
+      /*
+       * Don't set up a (new) loop device if we only remount - this left
        * stale assignments of files to loop devices. Nasty when used for
        * encryption.
        */
-      res = loop_check (&spec, &type, &flags, &loop, &loopdev, &loopfile);
+      res = loop_check(&spec, &types, &flags, &loop, &loopdev, &loopfile);
       if (res)
 	  return res;
   }
 
   /*
    * Call mount.TYPE for types that require a separate mount program.
-   * For the moment these types are ncp and smb.
+   * For the moment these types are ncpfs and smbfs. Maybe also vxfs.
+   * All such special things must occur isolated in the types string.
    */
-  if (check_special_mountprog (spec, node, type, flags, extra_opts, &status))
+  if (check_special_mountprog (spec, node, types, flags, extra_opts, &status))
       return status;
 
-  if (!fake && type && streq (type, "nfs")) {
+  /*
+   * Also nfs requires a separate program, but it is built in.
+   */
+  if (!fake && types && streq (types, "nfs")) {
 #ifdef HAVE_NFS
 retry_nfs:
-    mnt_err = nfsmount (spec, node, &flags, &extra_opts, &mount_opts, bg);
+    mnt_err = nfsmount (spec, node, &flags, &extra_opts, &mount_opts,
+			&nfs_mount_version, bg);
     if (mnt_err)
       return mnt_err;
 #else
@@ -780,7 +825,7 @@ retry_nfs:
   block_signals (SIG_BLOCK);
 
   if (!fake)
-    mnt5_res = guess_fstype_and_mount (spec, node, &type, flags & ~MS_NOSYS,
+    mnt5_res = guess_fstype_and_mount (spec, node, &types, flags & ~MS_NOSYS,
 				       mount_opts);
 
   if (fake || mnt5_res == 0) {
@@ -790,7 +835,7 @@ retry_nfs:
 
       update_mtab_entry(loop ? loopfile : spec,
 			node,
-			type ? type : "unknown",
+			types ? types : "unknown",
 			fix_opts_string (flags & ~MS_NOMTAB, extra_opts, user),
 			flags,
 			freq,
@@ -808,9 +853,8 @@ retry_nfs:
   block_signals (SIG_UNBLOCK);
 
 #ifdef HAVE_NFS
-  if (mnt_err && type && streq (type, "nfs")) {
-      extern int nfs_mount_version;
-      if (nfs_mount_version == 4) {
+  if (mnt_err && types && streq (types, "nfs")) {
+      if (nfs_mount_version == 4 && mnt_err != EBUSY && mnt_err != ENOENT) {
 	  if (verbose)
 	    printf(_("mount: failed with nfs mount version 4, trying 3..\n"));
 	  nfs_mount_version = 3;
@@ -821,7 +865,7 @@ retry_nfs:
 
   /* Mount failed, complain, but don't die.  */
 
-  if (type == 0) {
+  if (types == 0) {
     if (suid)
       error (_("mount: I could not determine the filesystem type, "
 	       "and none was specified"));
@@ -844,11 +888,11 @@ retry_nfs:
     case EBUSY:
       if (flags & MS_REMOUNT) {
 	error (_("mount: %s is busy"), node);
-      } else if (!strcmp(type, "proc") && !strcmp(node, "/proc")) {
+      } else if (!strcmp(types, "proc") && !strcmp(node, "/proc")) {
 	/* heuristic: if /proc/version exists, then probably proc is mounted */
 	if (stat ("/proc/version", &statbuf))   /* proc mounted? */
 	   error (_("mount: %s is busy"), node);   /* no */
-	else if(!all || verbose)                /* yes, don't mention it */
+	else if (!mount_all || verbose)            /* yes, don't mention it */
 	   error (_("mount: proc already mounted"));
       } else {
 	error (_("mount: %s already mounted or %s busy"), spec, node);
@@ -881,7 +925,7 @@ retry_nfs:
       break;
     case EINVAL:
     { int fd;
-      long size;
+      unsigned long size;
       int warned=0;
 
       if (flags & MS_REMOUNT) {
@@ -891,9 +935,9 @@ retry_nfs:
 	       "       or too many mounted file systems"),
 	       spec);
 
-	if (stat (spec, &statbuf) == 0 && S_ISBLK(statbuf.st_mode)
+	if (stat(spec, &statbuf) == 0 && S_ISBLK(statbuf.st_mode)
 	   && (fd = open(spec, O_RDONLY | O_NONBLOCK)) >= 0) {
-	  if(ioctl(fd, BLKGETSIZE, &size) == 0) {
+	  if (ioctl(fd, BLKGETSIZE, &size) == 0) {
 	    if (size == 0) {
 	      warned++;
 	  error ("       (could this be the IDE device where you in fact use\n"
@@ -926,16 +970,16 @@ retry_nfs:
       error (_("mount: %s: can't read superblock"), spec); break;
     case ENODEV:
     { int pfs;
-      if ((pfs = is_in_procfs(type)) == 1 || !strcmp(type, "guess"))
+      if ((pfs = is_in_procfs(types)) == 1 || !strcmp(types, "guess"))
         error(_("mount: %s: unknown device"), spec);
       else if (pfs == 0) {
 	char *lowtype, *p;
 	int u;
 
-	error (_("mount: fs type %s not supported by kernel"), type);
+	error (_("mount: fs type %s not supported by kernel"), types);
 
 	/* maybe this loser asked for FAT or ISO9660 or isofs */
-	lowtype = xstrdup(type);
+	lowtype = xstrdup(types);
 	u = 0;
 	for(p=lowtype; *p; p++) {
 	  if(tolower(*p) != *p) {
@@ -950,7 +994,7 @@ retry_nfs:
 	free(lowtype);
       } else
 	error (_("mount: %s has wrong device number or fs type %s not supported"),
-	       spec, type);
+	       spec, types);
       break;
     }
     case ENOTBLK:
@@ -981,7 +1025,7 @@ retry_nfs:
       } else {
 	 if (loop) {
 	     opts = opts0;
-	     type = type0;
+	     types = types0;
 	 }
          if (opts) {
 	     char *opts1 = realloc(xstrdup(opts), strlen(opts)+4);
@@ -989,11 +1033,11 @@ retry_nfs:
 	     opts = opts1;
          } else
              opts = "ro";
-	 if (type && !strcmp(type, "guess"))
-	     type = 0;
+	 if (types && !strcmp(types, "guess"))
+	     types = 0;
          error (_("mount: %s%s is write-protected, mounting read-only"),
 		bd, spec0);
-	 return try_mount_one (spec0, node0, type, opts, freq, pass, bg, 1);
+	 return try_mount_one (spec0, node0, types, opts, freq, pass, bg, 1);
       }
       break;
     }
@@ -1051,7 +1095,7 @@ usersubst(const char *opts) {
  * Return 0 for success (either mounted sth or -a and NOAUTO was given)
  */
 static int
-mount_one (const char *spec, const char *node, char *type, const char *opts,
+mount_one (const char *spec, const char *node, char *types, const char *opts,
 	   char *cmdlineopts, int freq, int pass) {
   int status;
   int status2;
@@ -1068,11 +1112,23 @@ mount_one (const char *spec, const char *node, char *type, const char *opts,
        opts = xstrconcat3(opts, ",", cmdlineopts);
 
   if (!strncmp(spec, "UUID=", 5)) {
-      nspec = get_spec_by_uuid(spec+5);
       specset = 1;
+      nspec = get_spec_by_uuid(spec+5);
   } else if (!strncmp(spec, "LABEL=", 6)) {
-      nspec = get_spec_by_volume_label(spec+6);
+      const char *nspec2 = NULL;
       specset = 2;
+      nspec = get_spec_by_volume_label(spec+6);
+      if (nspec)
+	      nspec2 = second_occurrence_of_vol_label(spec+6);
+      if (nspec2) {
+	      if (verbose)
+		      printf(_("mount: the label %s occurs on "
+			       "both %s and %s\n"),
+			     spec+6, nspec, nspec2);
+	      die (EX_FAIL,
+		   _("mount: %s duplicate - not mounted"),
+		   spec, _PATH_FSTAB);
+      }
   } else
       nspec = 0;		/* just for gcc */
 
@@ -1082,17 +1138,22 @@ mount_one (const char *spec, const char *node, char *type, const char *opts,
 	  if (verbose > 1)
 		  printf(_("mount: going to mount %s by %s\n"), spec,
 			 (specset==1) ? _("UUID") : _("label"));
-      } else if(!all)
+      } else if(!mount_all)
           die (EX_USAGE, _("mount: no such partition found"));
       /* if -a then we may be rescued by a noauto option */
   }
 
-  if (type == NULL && !mounttype) {
+  if (types == NULL && !mounttype) {
       if (strchr (spec, ':') != NULL) {
-	type = "nfs";
+	types = "nfs";
 	if (verbose)
 	  printf(_("mount: no type was given - "
 		 "I'll assume nfs because of the colon\n"));
+      } else if(!strncmp(spec, "//", 2)) {
+	types = "smbfs";
+	if (verbose)
+	  printf(_("mount: no type was given - "
+		   "I'll assume smbfs because of the // prefix\n"));
       }
   }
 
@@ -1100,7 +1161,7 @@ mount_one (const char *spec, const char *node, char *type, const char *opts,
    * Try to mount the file system. When the exit status is EX_BG,
    * we will retry in the background. Otherwise, we're done.
    */
-  status = try_mount_one (spec, node, type, opts, freq, pass, 0, 0);
+  status = try_mount_one (spec, node, types, opts, freq, pass, 0, 0);
   if (status != EX_BG)
     return status;
 
@@ -1113,10 +1174,10 @@ mount_one (const char *spec, const char *node, char *type, const char *opts,
     return 0;			/* parent returns "success" */
   spec = xstrdup(spec);		/* arguments will be destroyed */
   node = xstrdup(node);		/* by set_proc_name()          */
-  type = xstrdup(type);
+  types = xstrdup(types);
   opts = xstrdup(opts);
   set_proc_name (spec);		/* make a nice "ps" listing */
-  status2 = try_mount_one (spec, node, type, opts, freq, pass, 1, 0);
+  status2 = try_mount_one (spec, node, types, opts, freq, pass, 1, 0);
   if (verbose && status2)
     printf (_("mount: giving up \"%s\"\n"), spec);
   exit (0);			/* child stops here */
@@ -1146,6 +1207,19 @@ mounted (char *spec, char *node) {
      return 0;
 }
 
+/* avoid using stat() on things we are not going to mount anyway.. */
+static int
+has_noauto (char *opts) {
+	char *s;
+
+	if (!opts)
+		return 0;
+	s = strstr(opts, "noauto");
+	if (!s)
+		return 0;
+	return (s == opts || s[-1] == ',') && (s[6] == 0 || s[6] == ',');
+}
+
 /* Mount all filesystems of the specified types except swap and root.  */
 /* With the --fork option: fork and let different incarnations of
    mount handle different filesystems.  However, try to avoid several
@@ -1153,118 +1227,132 @@ mounted (char *spec, char *node) {
 #define DISKMAJOR(m)	(((int) m) & ~0xf)
 
 static int
-mount_all (string_list types, char *options) {
-     struct mntentchn *mc, *mc0, *mtmp;
-     int status = 0;
-     struct stat statbuf;
-     struct child {
-	  pid_t pid;
-	  char *group;
-	  struct mntentchn *mec;
-	  struct mntentchn *meclast;
-	  struct child *nxt;
-     } childhead, *childtail, *cp;
-     char major[22];
-     char *g, *colon;
+do_mount_all (char *types, char *options, char *test_opts) {
+	struct mntentchn *mc, *mc0, *mtmp;
+	int status = 0;
+	struct stat statbuf;
+	struct child {
+		pid_t pid;
+		char *group;
+		struct mntentchn *mec;
+		struct mntentchn *meclast;
+		struct child *nxt;
+	} childhead, *childtail, *cp;
+	char major[22];
+	char *g, *colon;
 
-     /* build a chain of what we have to do, or maybe
-	several chains, one for each major or NFS host */
-     childhead.nxt = 0;
-     childtail = &childhead;
-     mc0 = fstab_head();
-     for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt) {
-	  if (matching_type (mc->m.mnt_type, types)
-	      && !streq (mc->m.mnt_dir, "/")
-	      && !streq (mc->m.mnt_dir, "root")) {
-	       if (mounted (mc->m.mnt_fsname, mc->m.mnt_dir)) {
-		    if (verbose)
-			 printf(_("mount: %s already mounted on %s\n"),
-				mc->m.mnt_fsname, mc->m.mnt_dir);
-	       } else {
-		    mtmp = (struct mntentchn *) xmalloc(sizeof(*mtmp));
-		    *mtmp = *mc;
-		    mtmp->nxt = 0;
-		    g = NULL;
-		    if (optfork) {
-			 if (stat(mc->m.mnt_fsname, &statbuf) == 0 &&
-			     S_ISBLK(statbuf.st_mode)) {
-			      sprintf(major, "#%x", DISKMAJOR(statbuf.st_rdev));
-			      g = major;
-			 }
+	/* build a chain of what we have to do, or maybe
+	   several chains, one for each major or NFS host */
+	childhead.nxt = 0;
+	childtail = &childhead;
+	mc0 = fstab_head();
+	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt) {
+		if (has_noauto (mc->m.mnt_opts))
+			continue;
+		if (matching_type (mc->m.mnt_type, types)
+		    && matching_opts (mc->m.mnt_opts, test_opts)
+		    && !streq (mc->m.mnt_dir, "/")
+		    && !streq (mc->m.mnt_dir, "root")) {
+
+			if (mounted (mc->m.mnt_fsname, mc->m.mnt_dir)) {
+				if (verbose)
+					printf(_("mount: %s already mounted "
+						 "on %s\n"),
+					       mc->m.mnt_fsname,
+					       mc->m.mnt_dir);
+				continue;
+			}
+
+			mtmp = (struct mntentchn *) xmalloc(sizeof(*mtmp));
+			*mtmp = *mc;
+			mtmp->nxt = 0;
+			g = NULL;
+			if (optfork) {
+				if (stat(mc->m.mnt_fsname, &statbuf) == 0 &&
+				    S_ISBLK(statbuf.st_mode)) {
+					sprintf(major, "#%x",
+						DISKMAJOR(statbuf.st_rdev));
+					g = major;
+				}
 #ifdef HAVE_NFS
-			 if (strcmp(mc->m.mnt_type, "nfs") == 0) {
-			      g = xstrdup(mc->m.mnt_fsname);
-			      colon = strchr(g, ':');
-			      if (colon)
-				   *colon = '\0';
-			 }
+				if (strcmp(mc->m.mnt_type, "nfs") == 0) {
+					g = xstrdup(mc->m.mnt_fsname);
+					colon = strchr(g, ':');
+					if (colon)
+						*colon = '\0';
+				}
 #endif
-		    }
-		    if (g) {
-			 for (cp = childhead.nxt; cp; cp = cp->nxt)
-			      if (cp->group && strcmp(cp->group, g) == 0) {
-				   cp->meclast->nxt = mtmp;
-				   cp->meclast = mtmp;
-				   goto fnd;
-			      }
-		    }
-		    cp = (struct child *) xmalloc(sizeof *cp);
-		    cp->nxt = 0;
-		    cp->mec = cp->meclast = mtmp;
-		    cp->group = xstrdup(g);
-		    cp->pid = 0;
-		    childtail->nxt = cp;
-		    childtail = cp;
-	       fnd:;
-	       }
-	  }
-     }
+			}
+			if (g) {
+				for (cp = childhead.nxt; cp; cp = cp->nxt)
+					if (cp->group &&
+					    strcmp(cp->group, g) == 0) {
+						cp->meclast->nxt = mtmp;
+						cp->meclast = mtmp;
+						goto fnd;
+					}
+			}
+			cp = (struct child *) xmalloc(sizeof *cp);
+			cp->nxt = 0;
+			cp->mec = cp->meclast = mtmp;
+			cp->group = xstrdup(g);
+			cp->pid = 0;
+			childtail->nxt = cp;
+			childtail = cp;
+		fnd:;
+
+		}
+	}
 			      
-     /* now do everything */
-     for (cp = childhead.nxt; cp; cp = cp->nxt) {
-	  pid_t p = -1;
-	  if (optfork) {
-	       p = fork();
-	       if (p == -1) {
-	       	    int errsv = errno;
-		    error(_("mount: cannot fork: %s"), strerror (errsv));
-	       }
-	       else if (p != 0)
-		    cp->pid = p;
-	  }
+	/* now do everything */
+	for (cp = childhead.nxt; cp; cp = cp->nxt) {
+		pid_t p = -1;
+		if (optfork) {
+			p = fork();
+			if (p == -1) {
+				int errsv = errno;
+				error(_("mount: cannot fork: %s"),
+				      strerror (errsv));
+			}
+			else if (p != 0)
+				cp->pid = p;
+		}
 
-	  /* if child, or not forked, do the mounting */
-	  if (p == 0 || p == -1) {
-	       for (mc = cp->mec; mc; mc = mc->nxt)
-		    status |= mount_one (mc->m.mnt_fsname, mc->m.mnt_dir,
-					 mc->m.mnt_type, mc->m.mnt_opts,
-					 options, 0, 0);
-	       if (mountcount)
-		    status |= EX_SOMEOK;
-	       if (p == 0)
-		    exit(status);
-	  }
-     }
+		/* if child, or not forked, do the mounting */
+		if (p == 0 || p == -1) {
+			for (mc = cp->mec; mc; mc = mc->nxt) {
+				status |= mount_one (mc->m.mnt_fsname,
+						     mc->m.mnt_dir,
+						     mc->m.mnt_type,
+						     mc->m.mnt_opts,
+						     options, 0, 0);
+			}
+			if (mountcount)
+				status |= EX_SOMEOK;
+			if (p == 0)
+				exit(status);
+		}
+	}
 
-     /* wait for children, if any */
-     while ((cp = childhead.nxt) != NULL) {
-	  childhead.nxt = cp->nxt;
-	  if (cp->pid) {
-	       int ret;
-	  keep_waiting:
-	       if(waitpid(cp->pid, &ret, 0) == -1) {
-		    if (errno == EINTR)
-			 goto keep_waiting;
-		    perror("waitpid");
-	       } else if (WIFEXITED(ret))
-		    status |= WEXITSTATUS(ret);
-	       else
-		    status |= EX_SYSERR;
-	  }
-     }
-     if (mountcount)
-	  status |= EX_SOMEOK;
-     return status;
+	/* wait for children, if any */
+	while ((cp = childhead.nxt) != NULL) {
+		childhead.nxt = cp->nxt;
+		if (cp->pid) {
+			int ret;
+		keep_waiting:
+			if(waitpid(cp->pid, &ret, 0) == -1) {
+				if (errno == EINTR)
+					goto keep_waiting;
+				perror("waitpid");
+			} else if (WIFEXITED(ret))
+				status |= WEXITSTATUS(ret);
+			else
+				status |= EX_SYSERR;
+		}
+	}
+	if (mountcount)
+		status |= EX_SOMEOK;
+	return status;
 }
 
 extern char version[];
@@ -1281,6 +1369,8 @@ static struct option longopts[] = {
 	{ "read-write", 0, 0, 'w' },
 	{ "rw", 0, 0, 'w' },
 	{ "options", 1, 0, 'o' },
+	{ "test-opts", 1, 0, 'O' },
+	{ "pass-fd", 1, 0, 'p' },
 	{ "types", 1, 0, 't' },
 	{ "bind", 0, 0, 128 },
 	{ "replace", 0, 0, 129 },
@@ -1288,6 +1378,9 @@ static struct option longopts[] = {
 	{ "before", 0, 0, 131 },
 	{ "over", 0, 0, 132 },
 	{ "move", 0, 0, 133 },
+	{ "guess-fstype", 1, 0, 134 },
+	{ "rbind", 0, 0, 135 },
+	{ "internal-only", 0, 0, 'i' },
 	{ NULL, 0, 0, 0 }
 };
 
@@ -1303,7 +1396,7 @@ usage (FILE *fp, int n) {
 	  "So far the informational part. Next the mounting.\n"
 	  "The command is `mount [-t fstype] something somewhere'.\n"
 	  "Details found in /etc/fstab may be omitted.\n"
-	  "       mount -a                 : mount all stuff from /etc/fstab\n"
+	  "       mount -a [-t|-O] ...     : mount all stuff from /etc/fstab\n"
 	  "       mount device             : mount device at the known place\n"
 	  "       mount directory          : mount known device here\n"
 	  "       mount -t type dev dir    : ordinary mount command\n"
@@ -1311,9 +1404,11 @@ usage (FILE *fp, int n) {
 	  "a filesystem (of the given type) found on the device.\n"
 	  "One can also mount an already visible directory tree elsewhere:\n"
 	  "       mount --bind olddir newdir\n"
+	  "or move a subtree:\n"
+	  "       mount --move olddir newdir\n"
 	  "A device can be given by name, say /dev/hda1 or /dev/cdrom,\n"
 	  "or by label, using  -L label  or by uuid, using  -U uuid .\n"
-	  "Other options: [-nfFrsvw] [-o options].\n"
+	  "Other options: [-nfFrsvw] [-o options] [-p passwdfd].\n"
 	  "For many more details, say  man 8 mount .\n"
 	));
 /*
@@ -1327,17 +1422,20 @@ usage (FILE *fp, int n) {
 int
 main (int argc, char *argv[]) {
 	int c, result = 0, specseen;
-	char *options = NULL, *spec, *node;
+	char *options = NULL, *test_opts = NULL, *spec, *node;
 	char *volumelabel = NULL;
 	char *uuid = NULL;
-	string_list types = NULL;
+	char *types = NULL;
 	struct mntentchn *mc;
 	int fd;
 
+	sanitize_env();
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 
+	umask(033);
+
 	/* People report that a mount called from init without console
 	   writes error messages to /etc/mtab
 	   Let us try to avoid getting fd's 0,1,2 */
@@ -1349,13 +1447,13 @@ main (int argc, char *argv[]) {
 	initproctitle(argc, argv);
 #endif
 
-	while ((c = getopt_long (argc, argv, "afFhlL:no:rsU:vVwt:",
-				 longopts, NULL)) != EOF) {
+	while ((c = getopt_long (argc, argv, "afFhilL:no:O:p:rsU:vVwt:",
+				 longopts, NULL)) != -1) {
 		switch (c) {
-		case 'a':		/* mount everything in fstab */
-			++all;
+		case 'a':	       /* mount everything in fstab */
+			++mount_all;
 			break;
-		case 'f':		/* fake: don't actually call mount(2) */
+		case 'f':	       /* fake: don't actually call mount(2) */
 			++fake;
 			break;
 		case 'F':
@@ -1364,6 +1462,9 @@ main (int argc, char *argv[]) {
 		case 'h':		/* help */
 			usage (stdout, 0);
 			break;
+		case 'i':
+			external_allowed = 0;
+			break;
 		case 'l':
 			list_with_volumelabel = 1;
 			break;
@@ -1379,6 +1480,15 @@ main (int argc, char *argv[]) {
 			else
 				options = xstrdup(optarg);
 			break;
+		case 'O':		/* with -t: mount only if (not) opt */
+			if (test_opts)
+				test_opts = xstrconcat3(test_opts, ",", optarg);
+			else
+				test_opts = xstrdup(optarg);
+			break;
+		case 'p':		/* fd on which to read passwd */
+			set_pfd(optarg);
+			break;
 		case 'r':		/* mount readonly */
 			readonly = 1;
 			readwrite = 0;
@@ -1387,7 +1497,7 @@ main (int argc, char *argv[]) {
 			sloppy = 1;
 			break;
 		case 't':		/* specify file system types */
-			types = parse_list (optarg);
+			types = optarg;
 			break;
 		case 'U':
 			uuid = optarg;
@@ -1423,7 +1533,20 @@ main (int argc, char *argv[]) {
 		case 133: /* move */
 			mounttype = MS_MOVE;
 			break;
-	      
+		case 134:
+			/* undocumented, may go away again:
+			   call: mount --guess-fstype device
+			   use only for testing purposes -
+			   the guessing is not reliable at all */
+		    {
+			char *fstype;
+			fstype = do_guess_fstype(optarg);
+			printf("%s\n", fstype ? fstype : "unknown");
+			exit(fstype ? 0 : EX_FAIL);
+		    }
+		case 135:
+			mounttype = (MS_BIND | MS_REC);
+			break;
 		case '?':
 		default:
 			usage (stderr, EX_USAGE);
@@ -1435,7 +1558,7 @@ main (int argc, char *argv[]) {
 
 	specseen = (uuid || volumelabel) ? 1 : 0; 	/* yes, .. i know */
 
-	if (argc+specseen == 0 && !all) {
+	if (argc+specseen == 0 && !mount_all) {
 		if (options || mounttype)
 			usage (stderr, EX_USAGE);
 		return print_all (types);
@@ -1443,7 +1566,7 @@ main (int argc, char *argv[]) {
 
 	if (getuid () != geteuid ()) {
 		suid = 1;
-		if (types || options || readwrite || nomtab || all ||
+		if (types || options || readwrite || nomtab || mount_all ||
 		    fake || mounttype || (argc + specseen) != 1)
 			die (EX_USAGE, _("mount: only root can do that"));
 	}
@@ -1458,8 +1581,16 @@ main (int argc, char *argv[]) {
 	if (specseen) {
 		if (uuid)
 			spec = get_spec_by_uuid(uuid);
-		else
+		else {
+			const char *spec2;
 			spec = get_spec_by_volume_label(volumelabel);
+			spec2 = second_occurrence_of_vol_label(volumelabel);
+			if (spec2)
+				die (EX_FAIL,
+				     _("mount: the label %s occurs on "
+				       "both %s and %s - not mounted\n"),
+				     volumelabel, spec, spec2);
+		}
 		if (!spec)
 			die (EX_USAGE, _("mount: no such partition found"));
 		if (verbose)
@@ -1470,9 +1601,9 @@ main (int argc, char *argv[]) {
 	switch (argc+specseen) {
 	case 0:
 		/* mount -a */
-		result = mount_all (types, options);
+		result = do_mount_all (types, options, test_opts);
 		if (result == 0 && verbose)
-			error(_("not mounted anything"));
+			error(_("nothing was mounted"));
 		break;
 
 	case 1:
@@ -1524,14 +1655,7 @@ main (int argc, char *argv[]) {
 			spec = argv[0];
 			node = argv[1];
 		}
-		if (types == NULL)
-			result = mount_one (spec, node, NULL, NULL,
-					    options, 0, 0);
-		else if (cdr (types) == NULL)
-			result = mount_one (spec, node, car (types), NULL,
-					    options, 0, 0);
-		else
-			usage (stderr, EX_USAGE);
+		result = mount_one (spec, node, types, NULL, options, 0, 0);
 		break;
       
 	default:
