@@ -18,11 +18,14 @@
  *  so useful anymore.]
  *
  * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
- * - added Native Language Support
+ * added Native Language Support
  *
- * Fri Dec  1 23:31:00 2000: Sepp Wijnands <mrrazz@garbage-coderz.net>
+ * 2000-12-01 Sepp Wijnands <mrrazz@garbage-coderz.net>
  * added probes for cramfs, hfs, hpfs and adfs.
  *
+ * 2001-10-26 Tim Launchbury
+ * added sysv magic.
+ *
  * aeb - many changes.
  *
  */
@@ -58,50 +61,94 @@ swapped(unsigned short a) {
 }
 
 /*
-    char *guess_fstype_from_superblock(const char *device);
+    char *guess_fstype(const char *device);
 
     Probes the device and attempts to determine the type of filesystem
     contained within.
 
     Original routine by <jmorriso@bogomips.ww.ubc.ca>; made into a function
     for mount(8) by Mike Grupenhoff <kashmir@umiacs.umd.edu>.
-    Read the superblock only once - aeb
-    Added iso9660, romfs, qnx4, udf, swap - aeb
-    Added a test for high sierra (iso9660) - quinlan@bucknell.edu
     Corrected the test for xiafs - aeb
+    Read the superblock only once - aeb
+    Added a very weak heuristic for vfat - aeb
+    Added iso9660, minix-v2, romfs, qnx4, udf, vxfs, swap - aeb
+    Added a test for high sierra (iso9660) - quinlan@bucknell.edu
     Added ufs from a patch by jj. But maybe there are several types of ufs?
     Added ntfs from a patch by Richard Russon.
-    Added a very weak heuristic for vfat - aeb
     Added xfs - 2000-03-21 Martin K. Petersen <mkp@linuxcare.com>
     Added cramfs, hfs, hpfs, adfs - Sepp Wijnands <mrrazz@garbage-coderz.net>
+    Added ext3 - Andrew Morton
+    Added jfs - Christoph Hellwig
+    Added sysv - Tim Launchbury
+    Added udf - Bryce Nesbitt
 */
 static char
 *magic_known[] = {
-	"adfs", "bfs", "cramfs", "ext", "ext2",
-	"hfs", "hpfs", "iso9660", "minix", "ntfs",
-	"qnx4", "romfs", "swap", "udf", "ufs",
-	"xfs", "xiafs"
+	"adfs", "bfs", "cramfs", "ext", "ext2", "ext3",
+	"hfs", "hpfs", "iso9660", "jfs", "minix", "ntfs",
+	"qnx4", "reiserfs", "romfs", "swap", "sysv", "udf", "ufs",
+	"vxfs", "xfs", "xiafs"
 };
 
-static int
-tested(const char *device) {
-    char **m;
+static struct tried {
+	struct tried *next;
+	char *type;
+} *tried = NULL;
 
-    for (m = magic_known; m - magic_known < SIZE(magic_known); m++)
-        if (!strcmp(*m, device))
-	    return 1;
-    return 0;
+static int
+was_tested(const char *fstype) {
+	char **m;
+	struct tried *t;
+
+	for (m = magic_known; m - magic_known < SIZE(magic_known); m++)
+		if (!strcmp(*m, fstype))
+			return 1;
+	for (t = tried; t; t = t->next) {
+		if (!strcmp(t->type, fstype))
+			return 1;
+	}
+	return 0;
 }
 
-/* udf magic - I find that trying to mount garbage as an udf fs
-   causes a very large kernel delay, almost killing the machine.
-   So, we do not try udf unless there is positive evidence that it
-   might work. Try iso9660 first, it is much more likely.
-   Strings below taken from ECMA 167. */
+static void
+set_tested(const char *fstype) {
+	struct tried *t = xmalloc(sizeof(struct tried));
+
+	t->next = tried;
+	t->type = xstrdup(fstype);
+	tried = t;
+}
+
+static void
+free_tested(void) {
+	struct tried *t, *tt;
+
+	t = tried;
+	while(t) {
+		free(t->type);
+		tt = t->next;
+		free(t);
+		t = tt;
+	}
+	tried = NULL;
+}
+
+/*
+ * udf magic - I find that trying to mount garbage as an udf fs
+ * causes a very large kernel delay, almost killing the machine.
+ * So, we do not try udf unless there is positive evidence that it
+ * might work. Strings below taken from ECMA 167.
+ */
+/*
+ * It seems that before udf 2.00 the volume descriptor was not well
+ * defined.  For 2.00 you're supposed to keep scanning records until
+ * you find one NOT in this list.  (See ECMA 2/8.3.1).
+ */
 static char
 *udf_magic[] = { "BEA01", "BOOT2", "CD001", "CDW02", "NSR02",
 		 "NSR03", "TEA01" };
 
+
 static int
 may_be_udf(const char *id) {
     char **m;
@@ -112,6 +159,43 @@ may_be_udf(const char *id) {
     return 0;
 }
 
+/* we saw "CD001" - may be iso9660 or udf - Bryce Nesbitt */
+static int
+is_really_udf(int fd) {
+	int j, bs;
+	struct iso_volume_descriptor isosb;
+
+	/* determine the block size by scanning in 2K increments
+	   (block sizes larger than 2K will be null padded) */
+	for (bs = 1; bs < 16; bs++) {
+		lseek(fd, bs*2048+32768, SEEK_SET);
+		if (read(fd, (char *)&isosb, sizeof(isosb)) != sizeof(isosb))
+			return 0;
+		if (isosb.id[0])
+			break;
+	}
+
+	/* Scan up to another 64 blocks looking for additional VSD's */
+	for (j = 1; j < 64; j++) {
+		if (j > 1) {
+			lseek(fd, j*bs*2048+32768, SEEK_SET);
+			if (read(fd, (char *)&isosb, sizeof(isosb))
+			    != sizeof(isosb))
+				return 0;
+		}
+		/* If we find NSR0x then call it udf:
+		   NSR01 for UDF 1.00
+		   NSR02 for UDF 1.50
+		   NSR03 for UDF 2.00 */
+		if (!strncmp(isosb.id, "NSR0", 4))
+			return 1;
+		if (!may_be_udf(isosb.id))
+			return 0;
+	}
+
+	return 0;
+}
+
 static int
 may_be_swap(const char *s) {
 	return (strncmp(s-10, "SWAP-SPACE", 10) == 0 ||
@@ -132,15 +216,24 @@ may_be_adfs(const u_char *s) {
 	return (sum == p[511]);
 }
 
-static char *
-fstype(const char *device) {
+static int is_reiserfs_magic_string (struct reiserfs_super_block * rs)
+{
+    return (!strncmp (rs->s_magic, REISERFS_SUPER_MAGIC_STRING, 
+		      strlen ( REISERFS_SUPER_MAGIC_STRING)) ||
+	    !strncmp (rs->s_magic, REISER2FS_SUPER_MAGIC_STRING, 
+		      strlen ( REISER2FS_SUPER_MAGIC_STRING)));
+}
+
+char *
+do_guess_fstype(const char *device) {
     int fd;
     char *type = NULL;
     union {
 	struct minix_super_block ms;
 	struct ext_super_block es;
 	struct ext2_super_block e2s;
-    } sb;
+	struct vxfs_super_block vs;
+    } sb;			/* stuff at 1024 */
     union {
 	struct xiafs_super_block xiasb;
 	char romfs_magic[8];
@@ -156,67 +249,61 @@ fstype(const char *device) {
 	struct iso_volume_descriptor iso;
 	struct hs_volume_descriptor hs;
     } isosb;
+    struct reiserfs_super_block reiserfssb;	/* block 64 or 8 */
+    struct jfs_super_block jfssb;		/* block 32 */
     struct hfs_super_block hfssb;
     struct hpfs_super_block hpfssb;
     struct adfs_super_block adfssb;
+    struct sysv_super_block svsb;
     struct stat statbuf;
 
     /* opening and reading an arbitrary unknown path can have
        undesired side effects - first check that `device' refers
-       to a block device */
-    if (stat (device, &statbuf) || !S_ISBLK(statbuf.st_mode))
+       to a block device or ordinary file */
+    if (stat (device, &statbuf) ||
+	!(S_ISBLK(statbuf.st_mode) || S_ISREG(statbuf.st_mode)))
       return 0;
 
     fd = open(device, O_RDONLY);
     if (fd < 0)
       return 0;
 
-    if (lseek(fd, 1024, SEEK_SET) != 1024
-	|| read(fd, (char *) &sb, sizeof(sb)) != sizeof(sb))
-	 goto io_error;
-
-    /* ext2 has magic in little-endian on disk, so "swapped" is
-       superfluous; however, there have existed strange byteswapped
-       PPC ext2 systems */
-    if (ext2magic(sb.e2s) == EXT2_SUPER_MAGIC
-	|| ext2magic(sb.e2s) == EXT2_PRE_02B_MAGIC
-	|| ext2magic(sb.e2s) == swapped(EXT2_SUPER_MAGIC))
-	 type = "ext2";
-
-    else if (minixmagic(sb.ms) == MINIX_SUPER_MAGIC
-	     || minixmagic(sb.ms) == MINIX_SUPER_MAGIC2
-	     || minixmagic(sb.ms) == swapped(MINIX_SUPER_MAGIC2))
-	 type = "minix";
-
-    else if (extmagic(sb.es) == EXT_SUPER_MAGIC)
-	 type = "ext";
+    /* do seeks and reads in disk order, otherwise a very short
+       partition may cause a failure because of read error */
 
     if (!type) {
+	 /* block 0 */
 	 if (lseek(fd, 0, SEEK_SET) != 0
 	     || read(fd, (char *) &xsb, sizeof(xsb)) != sizeof(xsb))
-	      goto io_error;
+	      goto try_iso9660;
+	 /* Gyorgy Kovesdi: none of my photocds has a readable block 0 */
 
 	 if (xiafsmagic(xsb.xiasb) == _XIAFS_SUPER_MAGIC)
 	      type = "xiafs";
 	 else if(!strncmp(xsb.romfs_magic, "-rom1fs-", 8))
 	      type = "romfs";
-	 else if(!strncmp(xsb.xfsb.s_magic, XFS_SUPER_MAGIC, 4) ||
-		 !strncmp(xsb.xfsb.s_magic, XFS_SUPER_MAGIC2, 4))
+	 else if(!strncmp(xsb.xfsb.s_magic, XFS_SUPER_MAGIC, 4))
 	      type = "xfs";
 	 else if(!strncmp(xsb.qnx4fs_magic+4, "QNX4FS", 6))
-	      type = "qnx4fs";
+	      type = "qnx4";
 	 else if(xsb.bfs_magic == 0x1badface)
 	      type = "bfs";
 	 else if(!strncmp(xsb.ntfssb.s_magic, NTFS_SUPER_MAGIC,
 			  sizeof(xsb.ntfssb.s_magic)))
 	      type = "ntfs";
-	 else if(cramfsmagic(xsb.cramfssb) == CRAMFS_SUPER_MAGIC)
+	 else if(cramfsmagic(xsb.cramfssb) == CRAMFS_SUPER_MAGIC ||
+		 cramfsmagic(xsb.cramfssb) == CRAMFS_SUPER_MAGIC_BE)
 	      type = "cramfs";
 	 else if ((!strncmp(xsb.fatsb.s_os, "MSDOS", 5) ||
 		   !strncmp(xsb.fatsb.s_os, "MSWIN", 5) ||
 		   !strncmp(xsb.fatsb.s_os, "MTOOL", 5) ||
+		   !strncmp(xsb.fatsb.s_os, "IBM", 3) ||
+		   !strncmp(xsb.fatsb.s_os, "DRDOS", 5) ||
 		   !strncmp(xsb.fatsb.s_os, "mkdosfs", 7) ||
-		   !strncmp(xsb.fatsb.s_os, "kmkdosfs", 8))
+		   !strncmp(xsb.fatsb.s_os, "kmkdosfs", 8) ||
+		   /* Michal Svec: created by fdformat, old msdos utility for
+		      formatting large (1.7) floppy disks. */
+		   !strncmp(xsb.fatsb.s_os, "CH-FOR18", 8))
 		  && (!strncmp(xsb.fatsb.s_fs, "FAT12   ", 8) ||
 		      !strncmp(xsb.fatsb.s_fs, "FAT16   ", 8) ||
 		      !strncmp(xsb.fatsb.s_fs2, "FAT32   ", 8)))
@@ -224,52 +311,65 @@ fstype(const char *device) {
     }
 
     if (!type) {
-	 if (lseek(fd, 8192, SEEK_SET) != 8192
-	     || read(fd, (char *) &ufssb, sizeof(ufssb)) != sizeof(ufssb))
-	      goto io_error;
-
-	 if (ufsmagic(ufssb) == UFS_SUPER_MAGIC) /* also test swapped version? */
-	      type = "ufs";
+	    /* sector 1 */
+	    if (lseek(fd, 512 , SEEK_SET) != 512
+		|| read(fd, (char *) &svsb, sizeof(svsb)) != sizeof(svsb))
+		    goto io_error;
+	    if (sysvmagic(svsb) == SYSV_SUPER_MAGIC )
+		    type = "sysv";
     }
 
     if (!type) {
-	 if (lseek(fd, 0x8000, SEEK_SET) != 0x8000
-	     || read(fd, (char *) &isosb, sizeof(isosb)) != sizeof(isosb))
-	      goto io_error;
+	/* block 1 */
+	if (lseek(fd, 1024, SEEK_SET) != 1024 ||
+	    read(fd, (char *) &sb, sizeof(sb)) != sizeof(sb))
+		goto io_error;
 
-	 if(strncmp(isosb.iso.id, ISO_STANDARD_ID, sizeof(isosb.iso.id)) == 0
-	    || strncmp(isosb.hs.id, HS_STANDARD_ID, sizeof(isosb.hs.id)) == 0)
-	      type = "iso9660";
-	 else if (may_be_udf(isosb.iso.id))
-	      type = "udf";
+	/* ext2 has magic in little-endian on disk, so "swapped" is
+	   superfluous; however, there have existed strange byteswapped
+	   PPC ext2 systems */
+	if (ext2magic(sb.e2s) == EXT2_SUPER_MAGIC ||
+	    ext2magic(sb.e2s) == EXT2_PRE_02B_MAGIC ||
+	    ext2magic(sb.e2s) == swapped(EXT2_SUPER_MAGIC)) {
+		type = "ext2";
+
+	     /* maybe even ext3? */
+	     if ((assemble4le(sb.e2s.s_feature_compat)
+		  & EXT3_FEATURE_COMPAT_HAS_JOURNAL) &&
+		 assemble4le(sb.e2s.s_journal_inum) != 0)
+		     type = "ext3";	/* "ext3,ext2" */
+	}
+
+	else if (minixmagic(sb.ms) == MINIX_SUPER_MAGIC ||
+		 minixmagic(sb.ms) == MINIX_SUPER_MAGIC2 ||
+		 minixmagic(sb.ms) == swapped(MINIX_SUPER_MAGIC2) ||
+		 minixmagic(sb.ms) == MINIX2_SUPER_MAGIC ||
+		 minixmagic(sb.ms) == MINIX2_SUPER_MAGIC2)
+		type = "minix";
+
+	else if (extmagic(sb.es) == EXT_SUPER_MAGIC)
+		type = "ext";
+
+	else if (vxfsmagic(sb.vs) == VXFS_SUPER_MAGIC)
+		type = "vxfs";
     }
 
     if (!type) {
+	/* block 1 */
         if (lseek(fd, 0x400, SEEK_SET) != 0x400
             || read(fd, (char *) &hfssb, sizeof(hfssb)) != sizeof(hfssb))
              goto io_error;
 
         /* also check if block size is equal to 512 bytes,
-           since the hfs driver currently only has support
-           for block sizes of 512 bytes long, and to be
-           more accurate (sb magic is only a short int) */
-        if ((hfsmagic(hfssb) == HFS_SUPER_MAGIC &&
-	     hfsblksize(hfssb) == 0x20000) ||
-            (swapped(hfsmagic(hfssb)) == HFS_SUPER_MAGIC &&
-             hfsblksize(hfssb) == 0x200))
+	   or a multiple. (I see 1536 here.) */
+        if (hfsmagic(hfssb) == HFS_SUPER_MAGIC &&	/* always BE */
+	    hfsblksize(hfssb) != 0 &&
+	    (hfsblksize(hfssb) & 0x1ff) == 0)
              type = "hfs";
     }
 
     if (!type) {
-        if (lseek(fd, 0x2000, SEEK_SET) != 0x2000
-            || read(fd, (char *) &hpfssb, sizeof(hpfssb)) != sizeof(hpfssb))
-             goto io_error;
-
-        if (hpfsmagic(hpfssb) == HPFS_SUPER_MAGIC)
-             type = "hpfs";
-    }
-
-    if (!type) {
+	/* block 3 */
         if (lseek(fd, 0xc00, SEEK_SET) != 0xc00
             || read(fd, (char *) &adfssb, sizeof(adfssb)) != sizeof(adfssb))
              goto io_error;
@@ -281,6 +381,80 @@ fstype(const char *device) {
              type = "adfs";
     }
 
+    if (!type) {
+	 int mag;
+
+	 /* block 8 */
+	 if (lseek(fd, 8192, SEEK_SET) != 8192
+	     || read(fd, (char *) &ufssb, sizeof(ufssb)) != sizeof(ufssb))
+	      goto io_error;
+
+	 mag = ufsmagic(ufssb);
+	 if (mag == UFS_SUPER_MAGIC_LE || mag == UFS_SUPER_MAGIC_BE)
+	      type = "ufs";
+    }
+
+    if (!type) {
+	/* block 8 */
+	if (lseek(fd, REISERFS_OLD_DISK_OFFSET_IN_BYTES, SEEK_SET) !=
+				REISERFS_OLD_DISK_OFFSET_IN_BYTES
+	    || read(fd, (char *) &reiserfssb, sizeof(reiserfssb)) !=
+		sizeof(reiserfssb))
+	    goto io_error;
+	if (is_reiserfs_magic_string(&reiserfssb))
+	    type = "reiserfs";
+    }
+
+    if (!type) {
+	/* block 8 */
+        if (lseek(fd, 0x2000, SEEK_SET) != 0x2000
+            || read(fd, (char *) &hpfssb, sizeof(hpfssb)) != sizeof(hpfssb))
+             goto io_error;
+
+        if (hpfsmagic(hpfssb) == HPFS_SUPER_MAGIC)
+             type = "hpfs";
+    }
+
+    if (!type) {
+	 /* block 32 */
+	 if (lseek(fd, JFS_SUPER1_OFF, SEEK_SET) != JFS_SUPER1_OFF
+	     || read(fd, (char *) &jfssb, sizeof(jfssb)) != sizeof(jfssb))
+	      goto io_error;
+	 if (!strncmp(jfssb.s_magic, JFS_MAGIC, 4))
+	      type = "jfs";
+    }
+
+    if (!type) {
+	 /* block 32 */
+    try_iso9660:
+	 if (lseek(fd, 0x8000, SEEK_SET) != 0x8000
+	     || read(fd, (char *) &isosb, sizeof(isosb)) != sizeof(isosb))
+	      goto io_error;
+
+	 if (strncmp(isosb.hs.id, HS_STANDARD_ID, sizeof(isosb.hs.id)) == 0) {
+		 /* "CDROM" */
+		 type = "iso9660";
+	 } else if (strncmp(isosb.iso.id, ISO_STANDARD_ID,
+			  sizeof(isosb.iso.id)) == 0) {
+		 /* CD001 */
+		 type = "iso9660";
+		 if (is_really_udf(fd))
+			 type = "udf";
+	 } else if (may_be_udf(isosb.iso.id))
+		 type = "udf";
+    }
+
+    if (!type) {
+	/* block 64 */
+	if (lseek(fd, REISERFS_DISK_OFFSET_IN_BYTES, SEEK_SET) !=
+		REISERFS_DISK_OFFSET_IN_BYTES
+	    || read(fd, (char *) &reiserfssb, sizeof(reiserfssb)) !=
+		sizeof(reiserfssb))
+	    goto io_error;
+	if (is_reiserfs_magic_string(&reiserfssb))
+	    type = "reiserfs";
+    }
+
     if (!type) {
 	    /* perhaps the user tries to mount the swap space
 	       on a new disk; warn her before she does mke2fs on it */
@@ -305,14 +479,17 @@ fstype(const char *device) {
     return(type);
 
 io_error:
-    perror(device);
+    if (errno)
+	 perror(device);
+    else
+	 fprintf(stderr, _("mount: error while guessing filesystem type\n"));
     close(fd);
     return 0;
 }
 
 char *
-guess_fstype_from_superblock(const char *spec) {
-	char *type = fstype(spec);
+guess_fstype(const char *spec) {
+	char *type = do_guess_fstype(spec);
 	if (verbose) {
 	    printf (_("mount: you didn't specify a filesystem type for %s\n"),
 		    spec);
@@ -363,19 +540,28 @@ is_in_procfs(const char *type) {
     return ret;
 }
 
+/* Try all types in FILESYSTEMS, except those in *types,
+   in case *types starts with "no" */
 /* return: 0: OK, -1: error in errno, 1: type not found */
-/* when 1 is returned, *type is NULL */
+/* when 0 or -1 is returned, *types contains the type used */
+/* when 1 is returned, *types is NULL */
 int
 procfsloop(int (*mount_fn)(struct mountargs *), struct mountargs *args,
-	   char **type) {
+	   char **types) {
 	char *files[2] = { ETC_FILESYSTEMS, PROC_FILESYSTEMS };
 	FILE *procfs;
 	char *fsname;
+	char *notypes = NULL;
+	int no = 0;
 	int ret = 1;
 	int errsv = 0;
 	int i;
 
-	*type = NULL;
+	if (*types && !strncmp(*types, "no", 2)) {
+		no = 1;
+		notypes = (*types) + 2;
+	}
+	*types = NULL;
 
 	/* Use PROC_FILESYSTEMS only when ETC_FILESYSTEMS does not exist.
 	   In some cases trying a filesystem that the kernel knows about
@@ -394,25 +580,29 @@ procfsloop(int (*mount_fn)(struct mountargs *), struct mountargs *args,
 				fclose(procfs);
 				goto nexti;
 			}
-			if (tested (fsname))
+			if (was_tested (fsname))
 				continue;
+			if (no && matching_type(fsname, notypes))
+				continue;
+			set_tested (fsname);
 			args->type = fsname;
 			if (verbose) {
 				printf(_("Trying %s\n"), fsname);
 				fflush(stdout);
 			}
 			if ((*mount_fn) (args) == 0) {
-				*type = fsname;
+				*types = fsname;
 				ret = 0;
 				break;
 			} else if (errno != EINVAL &&
 				   is_in_procfs(fsname) == 1) {
-				*type = "guess";
+				*types = "guess";
 				ret = -1;
 				errsv = errno;
 				break;
 			}
 		}
+		free_tested();
 		fclose(procfs);
 		errno = errsv;
 		return ret;
