@@ -23,6 +23,7 @@
  *               in mtab, try them all, with last one tried first
  *             - Differentiate "user" and "users" key words in fstab
  * 001202: aeb - remove at most one line from /etc/mtab
+ * 010716: Michael K. Johnson <johnsonm@redhat.com: -a -O
  * 010914: Jamie Strandboge - use tcp if that was used for mount
  * 011005: hch - add lazy umount support
  * 020105: aeb - permission test owner umount
@@ -278,7 +279,9 @@ umount_one (const char *spec, const char *node, const char *type,
 	if (force) {		/* only supported for NFS */
 		res = umount2 (node, MNT_FORCE);
 		if (res == -1) {
+			int errsv = errno;
 			perror("umount2");
+			errno = errsv;
 			if (errno == ENOSYS) {
 				if (verbose)
 					printf(_("no umount2, trying umount...\n"));
@@ -399,15 +402,23 @@ umount_one (const char *spec, const char *node, const char *type,
  * In both cases, it is best to try the last occurrence first.
  */
 static int
-umount_one_bw (const char *file, struct mntentchn *mc) {
-     int res = 1;
+umount_one_bw (const char *file, struct mntentchn *mc0) {
+	struct mntentchn *mc;
+	int res = 1;
 
-     while (res && mc) {
-	  res = umount_one(mc->m.mnt_fsname, mc->m.mnt_dir,
-			   mc->m.mnt_type, mc->m.mnt_opts, mc);
-	  mc = getmntfilesbackward (file, mc);
-     }
-     return res;
+	mc = mc0;
+	while (res && mc) {
+		res = umount_one(mc->m.mnt_fsname, mc->m.mnt_dir,
+				 mc->m.mnt_type, mc->m.mnt_opts, mc);
+		mc = getmntdirbackward(file, mc);
+	}
+	mc = mc0;
+	while (res && mc) {
+		res = umount_one(mc->m.mnt_fsname, mc->m.mnt_dir,
+				 mc->m.mnt_type, mc->m.mnt_opts, mc);
+		mc = getmntdevbackward(file, mc);
+	}
+	return res;
 }
 
 /* Unmount all filesystems of type VFSTYPES found in mtab.  Since we are
@@ -416,7 +427,7 @@ umount_one_bw (const char *file, struct mntentchn *mc) {
    in any case it's important to umount mtab entries in reverse order
    to mount, e.g. /usr/spool before /usr.  */
 static int
-umount_all (char *types) {
+umount_all (char *types, char *test_opts) {
      struct mntentchn *mc, *hd;
      int errors = 0;
 
@@ -424,7 +435,8 @@ umount_all (char *types) {
      if (!hd->prev)
 	  die (2, _("umount: cannot find list of filesystems to unmount"));
      for (mc = hd->prev; mc != hd; mc = mc->prev) {
-	  if (matching_type (mc->m.mnt_type, types)) {
+	  if (matching_type (mc->m.mnt_type, types)
+	      && matching_opts (mc->m.mnt_opts, test_opts)) {
 	       errors |= umount_one (mc->m.mnt_fsname, mc->m.mnt_dir,
 				     mc->m.mnt_type, mc->m.mnt_opts, mc);
 	  }
@@ -441,6 +453,7 @@ static struct option longopts[] =
   { "force", 0, 0, 'f' },
   { "help", 0, 0, 'h' },
   { "no-mtab", 0, 0, 'n' },
+  { "test-opts", 1, 0, 'O' },
   { "verbose", 0, 0, 'v' },
   { "version", 0, 0, 'V' },
   { "read-only", 0, 0, 'r' },
@@ -452,7 +465,7 @@ static void
 usage (FILE *fp, int n)
 {
   fprintf (fp, _("Usage: umount [-hV]\n"
-	       "       umount -a [-f] [-r] [-n] [-v] [-t vfstypes]\n"
+	       "       umount -a [-f] [-r] [-n] [-v] [-t vfstypes] [-O opts]\n"
 	       "       umount [-f] [-r] [-n] [-v] special | node...\n"));
   exit (n);
 }
@@ -522,7 +535,6 @@ get_value(string_list list, char *s) {
 	}
 	return 0;
 }
-/*=======================================================================*/
 
 static int
 umount_file (char *arg) {
@@ -531,19 +543,27 @@ umount_file (char *arg) {
 	string_list options;
 	int fstab_has_user, fstab_has_users, fstab_has_owner, ok;
 
-	file = canonicalize (arg); /* mtab paths are canonicalized */
+	file = canonicalize(arg); /* mtab paths are canonicalized */
 	if (verbose > 1)
 		printf(_("Trying to umount %s\n"), file);
 
-	mc = getmntfilesbackward (file, NULL);
+	mc = getmntdirbackward(file, NULL);
+	if (!mc)
+		mc = getmntdevbackward(file, NULL);
 	if (!mc && verbose)
 		printf(_("Could not find %s in mtab\n"), file);
 
 	if (suid) {
+		char *mtab_user = NULL;
+
 		if (!mc)
-			die (2, _("umount: %s is not mounted (according to mtab)"), file);
-		if (getmntfilesbackward (file, mc))
-			die (2, _("umount: it seems %s is mounted multiple times"), file);
+			die(2,
+			    _("umount: %s is not mounted (according to mtab)"),
+			    file);
+		if (!is_mounted_once(file))
+			die(2,
+			    _("umount: it seems %s is mounted multiple times"),
+			    file);
 
 		/* If fstab contains the two lines
 		   /dev/sda1 /mnt/zip auto user,noauto  0 0
@@ -555,10 +575,12 @@ umount_file (char *arg) {
 		if (!fs) {
 			if (!getfsspec (file) && !getfsfile (file))
 				die (2,
-				     _("umount: %s is not in the fstab (and you are not root)"),
+				     _("umount: %s is not in the fstab "
+				       "(and you are not root)"),
 				     file);
 			else
-				die (2, _("umount: %s mount disagrees with the fstab"), file);
+				die (2, _("umount: %s mount disagrees with "
+					  "the fstab"), file);
 		}
 
 		/* User mounting and unmounting is allowed only
@@ -583,7 +605,6 @@ umount_file (char *arg) {
 
 		if (!ok && (fstab_has_user || fstab_has_owner)) {
 			char *user = getusername();
-			char *mtab_user;
 
 			options = parse_list (mc->m.mnt_opts);
 			mtab_user = get_value(options, "user=");
@@ -592,7 +613,8 @@ umount_file (char *arg) {
 				ok = 1;
 		}
 		if (!ok)
-			die (2, _("umount: only root can unmount %s from %s"),
+			die (2, _("umount: only %s can unmount %s from %s"),
+			     mtab_user ? mtab_user : "root",
 			     fs->m.mnt_fsname, fs->m.mnt_dir);
 	}
 
@@ -606,7 +628,7 @@ int
 main (int argc, char *argv[]) {
 	int c;
 	int all = 0;
-	char *types = NULL;
+	char *types = NULL, *test_opts = NULL;
 	int result = 0;
 
 	sanitize_env();
@@ -614,7 +636,9 @@ main (int argc, char *argv[]) {
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 
-	while ((c = getopt_long (argc, argv, "adfhlnrt:vV",
+	umask(033);
+
+	while ((c = getopt_long (argc, argv, "adfhlnrt:O:vV",
 				 longopts, NULL)) != -1)
 		switch (c) {
 		case 'a':		/* umount everything */
@@ -636,6 +660,9 @@ main (int argc, char *argv[]) {
 		case 'n':		/* do not write in /etc/mtab */
 			++nomtab;
 			break;
+		case 'O':		/* specify file system options */
+			test_opts = optarg;
+			break;
 		case 'r':		/* remount read-only if umount fails */
 			++remount;
 			break;
@@ -665,9 +692,10 @@ main (int argc, char *argv[]) {
 	argv += optind;
 
 	if (all) {
+		/* nodev stuff: sysfs, usbfs, oprofilefs, ... */
 		if (types == NULL)
-			types = "noproc";
-		result = umount_all (types);
+			types = "noproc,nodevfs,nodevpts";
+		result = umount_all (types, test_opts);
 	} else if (argc < 1) {
 		usage (stderr, 2);
 	} else while (argc--) {
