@@ -25,20 +25,29 @@
  * - 64bit clean patch
  * 3Feb2001 Andrew Morton <andrewm@uow.edu.au>
  * - -M option to write profile multiplier.
+ * 2001-11-07 Werner Almesberger <wa@almesberger.net>
+ * - byte order auto-detection and -n option
+ * 2001-11-09 Werner Almesberger <wa@almesberger.net>
+ * - skip step size (index 0)
+ * 2002-03-09 John Levon <moz@compsoc.man.ac.uk>
+ * - make maplineno do something
+ * 2002-11-28 Mads Martin Joergensen +
+ * - also try /boot/System.map-`uname -r`
+ * 2003-04-09 Werner Almesberger <wa@almesberger.net>
+ * - fixed off-by eight error and improved heuristics in byte order detection
  */
 
 #include <errno.h>
 #include <stdio.h>
+#include <fcntl.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <fcntl.h>
+#include <sys/utsname.h>
 #include "nls.h"
 
-#define RELEASE "2.0, May 1996"
-
 #define S_LEN 128
 
 static char *prgname;
@@ -46,53 +55,95 @@ static char *prgname;
 /* These are the defaults */
 static char defaultmap[]="/usr/src/linux/System.map";
 static char defaultpro[]="/proc/profile";
-static char optstring[]="M:m:p:itvarV";
+static char optstring[]="M:m:np:itvarVb";
 
-static void
-usage(void) {
-  fprintf(stderr,
-		  _("%s: Usage: \"%s [options]\n"
-		  "\t -m <mapfile>  (default = \"%s\")\n"
-		  "\t -p <pro-file> (default = \"%s\")\n"
-		  "\t -M <mult>     set the profiling multiplier to <mult>\n"
-		  "\t -i            print only info about the sampling step\n"
-		  "\t -v            print verbose data\n"
-		  "\t -a            print all symbols, even if count is 0\n"
-		  "\t -r            reset all the counters (root only)\n"
-		  "\t -V            print version and exit\n")
-		  ,prgname,prgname,defaultmap,defaultpro);
-  exit(1);
+static void *
+xmalloc (size_t size) {
+	void *t;
+
+	if (size == 0)
+		return NULL;
+
+	t = malloc (size);
+	if (t == NULL) {
+		fprintf(stderr, _("out of memory"));
+		exit(1);
+	}
+
+	return t;
 }
 
 static FILE *
 myopen(char *name, char *mode, int *flag) {
-	static char cmdline[S_LEN];
+	int len = strlen(name);
 
-	if (!strcmp(name+strlen(name)-3,".gz")) {
-		*flag=1;
-		sprintf(cmdline,"zcat %s", name);
-		return popen(cmdline,mode);
+	if (!strcmp(name+len-3,".gz")) {
+		FILE *res;
+		char *cmdline = xmalloc(len+6);
+		sprintf(cmdline, "zcat %s", name);
+		res = popen(cmdline,mode);
+		free(cmdline);
+		*flag = 1;
+		return res;
 	}
-	*flag=0;
+	*flag = 0;
 	return fopen(name,mode);
 }
 
+#ifndef BOOT_SYSTEM_MAP
+#define BOOT_SYSTEM_MAP "/boot/System.map-"
+#endif
+
+static char *
+boot_uname_r_str(void) {
+	struct utsname uname_info;
+	char *s;
+	size_t len;
+
+	if (uname(&uname_info))
+		return "";
+	len = strlen(BOOT_SYSTEM_MAP) + strlen(uname_info.release) + 1;
+	s = xmalloc(len);
+	strcpy(s, BOOT_SYSTEM_MAP);
+	strcat(s, uname_info.release);
+	return s;
+}
+
+static void
+usage(void) {
+	fprintf(stderr,
+		_("%s: Usage: \"%s [options]\n"
+		  "\t -m <mapfile>  (defaults: \"%s\" and\n\t\t\t\t  \"%s\")\n"
+		  "\t -p <pro-file> (default: \"%s\")\n"
+		  "\t -M <mult>     set the profiling multiplier to <mult>\n"
+		  "\t -i            print only info about the sampling step\n"
+		  "\t -v            print verbose data\n"
+		  "\t -a            print all symbols, even if count is 0\n"
+		  "\t -b            print individual histogram-bin counts\n"
+		  "\t -r            reset all the counters (root only)\n"
+		  "\t -n            disable byte order auto-detection\n"
+		  "\t -V            print version and exit\n"),
+		prgname, prgname, defaultmap, boot_uname_r_str(), defaultpro);
+	exit(1);
+}
+
 int
-main (int argc, char **argv) {
+main(int argc, char **argv) {
 	FILE *map;
 	int proFd;
 	char *mapFile, *proFile, *mult=0;
-	unsigned long len=0, add0=0, indx=0;
+	unsigned long len=0, add0=0, indx=1;
 	unsigned int step;
 	unsigned int *buf, total, fn_len;
 	unsigned long fn_add, next_add;          /* current and next address */
 	char fn_name[S_LEN], next_name[S_LEN];   /* current and next name */
 	char mode[8];
 	int c;
-	int optAll=0, optInfo=0, optReset=0, optVerbose=0;
+	int optAll=0, optInfo=0, optReset=0, optVerbose=0, optNative=0, optBins=0;
 	char mapline[S_LEN];
 	int maplineno=1;
 	int popenMap;   /* flag to tell if popen() has been used */
+	int header_printed;
 
 #define next (current^1)
 
@@ -100,22 +151,45 @@ main (int argc, char **argv) {
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 
-	prgname=argv[0];
-	proFile=defaultpro;
-	mapFile=defaultmap;
+	prgname = argv[0];
+	proFile = defaultpro;
+	mapFile = defaultmap;
 
-	while ((c=getopt(argc,argv,optstring))!=-1) {
+	while ((c = getopt(argc, argv, optstring)) != -1) {
 		switch(c) {
-		case 'm': mapFile=optarg; break;
-		case 'p': proFile=optarg; break;
-		case 'a': optAll++;       break;
-		case 'i': optInfo++;      break;
-		case 'M': mult=optarg;    break;
-		case 'r': optReset++;     break;
-		case 'v': optVerbose++;   break;
-		case 'V': printf(_("%s Version %s\n"),prgname,RELEASE);
+		case 'm':
+			mapFile = optarg;
+			break;
+		case 'n':
+			optNative++;
+			break;
+		case 'p':
+			proFile = optarg;
+			break;
+		case 'a':
+			optAll++;
+			break;
+		case 'b':
+			optBins++;
+			break;
+		case 'i':
+			optInfo++;
+			break;
+		case 'M':
+			mult = optarg;
+			break;
+		case 'r':
+			optReset++;
+			break;
+		case 'v':
+			optVerbose++;
+			break;
+		case 'V':
+			printf(_("%s version %s\n"), prgname,
+			       UTIL_LINUX_VERSION);
 			exit(0);
-		default: usage();
+		default:
+			usage();
 		}
 	}
 
@@ -152,15 +226,15 @@ main (int argc, char **argv) {
 	/*
 	 * Use an fd for the profiling buffer, to skip stdio overhead
 	 */
-	if ( ((proFd=open(proFile,O_RDONLY)) < 0)
+	if (((proFd=open(proFile,O_RDONLY)) < 0)
 	     || ((int)(len=lseek(proFd,0,SEEK_END)) < 0)
-	     || (lseek(proFd,0,SEEK_SET)<0) ) {
+	     || (lseek(proFd,0,SEEK_SET) < 0)) {
 		fprintf(stderr,"%s: %s: %s\n",prgname,proFile,strerror(errno));
 		exit(1);
 	}
 
-	if ( !(buf=malloc(len)) ) {
-		fprintf(stderr,"%s: malloc(): %s\n",prgname, strerror(errno));
+	if (!(buf=malloc(len))) {
+		fprintf(stderr,"%s: malloc(): %s\n", prgname, strerror(errno));
 		exit(1);
 	}
 
@@ -170,29 +244,64 @@ main (int argc, char **argv) {
 	}
 	close(proFd);
 
-	step=buf[0];
+	if (!optNative) {
+		int entries = len/sizeof(*buf);
+		int big = 0,small = 0,i;
+		unsigned *p;
+
+		for (p = buf+1; p < buf+entries; p++) {
+			if (*p & ~0U << (sizeof(*buf)*4))
+				big++;
+			if (*p & ((1 << (sizeof(*buf)*4))-1))
+				small++;
+		}
+		if (big > small) {
+			fprintf(stderr,"Assuming reversed byte order. "
+			   "Use -n to force native byte order.\n");
+			for (p = buf; p < buf+entries; p++)
+				for (i = 0; i < sizeof(*buf)/2; i++) {
+					unsigned char *b = (unsigned char *) p;
+					unsigned char tmp;
+
+					tmp = b[i];
+					b[i] = b[sizeof(*buf)-i-1];
+					b[sizeof(*buf)-i-1] = tmp;
+				}
+		}
+	}
+
+	step = buf[0];
 	if (optInfo) {
-		printf(_("Sampling_step: %i\n"),step);
+		printf(_("Sampling_step: %i\n"), step);
 		exit(0);
 	} 
 
-	total=0;
+	total = 0;
 
-	if (!(map=myopen(mapFile,"r",&popenMap))) {
-		fprintf(stderr,"%s: ",prgname);perror(mapFile);
+	map = myopen(mapFile, "r", &popenMap);
+	if (map == NULL && mapFile == defaultmap) {
+		mapFile = boot_uname_r_str();
+		map = myopen(mapFile, "r", &popenMap);
+	}
+	if (map == NULL) {
+		int errsv = errno;
+		fprintf(stderr, "%s: ", prgname);
+		errno = errsv;
+		perror(mapFile);
 		exit(1);
 	}
 
-	while(fgets(mapline,S_LEN,map)) {
-		if (sscanf(mapline,"%lx %s %s",&fn_add,mode,fn_name)!=3) {
+	while (fgets(mapline,S_LEN,map)) {
+		if (sscanf(mapline,"%lx %s %s",&fn_add,mode,fn_name) != 3) {
 			fprintf(stderr,_("%s: %s(%i): wrong map line\n"),
-				prgname,mapFile, maplineno);
+				prgname, mapFile, maplineno);
 			exit(1);
 		}
 		if (!strcmp(fn_name,"_stext")) /* only elf works like this */ {
-			add0=fn_add;
+			add0 = fn_add;
 			break;
 		}
+		maplineno++;
 	}
 
 	if (!add0) {
@@ -204,7 +313,7 @@ main (int argc, char **argv) {
 	/*
 	 * Main loop.
 	 */
-	while(fgets(mapline,S_LEN,map)) {
+	while (fgets(mapline,S_LEN,map)) {
 		unsigned int this=0;
 
 		if (sscanf(mapline,"%lx %s %s",&next_add,mode,next_name)!=3) {
@@ -212,26 +321,50 @@ main (int argc, char **argv) {
 				prgname,mapFile, maplineno);
 			exit(1);
 		}
+		header_printed = 0;
 
 		/* ignore any LEADING (before a '[tT]' symbol is found)
 		   Absolute symbols */
-		if (*mode == 'A' && total == 0) continue;
-		if (*mode!='T' && *mode!='t') break;/* only text is profiled */
+		if ((*mode == 'A' || *mode == '?') && total == 0) continue;
+		if (*mode != 'T' && *mode != 't' &&
+		    *mode != 'W' && *mode != 'w')
+			break;	/* only text is profiled */
 
-		while (indx < (next_add-add0)/step)
+		if (indx >= len / sizeof(*buf)) {
+			fprintf(stderr, _("%s: profile address out of range. "
+					  "Wrong map file?\n"), prgname);
+			exit(1);
+		}
+
+		while (indx < (next_add-add0)/step) {
+			if (optBins && (buf[indx] || optAll)) {
+				if (!header_printed) {
+					printf ("%s:\n", fn_name);
+					header_printed = 1;
+				}
+				printf ("\t%lx\t%u\n", (indx - 1)*step + add0, buf[indx]);
+			}
 			this += buf[indx++];
+		}
 		total += this;
 
-		fn_len = next_add-fn_add;
-		if (fn_len && (this || optAll)) {
-			if (optVerbose)
-				printf("%08lx %-40s %6i %8.4f\n", fn_add,
-				       fn_name,this,this/(double)fn_len);
-			else
-				printf("%6i %-40s %8.4f\n",
-				       this,fn_name,this/(double)fn_len);
+		if (optBins) {
+			if (optVerbose || this > 0)
+				printf ("  total\t\t\t\t%u\n", this);
+		} else {
+			fn_len = next_add-fn_add;
+			if (fn_len && (this || optAll)) {
+				if (optVerbose)
+					printf("%08lx %-40s %6i %8.4f\n", fn_add,
+					       fn_name,this,this/(double)fn_len);
+				else
+					printf("%6i %-40s %8.4f\n",
+					       this,fn_name,this/(double)fn_len);
+			}
 		}
 		fn_add=next_add; strcpy(fn_name,next_name);
+
+		maplineno++;
 	}
 	/* trailer */
 	if (optVerbose)
