@@ -57,12 +57,12 @@
 #include <locale.h>
 #include "xstrncpy.h"
 #include "nls.h"
+#include "widechar.h"
 
 #define _REGEX_RE_COMP
 #include <regex.h>
 #undef _REGEX_RE_COMP
 
-/* #define MOREHELPFILE "/usr/lib/more.help" */
 #define VI		"vi"	/* found on the user's path */
 
 #define Fopen(s,m)	(Currline = 0,file_pos=0,fopen(s,m))
@@ -98,7 +98,7 @@ int  pr(char *s1);
 int  printd (int n);
 void set_tty(void);
 void reset_tty(void);
-void ttyin (char buf[], register int nmax, char pchar);
+void ttyin (unsigned char buf[], register int nmax, char pchar);
 int  number(char *cmd);
 int  readch (void);
 int  get_line(register FILE *f, int *length);
@@ -508,7 +508,7 @@ checkf (fs, clearfirst)
 {
 	struct stat stbuf;
 	register FILE *f;
-	char c;
+	int c;
 
 	if (stat (fs, &stbuf) == -1) {
 		(void)fflush(stdout);
@@ -529,7 +529,7 @@ checkf (fs, clearfirst)
 	if (magic(f, fs))
 		return((FILE *)NULL);
 	c = Getc(f);
-	*clearfirst = c == '\f';
+	*clearfirst = (c == '\f');
 	Ungetc (c, f);
 	if ((file_size = stbuf.st_size) == 0)
 		file_size = LONG_MAX;
@@ -548,6 +548,10 @@ magic(f, fs)
 {
 	char twobytes[2];
 
+	/* don't try to look ahead if the input is unseekable */
+	if (fseek(f, 0L, SEEK_SET))
+		return(0);
+
 	if (fread(twobytes, 2, 1, f) == 1) {
 		switch(twobytes[0] + (twobytes[1]<<8)) {
 		case OMAGIC:	/* 0407 */
@@ -862,15 +866,30 @@ static void prompt (char *filename)
 }
 
 /*
-** Get a logical line
-*/
+ * Get a logical line
+ */
 
 int get_line(register FILE *f, int *length)
 {
-    register int	c;
-    register char	*p;
-    register int	column;
-    static int		colflg;
+    int	c;
+    char *p;
+    int	column;
+    static int colflg;
+
+#ifdef ENABLE_WIDECHAR
+    int i;
+    wchar_t wc;
+    int wc_width;
+    mbstate_t state, state_bak;		/* Current status of the stream. */
+    unsigned char mbc[MB_LEN_MAX];	/* Buffer for one multibyte char. */
+    size_t mblength;			/* Byte length of multibyte char. */
+    size_t mbc_pos = 0;			/* Postion of the MBC. */
+    int use_mbc_buffer_flag = 0;	/* If 1, mbc has data. */
+    int break_flag = 0;			/* If 1, exit while(). */
+    long file_pos_bak = Ftell (f);
+
+    memset (&state, '\0', sizeof (mbstate_t));
+#endif
 
     p = Line;
     column = 0;
@@ -880,6 +899,58 @@ int get_line(register FILE *f, int *length)
 	c = Getc (f);
     }
     while (p < &Line[LINSIZ - 1]) {
+#ifdef ENABLE_WIDECHAR
+	if (fold_opt && use_mbc_buffer_flag && MB_CUR_MAX > 1) {
+	    use_mbc_buffer_flag = 0;
+	    state_bak = state;
+	    mbc[mbc_pos++] = c;
+process_mbc:
+	    mblength = mbrtowc (&wc, mbc, mbc_pos, &state);
+
+	    switch (mblength) {
+	      case (size_t)-2:	  /* Incomplete multibyte character. */
+		use_mbc_buffer_flag = 1;
+		state = state_bak;
+		break;
+
+	      case (size_t)-1:	  /* Invalid as a multibyte character. */
+		*p++ = mbc[0];
+		state = state_bak;
+		column++;
+		file_pos_bak++;
+		
+		if (column >= Mcol) {
+		    Fseek (f, file_pos_bak);
+		} else {
+		    memmove (mbc, mbc + 1, --mbc_pos);
+		    if (mbc_pos > 0) {
+			mbc[mbc_pos] = '\0';
+			goto process_mbc;
+		    }
+		}
+		break;
+
+	      default:
+		wc_width = wcwidth (wc);
+
+		if (column + wc_width > Mcol) {
+		    Fseek (f, file_pos_bak);
+		    break_flag = 1;
+		} else {
+		    for (i = 0; i < mbc_pos; i++)
+		      *p++ = mbc[i];
+		    if (wc_width > 0)
+		      column += wc_width;
+		}
+	    }
+
+	    if (break_flag || column >= Mcol)
+	      break;
+
+	    c = Getc (f);
+	    continue;
+	}
+#endif 
 	if (c == EOF) {
 	    if (p > Line) {
 		*p = '\0';
@@ -893,6 +964,7 @@ int get_line(register FILE *f, int *length)
 	    Currline++;
 	    break;
 	}
+
 	*p++ = c;
 #if 0
 	if (c == '\033') {      /* ESC */
@@ -923,8 +995,7 @@ int get_line(register FILE *f, int *length)
 		    }
 		    if (column >= promptlen) promptlen = 0;
 		}
-	    }
-	    else
+	    } else
 		column = 1 + (column | 7);
 	} else if (c == '\b' && column > 0) {
 	    column--;
@@ -945,30 +1016,43 @@ int get_line(register FILE *f, int *length)
 	} else if (c == EOF) {
 	    *length = p - Line;
 	    return (column);
-	}
-#if 0
-	else {
-		char ch[2]; /* for mblen() */
+	} else {
+#ifdef ENABLE_WIDECHAR
+	    if (fold_opt && MB_CUR_MAX > 1) {
+		memset (mbc, '\0', MB_LEN_MAX);
+		mbc_pos = 0;
+		mbc[mbc_pos++] = c;
+		state_bak = state;
 
-		ch[0] = c;
-		ch[1] = '\0';
-		if (mblen(ch, 1) <= 0) {    /* multibyte, maybe 2 columns */
-			if (column == Mcol-1 && fold_opt) {
-				p--;
-				Ungetc(c, f);
-				break;
-			}
-			*p++ = Getc(f);
-			column += 2;
-		} else if (c & 0x80) {      /* maybe SJIS KANA, or utf-8 ? */
-			column++;
-		} else if (isprint(c))
-			column++;
-	}
-#else
-	else if (isprint(c))
-		column++;
+		mblength = mbrtowc (&wc, mbc, mbc_pos, &state);
+
+		/* The value of mblength is always less than 2 here. */
+		switch (mblength) {
+		  case (size_t)-2:
+		    p--;
+		    file_pos_bak = Ftell (f) - 1;
+		    state = state_bak;
+		    use_mbc_buffer_flag = 1;
+		    break;
+
+		  case (size_t)-1:
+		    state = state_bak;
+		    column++;
+		    break;
+
+		  default:
+		    wc_width = wcwidth (wc);
+		    if (wc_width > 0)
+		      column += wc_width;
+		}
+	    } else
 #endif
+	      {
+		if (isprint(c))
+		   column++;
+	      }
+	}
+
 	if (column >= Mcol && fold_opt)
 		break;
 	c = Getc (f);
@@ -1015,11 +1099,11 @@ void erasep (register int col)
 ** Erase the current line entirely
 */
 
-void kill_line ()
+void kill_line()
 {
-    erasep (0);
+    erasep(0);
     if (!eraseln || dumb)
-	putchar ('\r');
+	putchar('\r');
 }
 
 /*
@@ -1131,7 +1215,6 @@ int command (char *filename, register FILE *f)
     register int retval = 0;
     register int c;
     char colonch;
-    FILE *helpf;
     int done;
     char comchar, cmdbuf[80];
 
@@ -1182,12 +1265,10 @@ int command (char *filename, register FILE *f)
 		xprintf ("\n");
 		if (clreol)
 			cleareol ();
-		xprintf (_("...back %d page"), nlines);
-		if (nlines > 1)
-			pr ("s\n");
+		if (nlines != 1)
+			xprintf (_("...back %d pages"), nlines);
 		else
-			pr ("\n");
-
+			xprintf (_("...back 1 page"));
 		if (clreol)
 			cleareol ();
 		pr ("\n");
@@ -1220,6 +1301,7 @@ int command (char *filename, register FILE *f)
 	    end_it (0);
 	case 's':
 	case 'f':
+	case ctrl('F'):
 	    if (nlines == 0) nlines++;
 	    if (comchar == 'f')
 		nlines *= dlines;
@@ -1228,11 +1310,10 @@ int command (char *filename, register FILE *f)
 	    xprintf ("\n");
 	    if (clreol)
 		cleareol ();
-	    xprintf (_("...skipping %d line"), nlines);
-	    if (nlines > 1)
-		pr ("s\n");
+	    if (nlines == 1)
+		    xprintf (_("...skipping one line"));
 	    else
-		pr ("\n");
+		    xprintf (_("...skipping %d lines"), nlines);
 
 	    if (clreol)
 		cleareol ();
@@ -1306,35 +1387,47 @@ int command (char *filename, register FILE *f)
 	    break;
 	case '?':
 	case 'h':
-	    helpf = NULL;
-	    {
-		    char *lang;
-		    char hlpfile[sizeof(MOREHELPFILE)+4];
-
-		    lang = getenv("LANGUAGE");
-		    if (!lang || *lang == '\0')
-			    lang = setlocale(LC_MESSAGES, "");
-		    if (!lang || *lang == '\0')
-			    lang = getenv("LANG");
-		    if (lang && strlen(lang) > 1) {
-			    strcpy(hlpfile, MOREHELPFILE);
-			    strcat(hlpfile, ".");
-			    strncat(hlpfile, lang, 2);
-			    helpf = fopen (hlpfile, "r");
-		    }
-	    }
-	    if (helpf == NULL)
-		    helpf = fopen (MOREHELPFILE, "r");
-	    if (helpf == NULL)
-		error (_("Can't open help file"));
-	    if (noscroll) doclear ();
-	    copy_file (helpf);
-	    fclose (helpf);
-	    prompt (filename);
+	    if (noscroll) doclear();
+	    xprintf(_("\n"
+"Most commands optionally preceded by integer argument k.  "
+"Defaults in brackets.\n"
+"Star (*) indicates argument becomes new default.\n"));
+	    xprintf("---------------------------------------"
+		"----------------------------------------\n");
+	    xprintf(_(
+"<space>                 Display next k lines of text [current screen size]\n"
+"z                       Display next k lines of text [current screen size]*\n"
+"<return>                Display next k lines of text [1]*\n"
+"d or ctrl-D             Scroll k lines [current scroll size, initially 11]*\n"
+"q or Q or <interrupt>   Exit from more\n"
+"s                       Skip forward k lines of text [1]\n"
+"f                       Skip forward k screenfuls of text [1]\n"
+"b or ctrl-B             Skip backwards k screenfuls of text [1]\n"
+"'                       Go to place where previous search started\n"
+"=                       Display current line number\n"
+"/<regular expression>   Search for kth occurrence of regular expression [1]\n"
+"n                       Search for kth occurrence of last r.e [1]\n"
+"!<cmd> or :!<cmd>       Execute <cmd> in a subshell\n"
+"v                       Start up /usr/bin/vi at current line\n"
+"ctrl-L                  Redraw screen\n"
+":n                      Go to kth next file [1]\n"
+":p                      Go to kth previous file [1]\n"
+":f                      Display current file name and line number\n"
+".                       Repeat previous command\n"));
+	    xprintf("---------------------------------------"
+		"----------------------------------------\n");
+	    prompt(filename);
 	    break;
 	case 'v':	/* This case should go right before default */
 	    if (!no_intty) {
-		    char *editor;
+		    /*
+		     * Earlier: call vi +n file. This also works for emacs.
+		     * POSIX: call vi -c n file (when editor is vi or ex).
+		     */
+		    char *editor, *p;
+		    int n = (Currline - dlines <= 0 ? 1 :
+			     Currline - (dlines + 1) / 2);
+		    int split = 0;
 
 		    editor = getenv("VISUAL");
 		    if (editor == NULL || *editor == '\0')
@@ -1342,21 +1435,40 @@ int command (char *filename, register FILE *f)
 		    if (editor == NULL || *editor == '\0')
 			    editor = VI;
 
-		    kill_line ();
-		    cmdbuf[0] = '+';
-		    scanstr (Currline - dlines <= 0 ? 1
-			     : Currline - (dlines + 1) / 2, &cmdbuf[1]);
-		    pr (editor); putchar (' ');
-		    pr (cmdbuf); putchar (' '); pr (fnames[fnum]);
-		    execute (filename, editor, editor, cmdbuf, fnames[fnum], 0);
+		    p = strrchr(editor, '/');
+		    if (p)
+			    p++;
+		    else
+			    p = editor;
+		    if (!strcmp(p, "vi") || !strcmp(p, "ex")) {
+			    strcpy(cmdbuf, "-c ");
+			    scanstr(n, &cmdbuf[3]);
+			    split = 1;
+		    } else {
+			    cmdbuf[0] = '+';
+			    scanstr(n, &cmdbuf[1]);
+		    }
+
+		    kill_line();
+		    pr(editor); putchar(' ');
+		    pr(cmdbuf); putchar(' '); pr(fnames[fnum]);
+		    if (split) {
+			    cmdbuf[2] = 0;
+			    execute(filename, editor, editor, cmdbuf,
+				    cmdbuf+3, fnames[fnum], (char *)0);
+		    } else
+			    execute(filename, editor, editor,
+				    cmdbuf, fnames[fnum], (char *)0);
 		    break;
 	    }
+	    /* fall through */
 	default:
 	    if (dum_opt) {
    		kill_line ();
 		if (Senter && Sexit) {
 		    my_putstring (Senter);
-		    promptlen = pr (_("[Press 'h' for instructions.]")) + (2 * soglitch);
+		    promptlen = pr (_("[Press 'h' for instructions.]"))
+			    + 2 * soglitch;
 		    my_putstring (Sexit);
 		}
 		else
@@ -1577,7 +1689,7 @@ void search(char buf[], FILE *file, register int n)
 }
 
 /*VARARGS2*/
-void execute (char * filename, char * cmd, ...)
+void execute (char *filename, char *cmd, ...)
 {
 	int id;
 	int n;
@@ -1676,8 +1788,10 @@ void skipf (register int nskip)
     pr ("\n");
     if (clreol)
 	cleareol ();
-    pr (_("...Skipping "));
-    pr (nskip > 0 ? _("to file ") : _("back to file "));
+    if (nskip > 0)
+	    pr (_("...Skipping to file "));
+    else
+	    pr (_("...Skipping back to file "));
     pr (fnames[fnum]);
     pr ("\n");
     if (clreol)
@@ -1709,13 +1823,13 @@ retry:
 	     * Wait until we're in the foreground before we save the
 	     * the terminal modes.
 	     */
-	    if (ioctl(fileno(stdout), TIOCGPGRP, &tgrp) < 0) {
-		 perror("TIOCGPGRP");
-		 exit(1);
+	    if ((tgrp = tcgetpgrp(fileno(stdout))) < 0) {
+		perror("tcgetpgrp");
+		exit(1);
 	    }
 	    if (tgrp != getpgrp(0)) {
-		 kill(0, SIGTTOU);
-		 goto retry;
+		kill(0, SIGTTOU);
+		goto retry;
 	    }
 	}
 #endif
@@ -1815,10 +1929,10 @@ retry:
 }
 
 int readch () {
-	char c;
+	unsigned char c;
 
 	errno = 0;
-	if (read (2, &c, 1) <= 0) {
+	if (read (fileno(stderr), &c, 1) <= 0) {
 		if (errno != EINTR)
 			end_it(0);
 		else
@@ -1830,18 +1944,17 @@ int readch () {
 static char *BS = "\b";
 static char *BSB = "\b \b";
 static char *CARAT = "^";
-#define ERASEONECHAR \
+#define ERASEONECOLUMN \
     if (docrterase) \
 	errwrite(BSB); \
     else \
 	errwrite(BS);
 
-void ttyin (char buf[], register int nmax, char pchar) {
-    char *sp;
-    char c;
+void ttyin (unsigned char buf[], register int nmax, char pchar) {
+    unsigned char *sp;
+    int c;
     int slash = 0;
     int	maxlen;
-    char cbuf;
 
     sp = buf;
     maxlen = 0;
@@ -1853,12 +1966,59 @@ void ttyin (char buf[], register int nmax, char pchar) {
 	}
 	else if (((cc_t) c == otty.c_cc[VERASE]) && !slash) {
 	    if (sp > buf) {
-		--promptlen;
-		ERASEONECHAR
-		--sp;
+#ifdef ENABLE_WIDECHAR
+		if (MB_CUR_MAX > 1)
+		  {
+		    wchar_t wc;
+		    size_t pos = 0, mblength;
+		    mbstate_t state, state_bak;
+
+		    memset (&state, '\0', sizeof (mbstate_t));
+
+		    while (1) {
+			 state_bak = state;
+			 mblength = mbrtowc (&wc, buf + pos, sp - buf, &state);
+
+			 state = (mblength == (size_t)-2
+				 || mblength == (size_t)-1) ? state_bak : state;
+			 mblength = (mblength == (size_t)-2
+				     || mblength == (size_t)-1
+				     || mblength == 0) ? 1 : mblength;
+
+			 if (buf + pos + mblength >= sp)
+			 break;
+
+			 pos += mblength;
+		    }
+
+		    if (mblength == 1) {
+		      ERASEONECOLUMN
+		    }
+		    else {
+			int wc_width;
+			wc_width = wcwidth (wc);
+			wc_width = (wc_width < 1) ? 1 : wc_width;
+			while (wc_width--) {
+			    ERASEONECOLUMN
+			}
+		    }
+
+		    while (mblength--) {
+			--promptlen;
+			--sp;
+		    }
+		  }
+		else
+#endif
+		  {
+		    --promptlen;
+		    ERASEONECOLUMN
+		    --sp;
+		  }
+
 		if ((*sp < ' ' && *sp != '\n') || *sp == RUBOUT) {
 		    --promptlen;
-		    ERASEONECHAR
+		    ERASEONECOLUMN
 		}
 		continue;
 	    }
@@ -1889,7 +2049,7 @@ void ttyin (char buf[], register int nmax, char pchar) {
 	}
 	if (slash && ((cc_t) c == otty.c_cc[VKILL]
 		   || (cc_t) c == otty.c_cc[VERASE])) {
-	    ERASEONECHAR
+	    ERASEONECOLUMN
 	    --sp;
 	}
 	if (c != '\\')
@@ -1900,8 +2060,8 @@ void ttyin (char buf[], register int nmax, char pchar) {
 	    errwrite(CARAT);
 	    promptlen++;
 	}
-	cbuf = c;
 	if (c != '\n' && c != ESC) {
+	    char cbuf = c;
 	    errwrite1(&cbuf);
 	    promptlen++;
 	}
